Package: abba
Title: Batch Execution of R Programs on 'Kubernetes', 'SLURM', and
        'Posit Workbench'
Version: 0.2.0
Authors@R: 
    c(
    person(given = "Eli",
           family = "Miller",
           role = "aut",
           email = "Eli.Miller@AtorusResearch.com",
           comment = c(ORCID = "0000-0002-2127-9456")),
    person(given = "Mike",
           family = "Stackhouse",
           role = c("aut", "cre"),
           email = "mike.stackhouse@atorusresearch.com",
           comment = c(ORCID = "0000-0001-6030-723X")),
    person(given = "Ross",
           family = "Didenko",
           role = c("aut"),
           email = "Rostyslav.Didenko@atorusresearch.com"),
    person(given = "Yevhenii",
           family = "Boiko",
           role = c("aut"),
           email = "Yevhenii.Boiko@atorusresearch.com"),
    person("Atorus Research, Inc.", role = "cph")
           )
Description: Submit and monitor batch execution of R programs across
    distributed computing backends including 'Kubernetes', 'SLURM', and
    'Posit Workbench'. Provides end-user job submission functions, cluster
    interface functions using 'kubectl' and 'SLURM' commands, and a
    'plumber' API template for secure identity segregation. Supports
    parallel and sequential batch execution, file-based caching to skip
    unchanged programs, and 'logrx' integration for execution logging.
License: Apache License (>= 2)
URL: https://atorus-research.github.io/abba/,
        https://github.com/atorus-research/abba
BugReports: https://github.com/atorus-research/abba/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: yaml, uuid, stringr, httr2, magrittr, tools, rstudioapi,
        digest, tidyr
Suggests: testthat (>= 3.0.0), mockery, knitr, rmarkdown, logrx
Config/testthat/edition: 3
VignetteBuilder: knitr
SystemRequirements: kubectl (optional, for Kubernetes backend), SLURM
        (optional, sbatch/scontrol/sacct/squeue for SLURM backend),
        Posit Workbench (optional, for Workbench backend)
NeedsCompilation: no
Packaged: 2026-04-23 14:18:31 UTC; mstackhouse
Author: Eli Miller [aut] (ORCID: <https://orcid.org/0000-0002-2127-9456>),
  Mike Stackhouse [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6030-723X>),
  Ross Didenko [aut],
  Yevhenii Boiko [aut],
  Atorus Research, Inc. [cph]
Maintainer: Mike Stackhouse <mike.stackhouse@atorusresearch.com>
Repository: CRAN
Date/Publication: 2026-04-28 18:10:14 UTC
Built: R 4.5.2; ; 2026-05-01 18:31:31 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(abba_get_batch_log)
export(abba_get_batch_status)
export(abba_get_job_log)
export(abba_get_job_status)
export(abba_get_k8s_batch_log_local)
export(abba_get_k8s_batch_status_local)
export(abba_get_k8s_job_log_local)
export(abba_get_k8s_job_status_local)
export(abba_rslauncher_get_job_status_local)
export(abba_rslauncher_get_job_succeeded_local)
export(abba_rslauncher_submit_job_local)
export(abba_rslauncher_submit_logrx_job_local)
export(abba_rslauncher_watch_job_local)
export(abba_slurm_get_job_log)
export(abba_slurm_get_job_status)
export(abba_slurm_get_job_succeeded)
export(abba_slurm_submit_job)
export(abba_slurm_watch_job)
export(abba_submit_and_get_log)
export(abba_submit_batch)
export(abba_submit_batch_and_get_results)
export(abba_submit_job)
export(abba_submit_k8s_job_and_poll_local)
export(abba_submit_k8s_job_local)
export(abba_wait_for_batch_log)
export(abba_wait_for_job_log)
export(abba_watch_k8s_batch_local)
export(abba_watch_k8s_job_local)
export(calculate_run_group)
export(create_batch_api)
export(create_batch_job)
export(submit_k8s_yaml)
importFrom(magrittr,"%>%")

<!-- README.md is generated from README.Rmd. Please edit that file -->

# **abba**

**abba** enables users to submit programs for background execution,
allowing you to continue work as usual and not concern yourself with the
job being terminated when you end your interactive session. Furthermore,
by combining **abba** with services like Posit Connect, you can schedule
jobs to run on a routine basis, such as nightly. While services like
Connect allow you to run your work within a designated R environment,
**abba** expands these capabilities by allowing you to send your work to
a server or cluster environment, such as Kubernetes, in a secure manner.

**abba** has three major components:

- End user job submission functions
- Cluster/server interface functions
- Server API template

The server API template serves as the driver for the actual submission
of jobs where they need to execute. This is intentionally separated to
ensure job submission can be done securely, without giving an end user
too much power or access.

The cluster/server interface functions simplify the management of job
submission into a cluster. Currently only Kubernetes is supported. These
methods manage that process and are utilized on within the API template.

The end user job submission functions are what the typical **abba** user
will utilize. These functions provide a simple interface to target what
the end user must concern themselves with, such as the program being
submitted, the resources necessary, the container environment to run
within, and necessary mounts.

# Run a Program

Using function in **abba**, users can send a program to run and retrieve
the results.

``` r
abba_submit_and_get_log(
  "/home/mike.stackhouse/repos/abba/test_programs/test_program.R"
)
```

# Create a Batch File

**abba** has some convenience functions to simplify the setup of a job
that you intend to schedule.

``` r
create_batch_job()
```

This creates an Rmd file in your local directory following a simple
template.

# Create an **abba** API

**abba** also makes it simple to create the necessary **plumber** API
when setting up within your own environment.

``` r
create_batch_api()
```

## Support

If in need of support, contact <support@atorusresearch.com>

if (!requireNamespace("logrx", quietly = TRUE)) {
  stop("logrx package is not installed. Install logrx package to submit programs via logrx.")
}

# commandArgs picks up the variables you pass from the command line
args <- commandArgs(trailingOnly = TRUE)

prog_path <- args[1]
log_path <- dirname(args[2])
log_name <- basename(args[2])
source_file <- args[3]

# remember current dir and ensure it is restored
cwd <- getwd()
on.exit(if (exists("cwd")) setwd(cwd), add = TRUE)

# source .Rprofile(or any other file) if supplied
if (!is.na(source_file)){
  # source the file from within its parent directory
  setwd(dirname(source_file))
  source(source_file)
}

# main program path should always come as a first argument. It will be executed/sourced
# last. Source the file from within its parent directory
setwd(dirname(prog_path))

logrx::axecute(prog_path, log_name = log_name, log_path = log_path)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(abba)

## ----submit_job, eval=FALSE---------------------------------------------------
# abba_slurm_submit_job(
#   "/home/mike.stackhouse/repos/abba/test_programs/test_program.R",
#   log_path = "/home/mike.stackhouse/test_logs"
#   )
# # [1] "38"

## ----get_status, eval=FALSE---------------------------------------------------
# abba_slurm_get_job_status("57")
# # 57
# # "COMPLETED"

## ----watch_job, eval=FALSE----------------------------------------------------
# abba_slurm_get_job_log("38")
# # [[1]]
# # [1] "[1] 4"                   "Warning message:"        "This is a test warning " "La-di-da"

## ----submit batch, eval=FALSE-------------------------------------------------
# abba_submit_batch(list("/home/mike.stackhouse/repos/abba/test_programs/test_program.R", "/home/mike.stackhouse/repos/abba/test_programs/test_program.R"),
#                   submit_func=abba_slurm_submit_job,
#                   wait_func=abba_slurm_watch_job,
#                   succeed_func=abba_slurm_get_job_succeeded,
#                   log_path="/home/mike.stackhouse/test_logs")
# # Submitting program:
# # 	/home/mike.stackhouse/repos/abba/test_programs/test_program.R
# # Submitting program:
# # 	/home/mike.stackhouse/repos/abba/test_programs/test_program.R
# # /home/mike.stackhouse/repos/abba/test_programs/test_program.R /home/mike.stackhouse/repos/abba/test_programs/test_program.R
#                                                           "48"                                                           "49"

# commandArgs picks up the variables you pass from the command line
args <- commandArgs(trailingOnly = TRUE)

print(args)
prog_path <- args[1]
source_file <- args[2]

# remember current dir
cwd <- getwd()

# source .Rprofile(or any other file) if supplied
if (!is.na(source_file)){
  # source the file from within it`s parent directory
  setwd(dirname(source_file))
  source(source_file)
}

# main program path should always come as a first argument. It will be executed/sourced
# last. Source the file from within it`s parent directory
setwd(dirname(prog_path))
source(prog_path)

# return to original directory
setwd(cwd)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(abba)

## ----submit_batch, eval=FALSE-------------------------------------------------
# abba_submit_batch(
#   list(c("/home/mike.stackhouse/test_progs/r_job_test.R",
#          "/home/mike.stackhouse/test_progs/r_job_test2.R"),
#        "/home/mike.stackhouse/test_progs/r_job_test3.R"),
#   submit_func=abba_rslauncher_submit_job_local,
#   wait_func=abba_rslauncher_watch_job_local,
#   log_path="/home/mike.stackhouse/test_logs",
#   timeout_seconds = 15
#   )
# #  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R
# #  "TG9jYWw6UitNSWpxN2FaQ3JyQXRqbGQ5TzVFUT09"  "TG9jYWw6eGVremgtbGRNMFF0VDFFVG5MTEEzdz09"  "TG9jYWw6eENZb2lBR2x1RVJyZVYyelo0Z040QT09"

## ----create_batch_list, eval=FALSE--------------------------------------------
# ds <- as.data.frame(list(program_name=c("/home/yevhenii.boiko/k8s_test/r_job_test.R", "/home/yevhenii.boiko/k8s_test/r_job_test2.R", "/home/yevhenii.boiko/k8s_test/r_job_test3.R"),
#                          inputs=c("raw.DM", "raw.EX,sdtm.DM", "raw.AE,sdtm.DM"),
#                          outputs=c("sdtm.DM", "sdtm.EX", "sdtm.AE")))
# calculate_run_group(ds, col_name='run_group')
# #                                  program_name         inputs outputs run_group
# # 1  /home/yevhenii.boiko/k8s_test/r_job_test.R         raw.DM sdtm.DM         1
# # 2 /home/yevhenii.boiko/k8s_test/r_job_test2.R raw.EX,sdtm.DM sdtm.EX         2
# # 3 /home/yevhenii.boiko/k8s_test/r_job_test3.R raw.AE,sdtm.DM sdtm.AE         2

## ----example_batch_execution, eval=FALSE--------------------------------------
# abba_submit_batch(calculate_run_group(ds, col_name='run_group'),
#                   submit_func=abba_rslauncher_submit_job_local,
#                   wait_func=abba_rslauncher_watch_job_local,
#                   log_path="/home/mike.stackhouse/test_logs"
#   )
# #  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R
# #  "TG9jYWw6UGVPUVFoeVVMKys2c3M0TnZZcm9ZUT09"  "TG9jYWw6VmgrUnlueDBOLXZuVkVIYWxRa1NEZz09"  "TG9jYWw6bGlWcjdsVTRwQzIwcmlLUThzUUN4UT09"

## ----get_batch_log, eval=FALSE------------------------------------------------
# abba_rslauncher_get_job_log_local(c("TG9jYWw6UGVPUVFoeVVMKys2c3M0TnZZcm9ZUT09", "TG9jYWw6VmgrUnlueDBOLXZuVkVIYWxRa1NEZz09", "TG9jYWw6bGlWcjdsVTRwQzIwcmlLUThzUUN4UT09"))
# # [[1]]
# # [1] "[1] \"Program r_job_test start run timestamp: 2024-03-28 16:48:39\"" "[1] \"Program r_job_test end run timestamp: 2024-03-28 16:48:40\""
# # [3] "Warning message:"                                                    "No useful work done here. "
# #
# # [[2]]
# # [1] "[1] \"Program r_job_test2 start run timestamp: 2024-03-28 16:48:41\"" "[1] \"Program r_job_test2 end run timestamp: 2024-03-28 16:48:42\""
# # [3] "Warning message:"                                                     "No useful work done here. "
# #
# # [[3]]
# # [1] "[1] \"Program r_job_test3 start run timestamp: 2024-03-28 16:48:41\"" "[1] \"Program r_job_test3 end run timestamp: 2024-03-28 16:48:42\""
# # [3] "Warning message:"                                                     "No useful work done here. "

## ----example_batch_execution_sequential, eval=FALSE---------------------------
# abba_submit_batch(calculate_run_group(ds, col_name='run_group'),
#                   submit_func=abba_rslauncher_submit_job_local,
#                   wait_func=abba_rslauncher_watch_job_local,
#                   log_path="/home/mike.stackhouse/test_logs",
#                   sequential=TRUE
#   )
# #  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R
# #  "TG9jYWw6cGppOTZSOXgtR3FDbGdpZGtzaVh1UT09"  "TG9jYWw6VlFhVEs4MXNlWmRmOUh1dG9PeEZFZz09"  "TG9jYWw6Um5YbzY5ZEU5bHN0ZlRTMVBZNm1DZz09"
# 
# abba_rslauncher_get_job_log_local(c("TG9jYWw6cGppOTZSOXgtR3FDbGdpZGtzaVh1UT09","TG9jYWw6VlFhVEs4MXNlWmRmOUh1dG9PeEZFZz09","TG9jYWw6Um5YbzY5ZEU5bHN0ZlRTMVBZNm1DZz09" ))
# # [[1]]
# # [1] "[1] \"Program r_job_test start run timestamp: 2024-03-28 16:52:01\"" "[1] \"Program r_job_test end run timestamp: 2024-03-28 16:52:02\""
# # [3] "Warning message:"                                                    "No useful work done here. "
# #
# # [[2]]
# # [1] "[1] \"Program r_job_test2 start run timestamp: 2024-03-28 16:52:04\"" "[1] \"Program r_job_test2 end run timestamp: 2024-03-28 16:52:05\""
# # [3] "Warning message:"                                                     "No useful work done here. "
# #
# # [[3]]
# # [1] "[1] \"Program r_job_test3 start run timestamp: 2024-03-28 16:52:06\"" "[1] \"Program r_job_test3 end run timestamp: 2024-03-28 16:52:07\""
# # [3] "Warning message:"                                                     "No useful work done here. "

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(abba)

## ----create_api, eval=FALSE---------------------------------------------------
# create_batch_api("path/to/api_directory")
# # Batch API file created at path/to/api_directory/plumber.R

## ----options, eval=FALSE------------------------------------------------------
# options(
#   abba.lower.cpu.limit=0.5,
#   abba.cpu.limit = 2,
#   abba.lower.memory.limit='128M',
#   abba.memory.limit='1G',
#   abba.permitted.containers = c("registry.com/openval_4.2.1_focal:2023.09.0.02",
#                                 "registry.com/openval_4.2.1_focal:latest",
#                                 "registry.com/openval_base_4.3.2_focal:2024.03.01",
#                                 "registry.com/openval_base_4.3.2_focal:latest",
#                                 "registry.com/openval-dev-focal:latest"),
#   abba.default.container = "registry.com/openval_4.2.1_focal:2023.09.0.02"
#   )

library(plumber)
library(abba)

options(
  abba.lower.cpu.limit=0.5,
  abba.cpu.limit = 2,
  abba.lower.memory.limit='128M',
  abba.memory.limit='1G',
  abba.permitted.containers = c("registry.io/default_image:version",
                                "registry.io/image1:version",
                                "registry.io/image1:version"),
  abba.default.container = "registry.io/default_image:version",
  abba.k8s.namespace='rstudio',
  abba.home.nfs.ip.address='0.0.0.0'
)

# Returns a list containing "user" and "groups" information
# populated by incoming request data.
getUserMetadata <- function(req) {
  rawUserData <- req[["HTTP_RSTUDIO_CONNECT_CREDENTIALS"]]
  if (!is.null(rawUserData)) {
    jsonlite::fromJSON(rawUserData)
  } else {
    list()
  }
}

#* @apiTitle {abba} API
#* @apiDescription Backend submission API for {abba}

#* Submit a job on Kubernetes
#' @param file_path Full path to R file
#' @param batch_group_id Group ID for batch processing
#' @param user_tag Optional; a string that describes what kind of job will be scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container A string containing a permitted container name. Default is specified by API administrator.
#' @param mounts Specifically formatted list with information bout volumes that container would have access to during the run
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_address IP address for mounting service user home directory
#' @param namespace Kubernetes namespace to put the job in
#* @post /submit-job
function(file_path,
         batch_group_id='',
         user_tag='',
         cpu_limit="1",
         memory_limit="512M",
         container=getOption('abba.default.container'),
         mounts='',
         auto_mount_home=FALSE,
         home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
         namespace=getOption('abba.k8s.namespace'),
         req,
         res) {

  user <- getUserMetadata(req)
  username <- user[["user"]]

  # Request may come in as empty string
  if (container == "") {
    container <- getOption('abba.default.container')
  }

  # If specified, is the container name an allowable choice?
  permitted_containers <- getOption('abba.permitted.containers')
  if (!is.null(permitted_containers) && !(container %in% permitted_containers)) {
    err_msg <-sprintf(
      "The container %s is not an permitted image. Permitted images are:\n\t- %s",
      container,
      paste0(permitted_containers, collapse = "\n\t- ")
    )

    res$status <- 400
    res$body$message = jsonlite::unbox(err_msg)
    return(list(
      message=jsonlite::unbox(err_msg),
      error=jsonlite::unbox(err_msg)
    ))
  }

  result <- abba_submit_k8s_job_local(
    file_path,
    batch_group_id=batch_group_id,
    user_tag=user_tag,
    cpu_limit=cpu_limit,
    memory_limit=memory_limit,
    container=container,
    mounts=mounts,
    auto_mount_home=auto_mount_home,
    home_nfs_ip_address=home_nfs_ip_address,
    namespace=namespace,
    username=username
  )

  return(result)
}

#* Get Job logs
#' @param job_ids list of job IDs
#' @param namespace Kubernetes namespace to put the job in
#* @get /job-log
function(job_ids,
         namespace=getOption('abba.k8s.namespace')) {

  result <- abba_get_k8s_job_log_local(job_ids,
                                       namespace=namespace)

  return(result)
}

#* Get Job logs of a batch
#' @param batch_id unique identifier for a batch
#' @param namespace Kubernetes namespace to put the job in
#* @get /batch-log
function(batch_id,
         namespace=getOption('abba.k8s.namespace')) {

  result <- abba_get_k8s_batch_log_local(batch_id,
                                         namespace=namespace)

  return(result)
}

#* Get status of every job in a batch
#* @param batch_id unique identifier for a batch
#' @param namespace Kubernetes namespace to put the job in
#* @get /batch-status
function(batch_id='',
         namespace=getOption('abba.k8s.namespace')) {

  result <- abba_get_k8s_batch_status_local(batch_id,
                                            namespace=namespace)
  return(result)
}

#* Get status of a job as a collection of statuses of its pods
#' @param job_id job id to get status for
#' @param namespace Kubernetes namespace to put the job in
#* @get /job-status
function(job_id='',
         namespace=getOption('abba.k8s.namespace')) {
  result <- abba_get_k8s_job_status_local(job_id,
                                          namespace=namespace)
  return(result)
}

#* Get the list of permitted containers available in the API
#* @get /permitted-containers
function() {
  return(sort(getOption('abba.permitted.containers')))
}

#' Checks that mount information is supplied in expected format
#'
#' @param mounts a named list representing k8s mount information
#'
#' @return TRUE if mounts structure is as expected
#' @noRd
#'
validate_k8s_mount <- function(mounts){
  # null is a default value for mount argument in configure_yaml function.
  # by default, we should not add new mounts to config
  if (is.null(mounts) || (is.character(mounts) && mounts == '')){return(FALSE)}

  if (!is.list(mounts)){
    stop(sprintf('Mounts should be a list, not %s',
                 typeof(mounts)))}
  if(!all(c('volumes', 'volumeMounts') %in% names(mounts))){
    stop(sprintf('Mounts should have 2 attributes: volumes and volumeMounts, but it instead has %s',
                 names(mounts)))}
  if(length(mounts$volumes) != length(mounts$volumeMounts)){
    stop(sprintf('volumes and volumeMounts attributes should be equal in length. Length(volumes)=%d, Length(volumeMounts)=%d',
                 length(mounts$volumes), length(mounts$volumeMounts)))
  }

  volumes_names <- sapply(mounts$volumes, function(x) x$name)
  volumeMounts_names <- sapply(mounts$volumeMounts, function(x) x$name)

  if(!identical(sort(volumes_names), sort(volumeMounts_names))){
    stop(sprintf('Mounts should be equally named in volumes and volumeMounts. Volume names: %s, volumeMounts names: %s',
                 volumes_names, volumeMounts_names))
  }

  return(TRUE)
}

#' Checks that container information supplied is valid
#'
#' @param container A character string with the path to a container image
#'
#' @return TRUE if container passes checks
#' @noRd
#'
validate_k8s_container <- function(container){
  # If specified, is the container name an allowable choice?
  permitted_containers <- getOption('abba.permitted.containers')
  if (!is.null(permitted_containers) && !(container %in% permitted_containers)) {
    err_msg <-sprintf(
      "The container %s is not an permitted image. Permitted images are:\n\t- %s",
      container,
      paste0(permitted_containers, collapse = "\n\t- ")
    )
    stop(err_msg)
  }
}

#' Check that cpu_limit is between lower and upper limits specified by abba package options
#'
#' @param cpu_limit amount of cores a job cannot exceed when assigning resources
#'
#' @return TRUE if cpu_limit is between abba.lower.cpu.limit and abba.cpu.limit, error otherwise
#' @noRd
#'
validate_cpu_limit <- function(cpu_limit){
  if (mcpu_to_cpu(cpu_limit) < mcpu_to_cpu(getOption('abba.lower.cpu.limit'))){
    stop(sprintf('CPU limit(%s) is below the lower limit set by abba.lower.cpu.limit(%s).',
                 cpu_limit, getOption('abba.lower.cpu.limit')))
  }
  else if (mcpu_to_cpu(cpu_limit) > mcpu_to_cpu(getOption('abba.cpu.limit'))){
    stop(sprintf('CPU limit(%s) exceed the upper limit set by abba.cpu.limit(%s).',
                 cpu_limit, getOption('abba.cpu.limit')))
  }
  # in case lower/upper limits were not triggered, return originally supplied cpu_limit value
  return(cpu_limit)
}

#' Make sure memory limit is between lower and upper limit set by package options
#'
#' @param memory_limit An integer or character value specifying amount of memory
#'
#' @return TRUE if memory_limit is between lower and upper limit set by abba package options
#' @noRd
validate_memory_limit <- function(memory_limit){
  if (memory_to_bytes(memory_limit) < memory_to_bytes(getOption('abba.lower.memory.limit'))){
    stop(sprintf('Requested memory limit(%s) is lower than the minimum set by abba.lower.memory.limit(%s).',
                 memory_limit, getOption('abba.lower.memory.limit')))
  }

  else if (memory_to_bytes(memory_limit) > memory_to_bytes(getOption('abba.memory.limit'))){
    stop(sprintf('Requested memory limit(%s) exceed the limit set by abba.memory.limit(%s).',
                 memory_limit, getOption('abba.memory.limit')))
    }

  # in case lower/upper limits were not triggered, return originally supplied memory_limit value
  return(memory_limit)

}

#' Evaluate batch_id passed by user
#'
#' @param batch_group_id a string to uniquely identify a group of programs
#'
#' @return TRUE if batch_group_id passed all checks, otherwise function errors out
#' @noRd
validate_batch_id <- function(batch_group_id){
  if (!is.character(batch_group_id)) {
    stop(sprintf("batch_group_id must be a string, not %s", typeof(batch_group_id)))
  }
  if (length(batch_group_id) > 1) {
    stop("batch_group_id must be a single string value")
  }
  return(TRUE)
}

#' Evaluate unit_type
#'
#' @param unit_type a string to specify unit - either a 'job' or 'batch'
#'
#' @return TRUE if unit_type passed all checks, otherwise function errors out
#' @noRd
validate_unit_type <- function(unit_type){
  if (!is.character(unit_type)) {
    stop(sprintf("unit_type must be a string, not %s", typeof(unit_type)))
  }
  if (!(unit_type %in% c('job', 'batch'))) {
    stop(sprintf("unit_type must be one of 'job', 'batch', not %s", unit_type))
  }
  return(TRUE)
}

test_that("Checks pass if no option is set", {
  op <- options(abba.permitted.containers = NULL)

  expect_silent(validate_k8s_container("bad_name"))

  options(op)
})

test_that("Error generates properly when container name is invalid", {
  permitted_containers <- c("path.io/container-1.0.0", "path.io/container2-1.0.0", "path.io/container3-1.0.0")

  op <- options(abba.permitted.containers = permitted_containers)

  expect_snapshot_error(validate_k8s_container("bad_name"))

  options(op)
})

test_that("Check passes properly when container name is valid", {
  permitted_containers <- c("path.io/container-1.0.0", "path.io/container2-1.0.0", "path.io/container3-1.0.0")

  op <- options(abba.permitted.containers = permitted_containers)

  expect_silent(validate_k8s_container("path.io/container-1.0.0"))

  options(op)
})

test_that("function errors when input is not a string", {
  expect_error(validate_batch_id(13435))
})

test_that("function errors when input length is bigger than 1", {
  expect_error(validate_batch_id(c('id1', 'id2')))
})

test_that("function returns TRUE when input is a string character of length 1", {
  expect_equal(validate_batch_id('id1'), TRUE)
})

test_that("function errors when memory limit is lower than abba.lower.memory.limit", {
  low_value <- memory_to_bytes(getOption("abba.lower.memory.limit")) * 0.5
  expect_error(validate_memory_limit(low_value))
})

test_that("function errors when memory limit is higher than abba.memory.limit", {
  high_value <- memory_to_bytes(getOption("abba.memory.limit")) * 2
  expect_error(validate_memory_limit(high_value))
})

test_that("function returns input value when the input argument is within limits", {
  proper_value <- mean(c(memory_to_bytes(getOption("abba.lower.memory.limit")),
                         memory_to_bytes(getOption("abba.memory.limit"))))
  expect_equal(validate_memory_limit(proper_value), proper_value)
})

test_that("function errors when mount information is not a list type", {
  expect_error(validate_k8s_mount(1))
})

test_that("function returns FALSE when the input argument is NULL", {
  expect_equal(validate_k8s_mount(NULL), FALSE)
})

test_that("function returns FALSE when the input argument is an empty string", {
  expect_equal(validate_k8s_mount(''), FALSE)
})

test_that("function errors when the input list does not have required named values", {
  expect_error(validate_k8s_mount(list(volumes=list(list(name='mount0')))))
})

test_that("function errors when mount names in volumes and volumeMounts differ", {
  mount_info <- list(volumes=list(list(name='mount0')),
                     volumeMounts=list(list(name='mount1')))
  expect_error(validate_k8s_mount(mount_info))
})

test_that("Mount names in volumes and volumeMounts should be the same", {
  mount_info <- list(volumes=list(list(name='mount0')),
                     volumeMounts=list(list(name='mount0')))
  expect_equal(validate_k8s_mount(mount_info), TRUE)
})

test_that("function errors when length of volumes and volumeMounts differ", {
  mount_info <- list(volumes=list(list(name='mount0'), list(name='mount1')),
                     volumeMounts=list(list(name='mount0')))
  expect_error(validate_k8s_mount(mount_info))
})

test_that("function returns TRUE when unit_type equals to 'job' or 'batch'", {
  expect_equal(TRUE, validate_unit_type('job'))
  expect_equal(TRUE, validate_unit_type('batch'))
})

test_that("function errors out when unit_type is not a character", {
  expect_error(validate_unit_type(1))
})

test_that("function errors out when unit_type is not one of 'job', 'batch'", {
  expect_error(validate_unit_type('not a job'))
})

read_r_versions <- function(){

  r_versions_path <- getOption('abba.r.versions.path', '/etc/rstudio/r-versions')
  if (!file.exists(r_versions_path)) {
    stop("Cannot read R versions: ", r_versions_path, " not found. ",
         "This function requires Posit Workbench. ",
         "Set options(abba.r.versions.path = ...) to configure the path.")
  }

  # read data from r-versions json
  prep <- utils::read.delim(r_versions_path, sep=":", header=FALSE,
                            comment.char="#", strip.white=TRUE)
  # calculate values for ID column(need for transforming data)
  id_lines <- Filter(function(x) !startsWith(x, '#'), readLines(r_versions_path))
  ids <- cumsum(id_lines == "")[id_lines!=""]

  # add ids to differentiate between information blobs belonging to different r installations
  prep$id <- ids
  # add column names
  colnames(prep) <- c("Property", "Value", "ID")

  # R installs can be specified as a list of folders without any labels
  # TODO: handle r installs that are specified as a single folder path

  # reshape data frame
  result <- as.data.frame(tidyr::pivot_wider(prep,
                                             names_from = "Property",
                                             values_from = "Value"))
  # add 'bin/R' to path
  result$Path <- sapply(result$Path, function(x) file.path(x, 'bin', 'R'))
  result$RScriptPath <- sapply(result$Path, function(x) file.path(x, 'bin', 'Rscript'))
  # remove any entries which don't actually exist
  result <- result[file.exists(result$Path),]

  return(result)
}

select_r_version <- function(r_version){
  # default to standard workbench R version if none specified by user
  if (is.null(r_version) || is.na(r_version) || r_version == ""){
    return(file.path(R.home("bin"), "R"))
  }

  # read list of r versions available for workbench
  av <- read_r_versions()
  # if label is supplied, return R executable path
  if (r_version %in% av$Label){
    return(av[av$Label == r_version,]$Path)
  }
  # if exe path is supplied, return it back if it is listed in workbench r version
  else if (r_version %in% av$Path){
    return(r_version)
  }
  # it is possible that an R version is not registered in Workbench but can still
  # be used to submit R programs
  else if (file.exists(r_version)){
    return(r_version)
  }
  else{stop(paste(
    sprintf("R version '%s' not found. Cannot submit job.", r_version),
    "List of available R versions:\n",
    paste(utils::capture.output(print(av, row.names = FALSE)), collapse="\n")))}

}

select_rscript_version <- function(r_version){
  # default to standard workbench R version if none specified by user
  if (is.null(r_version) || is.na(r_version) || r_version == ""){
    return(file.path(R.home("bin"), "Rscript"))
  }

  # read list of r versions available for workbench
  av <- read_r_versions()
  # if label is supplied, return R executable path
  if (r_version %in% av$Label){
    return(av[av$Label == r_version,]$RScriptPath)
  }
  # if exe path is supplied, return it back if it is listed in workbench r version
  else if (r_version %in% av$RScriptPath){
    return(r_version)
  }
  # it is possible that an R version is not registered in Workbench but can still
  # be used to submit R programs
  else if (file.exists(r_version)){
    return(r_version)
  }
  else{stop(paste(
    sprintf("R version '%s' not found. Cannot submit job.", r_version),
    "List of available R versions:\n",
    paste(utils::capture.output(print(av, row.names = FALSE)), collapse="\n")))}

}

#' Submit programs for execution in parallel. Each program in a list will be submitted
#' at the same time.
#'
#' @param prog_list A list of R program paths to execute
#' @param submit_func function that will be used to submit jobs
#' @param wait_func function that checks job status and returns when job finishes executing
#' @param ... arguments that will be passed to submit_func and wait_func functions
#'
#' @return list job IDs associated with executed programs
#' @noRd
#'
#' @examples \dontrun{
#' job_ids <- batch_submit_parallel(
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_dm.sas",
#'     "/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_ae.sas"))
#'  }
batch_submit_parallel <- function(prog_list,
                                  submit_func=abba_rslauncher_submit_job_local,
                                  wait_func=abba_rslauncher_watch_job_local,
                                  ...) {
  # Submit jobs for the programs and wait until everything has been executed
  job_ids <- sapply(prog_list, function(x) submit_func(x, ...))

  # wait until all programs finish executing
  results <- wait_func(job_ids, ...)

  # return job ids generated by submitting programs for execution in workbench
  return(job_ids)
}

#' Submit programs for execution in order defined by structure of input list.
#' Programs inside sublists will be executed in parallel, and sublists themselves
#' would be submitted sequentially.
#'
#' @param prog_list A list of R program paths to execute
#' @param sequential when sequential=TRUE, prog_list is flattened and everything is executed sequentially.
#' @param submit_func function that will be used to submit jobs
#' @param wait_func function that checks job status and returns when job finishes executing
#' @param succeed_func function that returns TRUE if job finished running without errors and FALSE otherwise
#' @param col_name Name of the column that contains run group numbers when prog_list is a data frame
#' @param halt_on_error If TRUE: if prog_list contains program inputs/outputs - programs that depend on failed
#' program will not be executed; if prog_list contains only program paths - when program fails,
#' entire batch will stop executing. TRUE by default
#' @param rerun_unchanged_programs If FALSE: will not re-run programs whose code
#' and inputs have not been modified since last batch run with update_cache=TRUE.
#' @param update_cache if TRUE, file hash for programs and their inputs will be
#' calculated and written to `cache_folder` after the batch run. If FALSE, the
#' cache will not be created or updated. FALSE by default.
#' @param cache_folder Path to the folder where hash-sums of programs and
#' their inputs will be stored. Required when `update_cache=TRUE` or
#' `rerun_unchanged_programs=FALSE`. The default value can be set via the
#' `abba.default_cache_folder` option; abba never falls back to writing caches
#' under the user's filespace.
#' @param ... arguments that will be passed to submit_func and wait_func functions
#'
#' @return list job IDs associated with executed programs
#' @export
#'
#' @examples \dontrun{
#' job_ids <- abba_submit_batch(list(
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/dm.sas"),
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/ae.sas"),
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_dm.sas",
#'     "/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_ae.sas")))
#'  }
abba_submit_batch <- function(prog_list,
                              sequential=FALSE,
                              submit_func=abba_rslauncher_submit_job_local,
                              wait_func=abba_rslauncher_watch_job_local,
                              succeed_func=abba_rslauncher_get_job_succeeded_local,
                              col_name='run_group',
                              halt_on_error=TRUE,
                              rerun_unchanged_programs=TRUE,
                              update_cache=FALSE,
                              cache_folder=getOption("abba.default_cache_folder"),
                              ...) {

  if ((update_cache || !rerun_unchanged_programs) && is.null(cache_folder)) {
    stop("cache_folder must be supplied when update_cache=TRUE or rerun_unchanged_programs=FALSE. Set it via the cache_folder argument or the abba.default_cache_folder option.")
  }

  # if sequential=TRUE is specified - flatten the list and this will execute everything sequentially
  # REGARDLESS of whether prog_list is a list, a character vector or a data frame
  if (sequential==TRUE){
    prog_list_converted <- unlist(dataframe_to_batch_list(prog_list))
  }
  else {prog_list_converted <- prog_list}

  # execute each item in the prog_list_converted sequentially. Programs inside
  # each list element, which can also be a vector of program paths,
  # will be executed in parallel, unless 'sequential=TRUE' is specified
  job_ids <- list()
  previous_run_ok <- TRUE
  previous_run_programs <- c()

  # check if run groups are present for data frame
  unique_run_groups <- get_run_groups(prog_list_converted)

  for (rg in unique_run_groups){

    # check if programs/their inputs had any changes and remove programs whose code
    # and inputs(if specified) have not changed since last hash collection(typicaly from previous batch run)
    if (!rerun_unchanged_programs){
      prog_list_converted <- remove_unchanged_programs(prog_list_converted,
                                                       run_group=rg,
                                                       col_name=col_name,
                                                       cache_folder=cache_folder)
    }
    # check if programs from previous run group were executed completely
    batch_check_results <-
      batch_run_control(prog_list_converted,
                        current_group=rg,
                        previous_run_programs=previous_run_programs,
                        previous_run_ok=previous_run_ok,
                        col_name='run_group',
                        halt_on_error=halt_on_error,
                        ...)

    prog_list_converted <- batch_check_results$prog_list

    # if batch input is a list, do not run the next group
    if (batch_check_results$stop == TRUE && halt_on_error){break}

    # select programs for running in parallel
    parallel_run <- select_parallel_run(prog_list_converted, rg, col_name=col_name)
    # message that would show batch progress. Only show it if there is at least one program to run
    if (length(parallel_run) > 1){
      message(sprintf("Submitting programs for parallel run:\n\t%s", paste(parallel_run, collapse='\n\t')))
    }
    else if (length(parallel_run) == 1){
      message(sprintf("Submitting program:\n\t%s", paste(parallel_run, collapse='\n\t')))
    }
    # submit the run
    new_run <- batch_submit_parallel(parallel_run,
                                     submit_func=submit_func,
                                     wait_func=wait_func,
                                     ...)
    # add executed jobs to the list
    job_ids <- c(job_ids, new_run)
    # collect information about this run group exit codes to decide if we should
    # continue with the batch run on the next loop
    previous_run_ok <- succeed_func(new_run)
    previous_run_programs <- parallel_run
  }
  # update program and inputs hash after batch has been run
  # only update hashes if re-run check has been explicitly enabled
  if (update_cache){
    update_program_hashes(prog_list, cache_folder=cache_folder)
    }

  return(unlist(job_ids))
}


#' Submit programs for execution in order defined by structure of input list.
#'
#' @param prog_list A list of R program paths to execute
#' @param sequential when sequential=TRUE, prog_list is flattened and everything is executed sequentially.
#' @param submit_func function that will be used to submit jobs
#' @param wait_func function that checks job status and returns when job finishes executing
#' @param succeed_func function that returns TRUE if job finished running without errors and FALSE otherwise
#' @param status_func function that returns descriptive job statuses
#' @param col_name Name of the column that contains run group numbers when prog_list is a data frame
#' @param halt_on_error If TRUE: if prog_list contains program inputs/outputs - programs that depend on failed
#' program will not be executed; if prog_list contains only program paths - when program fails,
#' entire batch will stop executing. TRUE by default
#' @param rerun_unchanged_programs If FALSE: will not re-run programs whose code
#' and inputs have not been modified since last batch run with update_cache=TRUE.
#' @param update_cache if TRUE, file hash for programs and their inputs will be
#' calculated and written to `cache_folder` after the batch run. If FALSE, the
#' cache will not be created or updated. FALSE by default.
#' @param cache_folder Path to the folder where hash-sums of programs and
#' their inputs will be stored. Required when `update_cache=TRUE` or
#' `rerun_unchanged_programs=FALSE`. The default value can be set via the
#' `abba.default_cache_folder` option; abba never falls back to writing caches
#' under the user's filespace.
#' @param ... arguments that will be passed to submit_func and wait_func functions
#'
#' @return Data frame containing program names, job ids, execution statuses
#' @export
#'
#' @examples \dontrun{
#' job_ids <- abba_submit_batch_and_get_results(list(
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/dm.sas"),
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/ae.sas"),
#'   c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_dm.sas",
#'     "/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_ae.sas")),
#'   sequential=TRUE)
#'  }
abba_submit_batch_and_get_results <- function(prog_list,
                                              sequential=FALSE,
                                              submit_func=abba_rslauncher_submit_job_local,
                                              wait_func=abba_rslauncher_watch_job_local,
                                              succeed_func=abba_rslauncher_get_job_succeeded_local,
                                              status_func=rslauncher_get_job_display_status,
                                              col_name='run_group',
                                              halt_on_error=TRUE,
                                              rerun_unchanged_programs=TRUE,
                                              update_cache=FALSE,
                                              cache_folder=getOption("abba.default_cache_folder"),
                                              ...) {
  # use universal batch runner to submit programs and wait for completion
  job_ids <- abba_submit_batch(prog_list,
                               sequential=sequential,
                               submit_func=submit_func,
                               wait_func=wait_func,
                               succeed_func=succeed_func,
                               col_name=col_name,
                               halt_on_error=halt_on_error,
                               rerun_unchanged_programs=rerun_unchanged_programs,
                               update_cache=update_cache,
                               cache_folder=cache_folder,
                               ...)
  # collect batch results in a data frame
  results <- compose_batch_results(
    job_ids=job_ids,
    prog_names=unlist(dataframe_to_batch_list(prog_list)),
    status_func=status_func)

  return(results)
}

library(mockery)

test_that("batch_submit_parallel functions as expected", {
  mock_submit <- mock('program1.R', 'program2.R')
  mock_wait <- mock('result3', 'result4')

  result_actual <- batch_submit_parallel(c('program1.R', 'program2.R'),
                                         submit_func=mock_submit,
                                         wait_func=mock_wait,
                                         arg1='arg1')
  expect_called(mock_submit, 2)
  expect_called(mock_wait, 1)

  expect_args(mock_submit, 1, 'program1.R', arg1='arg1')
  expect_args(mock_submit, 2, 'program2.R', arg1='arg1')

  expected_mock_args <- c('program1.R', 'program2.R')
  names(expected_mock_args) <- c('program1.R', 'program2.R')
  expect_args(mock_wait, 1, expected_mock_args, arg1='arg1')

  expect_equal(result_actual, expected_mock_args)
})


test_that("abba_submit_batch flattens the list if sequential=TRUE", {

  # define mock functions
  mock_dataframe_to_batch_list <- mock(list(c('prog1.R'), ('prog2.R')))
  mock_get_run_groups <- mock(c(1,2))
  mock_batch_run_control <- mock(list(prog_list=c('prog1.R'), stop=FALSE),
                                 list(prog_list=c('prog2.R'), stop=FALSE),
                                 cycle=TRUE)
  mock_select_parallel_run <- mock('prog1.R', 'prog2.R')
  mock_batch_submit_parallel <- mock('prog-id1', 'prog-id2')
  mock_submit <- mock()
  mock_wait <- mock()
  mock_succeed <- mock(TRUE, cycle=TRUE)
  mock_abba_rslauncher_get_job_succeeded_local <- mock(TRUE, cycle=TRUE)
  mock_update_hashes <- mock()

  # stub functions in target function
  stub(abba_submit_batch, 'dataframe_to_batch_list', mock_dataframe_to_batch_list)
  stub(abba_submit_batch, 'get_run_groups', mock_get_run_groups)
  stub(abba_submit_batch, 'batch_run_control', mock_batch_run_control)
  stub(abba_submit_batch, 'select_parallel_run', mock_select_parallel_run)
  stub(abba_submit_batch, 'batch_submit_parallel', mock_batch_submit_parallel)
  stub(abba_submit_batch, 'update_program_hashes', mock_update_hashes)

  # run the function
  result_actual <- abba_submit_batch(c('prog1.R', 'prog2.R'),
                                     sequential=TRUE,
                                     submit_func=mock_submit,
                                     wait_func=mock_wait,
                                     succeed_func=mock_succeed,
                                     arg1='arg1')

  # execute checks
  expect_called(mock_dataframe_to_batch_list, 1)
  expect_called(mock_batch_submit_parallel, 2)

  expect_args(mock_batch_submit_parallel, 1,
              'prog1.R',
              submit_func=mock_submit,
              wait_func=mock_wait,
              arg1='arg1')
  expect_args(mock_batch_submit_parallel, 2,
              'prog2.R',
              submit_func=mock_submit,
              wait_func=mock_wait,
              arg1='arg1')

  expected_result <- c('prog-id1', 'prog-id2')

  expect_equal(result_actual, expected_result)
})


test_that("abba_submit_batch breaks execution cycle if programs have errors in sequential mode", {

  # define mock functions
  mock_dataframe_to_batch_list <- mock(c('prog1.R', 'prog2.R'))
  mock_get_run_groups <- mock(c(1,2))
  mock_batch_run_control <- mock(list(prog_list=c('prog1.R'), stop=FALSE),
                                 list(prog_list=c('prog2.R'), stop=TRUE),
                                 cycle=TRUE)
  mock_select_parallel_run <- mock('prog1.R', 'prog2.R')
  mock_batch_submit_parallel <- mock('prog-id1', 'prog-id2')
  mock_submit <- mock()
  mock_wait <- mock()
  mock_succeed <- mock(TRUE, cycle=TRUE)
  mock_abba_rslauncher_get_job_succeeded_local <- mock(FALSE, cycle=TRUE)
  mock_update_hashes <- mock()

  # stub functions in target function
  stub(abba_submit_batch, 'dataframe_to_batch_list', mock_dataframe_to_batch_list)
  stub(abba_submit_batch, 'get_run_groups', mock_get_run_groups)
  stub(abba_submit_batch, 'batch_run_control', mock_batch_run_control)
  stub(abba_submit_batch, 'select_parallel_run', mock_select_parallel_run)
  stub(abba_submit_batch, 'batch_submit_parallel', mock_batch_submit_parallel)
  stub(abba_submit_batch, 'update_program_hashes', mock_update_hashes)

  # run the function
  result_actual <- abba_submit_batch(c('prog1.R', 'prog2.R'),
                                     sequential=TRUE,
                                     halt_on_error = TRUE,
                                     submit_func=mock_submit,
                                     wait_func=mock_wait,
                                     succeed_func=mock_succeed,
                                     arg1='arg1')

  # execute checks
  expect_called(mock_dataframe_to_batch_list, 1)
  expect_called(mock_batch_submit_parallel, 1)

  expect_args(mock_batch_submit_parallel, 1,
              'prog1.R',
              submit_func=mock_submit,
              wait_func=mock_wait,
              arg1='arg1')

  expected_result <- c('prog-id1')

  expect_equal(result_actual, expected_result)
})


test_that("abba_submit_batch works as expected in standard mode", {

  # define mock functions
  mock_dataframe_to_batch_list <- mock()
  mock_get_run_groups <- mock(c(1,2))
  mock_batch_run_control <- mock(list(prog_list=c('prog1.R', 'prog2.R'), stop=FALSE),
                                 list(prog_list=c('prog3.R'), stop=FALSE),
                                 cycle=TRUE)
  mock_select_parallel_run <- mock(c('prog1.R', 'prog2.R'), 'prog3.R')
  mock_batch_submit_parallel <- mock(c('prog-id1', 'prog-id2'), 'prog-id3')
  mock_submit <- mock()
  mock_wait <- mock()
  mock_succeed <- mock(TRUE, cycle=TRUE)
  mock_update_hashes <- mock()

  # stub functions in target function
  stub(abba_submit_batch, 'dataframe_to_batch_list', mock_dataframe_to_batch_list)
  stub(abba_submit_batch, 'get_run_groups', mock_get_run_groups)
  stub(abba_submit_batch, 'batch_run_control', mock_batch_run_control)
  stub(abba_submit_batch, 'select_parallel_run', mock_select_parallel_run)
  stub(abba_submit_batch, 'batch_submit_parallel', mock_batch_submit_parallel)
  stub(abba_submit_batch, 'update_program_hashes', mock_update_hashes)

  # run the function
  result_actual <- abba_submit_batch(list(c('prog1.R', 'prog2.R'), 'prog3.R'),
                                     submit_func=mock_submit,
                                     wait_func=mock_wait,
                                     succeed_func=mock_succeed,
                                     arg1='arg1')

  # execute checks
  expect_called(mock_dataframe_to_batch_list, 0)
  expect_called(mock_batch_submit_parallel, 2)

  expect_args(mock_batch_submit_parallel, 1,
              c('prog1.R','prog2.R'),
              submit_func=mock_submit,
              wait_func=mock_wait,
              arg1='arg1')
  expect_args(mock_batch_submit_parallel, 2,
              c('prog3.R'),
              submit_func=mock_submit,
              wait_func=mock_wait,
              arg1='arg1')

  expected_result <- c('prog-id1', 'prog-id2', 'prog-id3')

  expect_equal(result_actual, expected_result)
})


test_that("abba_submit_batch_and_get_results works as expected", {

  # define mock functions
  mock_abba_submit_batch <- mock('job_ids')
  mock_compose_batch_results <- mock('batch-results')
  mock_status <- mock()
  mock_submit <- mock()
  mock_wait <- mock()
  mock_succeed <- mock(TRUE, cycle=TRUE)

  # stub functions in target function
  stub(abba_submit_batch_and_get_results, 'abba_submit_batch', mock_abba_submit_batch)
  stub(abba_submit_batch_and_get_results, 'compose_batch_results', mock_compose_batch_results)

  # run the function
  result_actual <- abba_submit_batch_and_get_results(
    list(c('prog1.R')),
    submit_func=mock_submit,
    wait_func=mock_wait,
    succeed_func=mock_succeed,
    status_func=mock_status,
    arg1='arg1')

  # execute checks
  expect_called(mock_abba_submit_batch, 1)
  expect_called(mock_compose_batch_results, 1)

  expect_args(mock_abba_submit_batch, 1,
              list(c('prog1.R')),
              sequential=FALSE,
              submit_func=mock_submit,
              wait_func=mock_wait,
              succeed_func=mock_succeed,
              col_name='run_group',
              halt_on_error=TRUE,
              rerun_unchanged_programs=TRUE,
              update_cache=FALSE,
              cache_folder=NULL,
              arg1='arg1')
  expect_args(mock_compose_batch_results, 1,
              job_ids=c('job_ids'),
              prog_names=c('prog1.R'),
              status_func=mock_status)

  expected_result <- c('batch-results')

  expect_equal(result_actual, expected_result)
})

.onLoad <- function(libname, pkgname) {
  op <- options()
  op.abba <- list(
    abba.lower.cpu.limit = 0.001,
    abba.cpu.limit = 2,
    abba.lower.memory.limit = '128M',
    abba.memory.limit = '1G',
    abba.permitted.containers = NULL,
    abba.default.container = NULL,
    abba.k8s.namespace = 'rstudio',
    abba.home.nfs.ip.address = NULL,
    abba.r.versions.path = '/etc/rstudio/r-versions',
    abba.default_cache_folder = NULL,
    abba.slurm.cpu.cores = 1L,
    abba.slurm.memory = 1024L
  )
  toset <- !(names(op.abba) %in% names(op))
  if (any(toset)) options(op.abba[toset])
  invisible()
}

mcpu_to_cpu <- function(cpu){
  # don't decipher anything if cpu limit was supplied as numeric
  if (is.numeric(cpu)){
    return(round(cpu, digits=3))}

  # define conversion factor
  suffixes <- list(m=1/1000)
  # extract numerical part and suffix
  numeric_part <- gsub('[^0-9.]', '', cpu)
  suffix <- gsub('[^a-zA-Z]', '', cpu)

  # just return numeric part if no unit specified
  if (suffix==''){return(round(as.numeric(numeric_part), digits=3))}
  # return just numeric part without conversion if suffix is unrecognized
  else if (suffix != 'm'){
    stop('Unrecognized cpu unit. Allowed unis are: m')
  }
  # return converted value in CPU units
  rounded_converted <- round(as.numeric(numeric_part) * suffixes[[suffix]],
                             digits=3)
  return(rounded_converted)

}

memory_to_bytes <- function(memory){
  # don't decipher anything if memory limit was supplied as numeric
  if (! is.character(memory)){
    return(ceiling(memory))}

  # define values of conversion factors. Not going above the terabyte because
  # it is ridiculous
  suffixes <- list(k=1e3, M=1e6, G=1e9, T=1e12,
                   ki=1024, Mi=1024**2, Gi=1024**3, Ti=1024**4)
  # extract numerical part and suffix
  numeric_part <- gsub('[^0-9.]', '', memory)
  suffix <- gsub('[^a-zA-Z]', '', memory)

  # just return numeric part if no unit specified
  if (suffix==''){return(ceiling(as.numeric(numeric_part)))}
  # return just numeric part without conversion if suffix is unrecognized
  else if (!(suffix %in% names(suffixes))){
    stop('Unrecognized memory unit. Allowed units are: k, M, G, T, ki, Mi, Gi, Ti')
  }
  # return converted value in bytes
  return(ceiling(as.numeric(numeric_part) * suffixes[[suffix]]))
}

test_that("YAML fields are properly updated by configure_yaml", {

  mock_system2 <- mock('uid=123(test_username) gid=456(domain users)', cycle=TRUE)

  stub(configure_k8s_yaml, "system2", mock_system2, depth=2)

  yaml_file_configured <- configure_k8s_yaml(file_path="/tst/path/test.R",
                                             batch_group_id="group_A",
                                             user_tag='user_tag',
                                             cpu_limit= 2L,
                                             memory_limit='512M',
                                             username='test_username',
                                             namespace='test_namespace',
                                             auto_mount_home=TRUE,
                                             home_nfs_ip_address='0.1.2.3'
                                             )
  actual <- c(yaml_file_configured$metadata$name,
              yaml_file_configured$metadata$namespace,
              yaml_file_configured$metadata$labels$`batch-group`,
              yaml_file_configured$spec$template$metadata$labels$`batch-group`,
              yaml_file_configured$spec$template$metadata$annotations$USER_TAG_0,
              yaml_file_configured$spec$template$spec$containers[[1]]$args[[2]],
              yaml_file_configured$spec$template$spec$containers[[1]]$resources$limits$cpu,
              yaml_file_configured$spec$template$spec$containers[[1]]$resources$limits$memory,
              yaml_file_configured$spec$template$metadata$annotations$user,
              yaml_file_configured$spec$template$spec$volumes[[1]]$nfs$server,
              yaml_file_configured$spec$template$spec$securityContext$runAsUser,
              yaml_file_configured$spec$template$spec$securityContext$runAsGroup)

  expected <- c(yaml_file_configured$metadata$generateName,
                'test_namespace',
                "group_A",
                "group_A",
                "user_tag",
                paste0("cd ~ && R --slave --no-save --no-restore -f ", "/tst/path/test.R"),
                "2",
                "512M",
                'test_username',
                '0.1.2.3',
                '123',
                '456')

  expect_equal(actual, expected)

})

test_that("function errors when cpu limit is lower than minimum specified in options", {
  low_cpu_limit <- mcpu_to_cpu(getOption("abba.lower.cpu.limit")) * 0.1
  expect_error(configure_k8s_yaml(file_path="/tst/path/test.R",
                              cpu_limit= low_cpu_limit))
})

test_that("function errors when cpu limit is bigger than maximum cpu limit specified in options", {
  high_cpu_limit <- mcpu_to_cpu(getOption("abba.cpu.limit")) * 10
  expect_error(configure_k8s_yaml(file_path="/tst/path/test.R",
                              cpu_limit= high_cpu_limit))
})

test_that("function errors when memory limit is lower than minimum specified in options", {
  low_memory_limit <- memory_to_bytes(getOption("abba.lower.memory.limit")) * 0.1
  expect_error(abba_configure_k8s_yaml_local(file_path="/tst/path/test.R",
                              memory_limit = low_memory_limit))

})

test_that("function errors when memory limit is bigger than maximum specified in options", {
  high_memory_limit <- memory_to_bytes(getOption("abba.memory.limit")) * 10
  expect_error(abba_configure_k8s_yaml_local(file_path="/tst/path/test.R",
                              memory_limit = high_memory_limit))

})

test_that("Container image can be customized", {
  yaml_file_configured <- configure_k8s_yaml(file_path="/tst/path/test.R",
                                             container='custom_container_image')

  actual <- yaml_file_configured$spec$template$spec$containers[[1]]


  expect_equal(actual[['image']], 'custom_container_image')

})

test_that("User can add custom mounts; volumeMounts are updated properly", {
  new_volumes <- list(volumes=list(list(name='mount1',
                                        nfs=list(server='0.0.0.0', path='/mnt/mount1'))),
                      volumeMounts=list(list(name='mount1', mountPath='/mnt/mount1')))

  yaml_file_configured <- configure_k8s_yaml(file_path="/tst/path/test.R",
                                             mounts=new_volumes,
                                             auto_mount_home = TRUE)

  actual_volumeMounts <- yaml_file_configured$spec$template$spec$containers[[1]]$volumeMounts

  expect_volumeMounts <-
    c(yaml_file_configured$spec$template$spec$containers[[1]]$volumeMounts[1],
      new_volumes$volumeMounts)
  expect_equal(actual_volumeMounts, expect_volumeMounts)

})

test_that("User can add custom mounts; volumes information is updated properly", {
  new_volumes <- list(volumes=list(list(name='mount1',
                                        nfs=list(server='0.0.0.0', path='/mnt/mount1'))),
                      volumeMounts=list(list(name='mount1', mountPath='/mnt/mount1')))

  yaml_file_configured <- configure_k8s_yaml(file_path="/tst/path/test.R",
                                             mounts=new_volumes)

  actual_volumes <- yaml_file_configured$spec$template$spec$volumes
  expect_volumes <- new_volumes$volumes
  expect_equal(actual_volumes, expect_volumes)

})

test_that("Home mount is present when auto_home_mount is TRUE", {
  new_volumes <- list(volumes=list(list(name='mount1',
                                        nfs=list(server='0.0.0.0', path='/mnt/mount1'))),
                      volumeMounts=list(list(name='mount1', mountPath='/mnt/mount1')))

  yaml_file_configured <- configure_k8s_yaml(file_path="/tst/path/test.R",
                                             auto_mount_home=TRUE,
                                             mounts=new_volumes)

  actual_volumes <- yaml_file_configured$spec$template$spec$volumes
  expect_volumes <- c(yaml_file_configured$spec$template$spec$volumes[1],
                      new_volumes$volumes)
  expect_equal(actual_volumes, expect_volumes)

})

test_that("Numeric values are unchanged", {
  expect_equal(mcpu_to_cpu(2.5), 2.5)
})

test_that("Numeric values are rounded up to 1e-3", {
  expect_equal(mcpu_to_cpu(0.4237), 0.424)
})

test_that("Character values without suffix are simply converted to numeric", {
  expect_equal(mcpu_to_cpu('0.500'), 0.5)
})

test_that("mcpu units are correctly converted", {
  expect_equal(mcpu_to_cpu('123m'), 0.123)
})

test_that("mcpu units are correctly converted - 2", {
  expect_equal(mcpu_to_cpu('1200m'), 1.2)
})

test_that("function errors when units other than m are supplied", {
  expect_error(mcpu_to_cpu('1200Ti'))
})

test_that("Numeric values are unchanged", {
  expect_equal(memory_to_bytes(1000), 1000)
})

test_that("Ceiling value is used if fractions are supplied", {
  expect_equal(memory_to_bytes(100.5), 101)
})

test_that("Ceiling value is used if fractions are supplied - in character form", {
  expect_equal(memory_to_bytes('128.1'), 129)
})

test_that("Kilobyte units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('1k'), 1e3)
})

test_that("Megabyte units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('128M'), 128e6)
})

test_that("Gigabytes units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('2.1G'), 2.1e9)
})

test_that("Terabytes units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('0.5T'), 0.5e12)
})

test_that("Kikibyte units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('1ki'), 1024)
})

test_that("Mebibyte units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('128Mi'), 128*1024**2)
})

test_that("Gibibytes units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('2.1Gi'), 2.1*1024**3)
})

test_that("Tebibytes units are correctly converted to bytes", {
  expect_equal(memory_to_bytes('0.5Ti'), 0.5*1024**4)
})

test_that("Function errors when crazy/unrecognized units are supplied", {
  expect_error(memory_to_bytes('0.5P'))
})

#' Get numerical User ID/ Group ID on linux. By default, returns uid/gid of a current user.
#' 'user' argument can be supplied to get uid/gid of specific user
#'
#' @param user Optional; return uid/guid for a specified user, not for current user
#' @param ... other arguments that will be ignored
#'
#' @return a list of uid and gid number for user executing this function
#' @noRd
#' @examples
#' ids <- get_guid()
get_guid <- function(user='', ...){
  output <- system2("id", args=user, stdout=TRUE, stderr=TRUE)

  uid <- stringr::str_extract(output, stringr::regex("(?<=uid\\=)\\d+(?=\\()"))
  gid <- stringr::str_extract(output, stringr::regex("(?<=gid\\=)\\d+(?=\\()"))

  return(list(uid=uid, gid=gid))
}

#' Load the Kubernetes job template stored in package
#'
#' @return Imported YAML as list
#' @noRd
load_k8s_yaml_template <- function() {
  yaml::read_yaml(system.file("job.yaml", package="abba"))
}

test_that("save_yaml saves file when no file_path is specified", {
  config <- load_k8s_yaml_template()
  config_path <- save_yaml(config)
  expect_true(file.exists(config_path))
})

test_that("save_yaml saves file in a temporary directory when no file_path is specified", {
  config <- load_k8s_yaml_template()
  config_path <- save_yaml(config)
  expect_true(grepl(tempdir(), config_path, fixed=TRUE))
})

#' Extract variables created during file sourcing
#'
#' @param rproj_path path to .Rprofile or other file user wants to source before running the batch
#'
#' @return a list of variables and their values
#' @noRd
#'
read_rproj <- function(rproj_path){

  # process different cases that would lead to premature return
  if (is.list(rproj_path)){return(rproj_path)}

  else if (is.null(rproj_path)){return(list())}

  # stop execution if .Rproj file does not exist
  else if (!file.exists(rproj_path)){
    stop(sprintf("Rproject file %s not found.", rproj_path))
  }

  # get current directory and ensure it is restored even if an error occurs
  cwd <- getwd()
  on.exit(setwd(cwd), add = TRUE)

  # change working dir to that of the parent directory of target file
  setwd(dirname(rproj_path))
  # create environment that would store all variables from sourcing the target file
  rproj_env <- new.env()
  # source target file
  source(rproj_path, local=rproj_env)

  # return environment variables that were created during source of target file
  return(as.list(rproj_env))
}

#' Configure slurm submission script using it in sbatch command
#'
#' @param program_path full path to R program
#' @param log_path desirable parent folder for program's log file.
#' @param r_version Version of R that will be used to run the program. Can be specified as a full path to Rscript executable, or as a label of R version that is displayed in the Workbench GUI.
#' @param user_tag custom string that will be added to the job name.
#' @param cpu_cores Amount of CPU cores that will be requested for the job.
#' @param memory Amount of RAM in megabytes that will be requested for the job.
#' @param username user whose permission level is used to execute the script. Defaults to user submitting the job.
#' @param job_timeout time limit for a job. Must be specified in a format of "days-hours:minutes:seconds" If exceeded, job will be cancelled.
#'
#' @return a character vector representing submission script for sbatch command
#' @noRd
#'
configure_slurm_job <- function(program_path='',
                                log_path='',
                                rscript_path='',
                                user_tag='',
                                cpu_cores=getOption("abba.slurm.cpu.cores"),
                                memory=getOption("abba.slurm.memory"),
                                username=NULL,
                                job_timeout=3600
){

  slurm_config_obj <- load_slurm_template()

  program_name <- unlist(strsplit(basename(program_path), '.', fixed = TRUE))[1]

  # cannot have underscores in job name/generate name
  job_name <- paste(gsub('_', '-', program_name), user_tag, sep='-')
  generate_name <- paste0(job_name, '-', uuid::UUIDgenerate())

  # Pull supplied username if provided, otherwise default to local user
  if (is.null(username)){
    service_user <- Sys.info()[["user"]]
  } else {
    service_user <- username
  }
  guid <- get_guid(user=service_user)

  # a function that would try to replace all possible keywords inside the target string
  replace_func <- function(x){
    x <- gsub("SLURM_JOB_UID", guid$uid, x)
    x <- gsub("SLURM_JOB_TIMEOUT", job_timeout, x)
    x <- gsub("CPU_CORES", cpu_cores, x)
    x <- gsub("RAM_MB", memory, x)
    x <- gsub("SLURM_JOB_JOB_NAME", generate_name, x)
    x <- gsub("PROGRAM_LOG_PATH", log_path, x)
    x <- gsub("RSCRIPT_PATH", rscript_path, x)
    x <- gsub("R_PROGRAM_PATH", program_path, x)
    x <- gsub("R_PROGRAM_FOLDER_PATH", dirname(program_path), x)
    return(x)
  }

  # recursively walk the yaml and replace all placeholders with actual values
  recursive_replace <- function(l){
    sapply(l, function(x) if(is.list(x)) recursive_replace(x)
           else if(is.character(x)) replace_func(x)
           else x, USE.NAMES = FALSE)
  }

  slurm_config_obj <- recursive_replace(slurm_config_obj)

  return(slurm_config_obj)

}


# function for submitting slurm config. Takes in a file path to config as the argument
submit_slurm_job_config <- function(slurm_config_path){
  # send the job for execution
  output <- suppressWarnings(system2(command="sbatch",
                                     args=c(slurm_config_path),
                                     stdout=TRUE, stderr=TRUE))
  system_command_error_check(output, "Error submitting the job.")
  # read job id from config and return it for further tracking and reporting
  return(slurm_config_get_job_id(output))
}


# function to get job log path
get_slurm_job_log_path <- function(job_id, ...){
  # get job info
  output <- suppressWarnings(system2(command="scontrol",
                                     args=c("show job", job_id),
                                     stdout=TRUE, stderr=TRUE))
  system_command_error_check(output, "Error getting job log path.")

  parsed_output <- slurm_parse_scontrol_output(output)

  return(parsed_output$StdOut)
}


# decude slurm statuses
decode_slurm_job_status <- function(short_job_status){
  job_statuses <- list(BF  = "BOOT_FAIL",
                       CA  = "CANCELLED",
                       CD  = "COMPLETED",
                       CF  = "CONFIGURING",
                       CG  = "COMPLETING",
                       DL  = "DEADLINE",
                       F   = "FAILED",
                       NF  = "NODE_FAIL",
                       OOM = "OUT_OF_MEMORY",
                       PD  = "PENDING",
                       PR  = "PREEMPTED",
                       R   = "RUNNING",
                       RD  = "RESV_DEL_HOLD",
                       RF  = "REQUEUE_FED",
                       RH  = "REQUEUE_HOLD",
                       RQ  = "REQUEUED",
                       RS  = "RESIZING",
                       RV  = "REVOKED",
                       SI  = "SIGNALING",
                       SE  = "SPECIAL_EXIT",
                       SO  = "STAGE_OUT",
                       ST  = "STOPPED",
                       S   = "SUSPENDED",
                       TO  = "TIMEOUT")
  return (job_statuses[[short_job_status]])
}


# simple function to get slurm job status. Relies on configured SLURM accounting
slurm_get_job_status_sacct0 <- function(job_id, ...){
  output <- suppressWarnings(system2(command="sacct",
                                     args=c(" -j", job_id, "--format=JobID,ExitCode,State"),
                                     stdout=TRUE, stderr=TRUE))

  err_msg <- if(is.null(attr(output, "errmsg"))) "No error message provided" else attr(output, "errmsg")
  if (attr(output, "status") != 0){
    stop(sprintf("Error getting job status. %s.\nError message: %s", output, err_msg))
  }
  # get first line after headers
  job_info <- output[[3]]

  # split words in job_info string by spaces. Expected result is 3 words(job id, exit code, and state)
  status <- strsplit(job_info, "\\s+")[[1]]
  return(status[[length(status)]])
}


# parse the StdOut path from scontrol command output
slurm_parse_scontrol_output <- function(output){
  # only easily parsed information is currently retained
  split_result <- stringr::str_split(stringr::str_trim(output), " ", simplify=TRUE)
  parsed_output <- split_result[stringr::str_count(split_result, "=") == 1]

  split_output <- stringr::str_split(parsed_output, "=")

  output_names <- sapply(split_output, function(x){x[[1]]})
  output_values <- lapply(split_output, function(x){x[[2]]})

  # create a named list with names being job attribute/parameter names
  stats::setNames(object=output_values, output_names)
}


# parse the StdOut path from scontrol command output
slurm_parse_squeue_output <- function(output){
  # squeue output is structured like a csv file with whitespace delimiter
  utils::read.table(text=output, header=TRUE, sep="")

}


# Function to determine where to place program logs depending on supplied log path/program path
slurm_config_determine_log_path <- function(log_path=NULL,
                                            program_path=NULL){
  if (is.null(program_path) || program_path == ''){
    stop(sprintf("Program_path parameter should be a real path, not %s", typeof(program_path)))
  }
  # put log file in r script folder if no log path is supplied
  prog_name <- paste0(tools::file_path_sans_ext(basename(program_path)), '.log')
  if (is.null(log_path) || log_path == ''){
    log_path <- file.path(dirname(program_path), prog_name)
  }
  # if log_path is a folder, name a log file after program name and put it in supplied log_path folder
  else if (!grepl(".", basename(log_path), fixed = TRUE)){
    log_path <- file.path(log_path, prog_name)
    }

  # if log_path is a full file path - return just log_path

  return (log_path)
}


# Function for parsing slurm config to extract job id
slurm_config_get_job_id <- function(output){
  job_id <- stringr::str_extract(output, stringr::regex("(?<=job )\\d+$"))
  return(job_id)
}

test_that("job_statuses deciphers job statuses used by slurm", {
  coded_job_statuses <- c("BF", "CA", "CD", "CF", "CG", "DL", "F", "NF", "OOM", "PD", "PR", "R", "RD", "RF", "RH", "RQ", "RS", "RV", "SI", "SE", "SO", "ST", "S", "TO")
  expected_result <- c("BOOT_FAIL", "CANCELLED", "COMPLETED", "CONFIGURING", "COMPLETING", "DEADLINE", "FAILED", "NODE_FAIL", "OUT_OF_MEMORY", "PENDING", "PREEMPTED", "RUNNING",
                       "RESV_DEL_HOLD", "REQUEUE_FED", "REQUEUE_HOLD", "REQUEUED", "RESIZING", "REVOKED", "SIGNALING", "SPECIAL_EXIT", "STAGE_OUT", "STOPPED", "SUSPENDED", "TIMEOUT")
  actual_result <- sapply(coded_job_statuses, decode_slurm_job_status)
  names(actual_result ) <- NULL
  expect_equal(actual_result, expected_result)
})


test_that("slurm_parse_scontrol_output correctly parses output of scontrol command", {

  output <- c("JobId=71701 Name=hostname",
              "UserId=da(1000) GroupId=da(1000)",
              "Priority=66264 Account=none QOS=normal WCKey=*123",
              "JobState=COMPLETED Reason=None Dependency=(null)",
              "TimeLimit=UNLIMITED Requeue=1 Restarts=0 BatchFlag=0 ExitCode=0:0",
              "SubmitTime=2010-01-05T10:58:40 EligibleTime=2010-01-05T10:58:40",
              "StartTime=2010-01-05T10:58:40 EndTime=2010-01-05T10:58:40",
              "SuspendTime=None SecsPreSuspend=0",
              "Partition=debug AllocNode:Sid=snowflake:4702",
              "ReqNodeList=(null) ExcNodeList=(null)",
              "NodeList=snowflake0",
              "NumNodes=1 NumCPUs=10 CPUs/Task=2 ReqS:C:T=1:1:1",
              "MinCPUsNode=2 MinMemoryNode=0 MinTmpDiskNode=0",
              "Features=(null) Reservation=(null)",
              "OverSubscribe=OK Contiguous=0 Licenses=(null) Network=(null)")

  actual_result <- slurm_parse_scontrol_output(output)

  expect_equal(actual_result$JobId, "71701")
  expect_equal(actual_result$JobState, "COMPLETED")
  expect_equal(40, length(actual_result))
})


test_that("slurm_command_error_check raises an error if input has a non-zero status attribute", {

  output <- c("command output")
  attr(output, "status") <- 1

  expect_error(slurm_command_error_check(output, "Message in case of error"))
})


test_that("slurm_config_determine_log_folder returns program's parent folder if log_path=NULL", {

  actual_result <- slurm_config_determine_log_path(log_path=NULL,
                                                   program_path="path/to/program/prog_name.R")

  expect_equal("path/to/program/prog_name.log", actual_result)
})


test_that("slurm_config_determine_log_folder returns log path + program name if log_path is a path to folder", {

  actual_result <- slurm_config_determine_log_path(log_path="path/to/log",
                                                   program_path="path/to/program/prog_name.R")

  expect_equal("path/to/log/prog_name.log", actual_result)
})


test_that("slurm_config_determine_log_folder returns log path if log_path is a path to file", {

  actual_result <- slurm_config_determine_log_path(log_path="path/to/log/log_name.log",
                                                   program_path="path/to/program/prog_name.R")

  expect_equal("path/to/log/log_name.log", actual_result)
})


test_that("slurm_config_determine_log_folder gives error if program_path is an empty string or a NULL", {

  expect_error(slurm_config_determine_log_path(program_path=NULL))
  expect_error(slurm_config_determine_log_path(program_path=''))
})


test_that("get_slurm_job_log_path returns job log path", {

  mock_system2 <- mock( c("JobId=9998 Name=hostname",
                          "StdOut=/path/to/log/prog9998.log"),
                        cycle = TRUE)
  stub(get_slurm_job_log_path, "system2", mock_system2, depth=2)

  actual_result <- get_slurm_job_log_path("9998")

  expect_equal("/path/to/log/prog9998.log", actual_result)
})

#' Save Slurm SBATCH job file in temporary directory.
#'
#' @param slurm_obj A character vector representing Slurm SBATCH job file
#' @param file_path Optional. Full filepath for to-be-saved SBATCH job config
#'
#' @return Full path to the saved SBATCH job config
#' @noRd
save_slurm_template <- function(slurm_obj, file_path=''){

  config_path <- if(file_path=='') tempfile(fileext = '.submit') else file_path

  writeLines(slurm_obj, config_path)

  return(config_path)
}

#' Replace placeholders in imported YAML config with values supplied by user
#'
#' @param file_path Full path to R file
#' @param batch_group_id A tag to mark jobs inside one batch(i.e. SDTM, ADaM etc.)
#' @param user_tag String that describes what kind of job will be scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container A character string containing a valid container path
#' @param mounts Specifically formatted list with information bout volumes that container would have access to during the run
#' @param namespace Kubernetes namespace to put the job in
#' @param username user whose authority will be used to run the program
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_ip_address IP address for mounting service user home directory
#'
#' @return A nested named list, yaml_file_obj, with placeholders replaced by actual values
#' @noRd
#' @examples
#' config <- configure_k8s_yaml(file_path="/path/to/file.R", user_tag="test program")
configure_k8s_yaml <- function(file_path='',
                               batch_group_id='',
                               user_tag='',
                               cpu_limit= 1L,
                               memory_limit='512M',
                               container=NULL,
                               mounts=NULL,
                               namespace=getOption('abba.k8s.namespace'),
                               username=NULL,
                               auto_mount_home=FALSE,
                               home_nfs_ip_address=getOption('abba.home.nfs.ip.address')){

  yaml_file_obj <- load_k8s_yaml_template()

  program_name <- unlist(strsplit(basename(file_path), '.', fixed = TRUE))[1]

  # By default let 'batch-group' be the lowest possible level - the program name.
  # Otherwise, keep what user has specified.
  if (is.null(batch_group_id) || batch_group_id == '') {
    batch_group_id <- program_name
  }

  # remove home mount if mount_home is FALSE
  if (auto_mount_home==FALSE){
    yaml_file_obj$spec$template$spec$volumes <- NULL
    yaml_file_obj$spec$template$spec$containers[[1]]$volumeMounts <- NULL
  }

  # cannot have underscores in job name/generate name
  job_name <- gsub('_', '-', program_name)
  generate_name <- paste0(job_name, '-', uuid::UUIDgenerate())
  # Pull supplied username if provided, otherwise default to local user
  if (is.null(username)){
    service_user <- Sys.info()[["user"]]
  } else {
    service_user <- username
  }
  guid <- get_guid(user=service_user)

  # Enforce lower and upper limits on cpu resource
  # use mcpu_to_cpu function to make sure compared values have equal units
  cpu_limit <- validate_cpu_limit(cpu_limit)

  # Enforce lower and upper limits on memory
  # use memory_to_bytes function to make sure compared values have equal units
  memory_limit <- validate_memory_limit(memory_limit)

  # update container info in configuration file
  yaml_file_obj <- update_k8s_container(yaml_file_obj, container)

  # add specified mounts to configuration file
  yaml_file_obj <- update_k8s_mounts(yaml_file_obj, mounts)

  # a function that would try to replace all possible keywords inside the target string
  replace_func <- function(x){
    x <- gsub("JOB_NAME", generate_name, x)
    x <- gsub("GENERATE_NAME", generate_name, x)
    x <- gsub("BATCH_GROUP_ID", batch_group_id, x)
    x <- gsub("USER_TAG", user_tag, x)
    x <- gsub("PROGRAM_FULL_PATH", file_path, x)
    x <- gsub("PROGRAM_BASE_NAME", program_name, x)
    x <- gsub("SERVICE_USER", service_user, x)
    x <- gsub("CPU_LIMIT", as.character(cpu_limit), x)
    x <- gsub("MEMORY_LIMIT", memory_limit, x)
    x <- gsub("RUN_AS_USER", guid$uid, x)
    x <- gsub("RUN_AS_GROUP", guid$gid, x)
    x <- gsub("K8S_NAMESPACE", namespace, x)
    if (!is.null(home_nfs_ip_address)) {
      x <- gsub("DEFAULT_NFS_MOUNT_IP_ADDRESS", home_nfs_ip_address, x)
    }
    return(x)
  }

  # recursively walk the yaml and replace all placeholders with actual values
  recursive_replace <- function(l){
    lapply(l, function(x) if(is.list(x)) recursive_replace(x)
                          else if(is.character(x)) replace_func(x)
                          else x)
  }

  yaml_file_obj <- recursive_replace(yaml_file_obj)

  # convert to integer after replacement with strings
  yaml_file_obj$spec$template$spec$securityContext$runAsUser <-
    as.integer(yaml_file_obj$spec$template$spec$securityContext$runAsUser)
  yaml_file_obj$spec$template$spec$securityContext$runAsGroup <-
    as.integer(yaml_file_obj$spec$template$spec$securityContext$runAsGroup)

  return(yaml_file_obj)

}

#' Update mounts information in yaml config. Mounts information should follow
#' specific format: it has to be a list with 2 attributes, volumes and volumeMounts.
#' The yaml object will not be updated if mounts object fails validation. Validation
#' is carried out by mount.is.valid function.
#' Example of proper mounts object:
#' mounts=list(volumes=list(list(name='mount1',
#'                               nfs=list(server='0.0.0.0', path='/mnt/mount1'))),
#'             volumeMounts=list(list(name='mount1', mountPath='/mnt/mount1')))
#'
#' @param yaml yaml file representation in a form of nested list
#' @param mounts list that contains information about volumes that user wants to mount.
#'
#' @return updated
#' @noRd
#'
#' @examples \dontrun{
#' yaml <- update_k8s_mounts(yaml,
#'                       list(volumes=list(list(name='mount1',
#'                                              nfs=list(server='0.0.0.0',
#'                                                       path='/mnt/mount1'))),
#'                            volumeMounts=list(list(name='mount1',
#'                                                   mountPath='/mnt/mount1'))))}
update_k8s_mounts <- function(yaml, mounts){

  # validate mounts
  if (!validate_k8s_mount(mounts)){return(yaml)}

  # update the fields in yaml after all checks are successful
  yaml$spec$template$spec$volumes <- c(yaml$spec$template$spec$volumes, mounts$volumes)

  yaml$spec$template$spec$containers[[1]]$volumeMounts <-
    c(yaml$spec$template$spec$containers[[1]]$volumeMounts, mounts$volumeMounts)

  return(yaml)
}

#' Update container information in yaml configuration object
#'
#' @param yaml yaml file representation in a form of nested list
#' @param container A character string containing a valid container path
#'
#' @return updated yaml object
#' @noRd
#'
#' @examples \dontrun{
#' yaml <- update_k8s_container(yaml, 'mycompany.registry.io/container-1.0.0')}
update_k8s_container <- function(yaml, container){
  # return unmodified yaml if supplied container information is not correctly specified
  validate_k8s_container(container)
  yaml$spec$template$spec$containers[[1]]$image <- container
  return(yaml)
}

# check if command executed via system2 produced any errors
# Generate R error with user specified message if exit code is not 0
system_command_error_check <- function(cmd_output, msg){
  err_msg <- if(is.null(attr(cmd_output, "errmsg"))) "" else paste("Error details:", attr(cmd_output, "errmsg"))
  if (!is.null(attr(cmd_output, "status")) && attr(cmd_output, "status") != 0){
    stop(sprintf(paste0(msg, " %s.\n%s"), cmd_output, err_msg))
  }
}

#' Send POST request to submit-job endpoint
#'
#' @param file_path Full path to R file
#' @param batch_group_id Group ID for batch processing
#' @param user_tag Optional; a string that describes what kind of job will be
#'   scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container A string containing a permitted container name.
#' @param mounts Specifically formatted list with information bout volumes that
#'   container would have access to during the run
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_ip_address IP address for mounting service user home directory
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY.
#' @param ... Other arguments that will be ignored
#'
#' @return body of request`s response in a list format
#' @export
#'
#' @examples \dontrun{
#' response <- abba_submit_job('/path/to/R/program.R')}
abba_submit_job <-
  function(file_path,
           batch_group_id='',
           user_tag='',
           cpu_limit=1L,
           memory_limit='512M',
           container='',
           mounts='',
           auto_mount_home=FALSE,
           home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY"),
           ...) {

    # address for a submit_job_and_watch endpoint
    req <- httr2::request(paste(api_address, 'submit-job', sep='/')) %>%
      add_api_key_to_header(api_key=api_key)

    # add a json body with all parameters
    req <- httr2::req_body_json(req, list(file_path=file_path,
                                          batch_group_id=batch_group_id,
                                          user_tag=user_tag,
                                          cpu_limit=cpu_limit,
                                          memory_limit=memory_limit,
                                          mounts=mounts,
                                          container=container,
                                          auto_mount_home=auto_mount_home,
                                          home_nfs_ip_address=home_nfs_ip_address))
    # send the request to API
    resp <- httr2::req_error(req, body = submit_job_error_body) %>% httr2::req_perform()
    result <- httr2::resp_body_json(resp)
    # unlist values inside the results
    result$job_id <- unlist(result$job_id)
    result$batch_id <- unlist(result$batch_id)
    # return response as a list
    return(result)
  }


#' Send job and poll for status. This function sends multiple requests so it
#' won't time out on heavy jobs
#'
#' @param file_path Full path to R file
#' @param batch_group_id Group ID for batch processing
#' @param user_tag Optional; a string that describes what kind of job will be
#'   scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container list that contains container name and image name
#' @param mounts Specifically formatted list with information bout volumes that
#'   container would have access to during the run
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_ip_address IP address for mounting service user home directory
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param poll_interval_seconds Total time to wait before timeout in seconds
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY.
#'
#' @return list with 2 attributes: job_id for submitted job`s id, and its logs
#' @export
#'
#' @examples \dontrun{
#' response <- abba_submit_and_get_log('/path/to/R/program.R', batch_group_id='SDTM')}
abba_submit_and_get_log <-
  function(file_path,
           batch_group_id='',
           user_tag='',
           cpu_limit=1L,
           memory_limit='512M',
           container='',
           mounts='',
           auto_mount_home=FALSE,
           home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
           poll_interval_seconds = 3,
           timeout_seconds = 600,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    # submit the job
    job_id <- abba_submit_job(file_path=file_path,
                              batch_group_id=batch_group_id,
                              user_tag=user_tag,
                              cpu_limit=cpu_limit,
                              memory_limit=memory_limit,
                              mounts=mounts,
                              container=container,
                              auto_mount_home=auto_mount_home,
                              home_nfs_ip_address=home_nfs_ip_address,
                              api_address=api_address,
                              api_key=api_key)

    start_time <- Sys.time()

    # Poll for job status in the specified batch group
    while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

      # Get the status
      job_details <- abba_get_job_status(job_id$job_id,
                                         api_address=api_address,
                                         api_key=api_key)

      # Get the names of the outer list in job_details
      status_names <- names(job_details)

      # Check if neither "Pending" nor "Running" is a name in job_details
      if (!"Pending" %in% status_names && !"Running" %in% status_names) {
        break # Break if no "Pending" or "Running" in the names of job_details
      }

      # Wait for the specified interval before polling again
      Sys.sleep(poll_interval_seconds)
    }

    # get logs after job is no long in pending/running stage
    program_name <- tools::file_path_sans_ext(basename(job_details[[1]]$Jobs[[1]]$path))
    logs <- abba_get_job_log(job_id$job_id,
                             api_address = api_address,
                             api_key=api_key)
    result = list()
    result[[program_name]] = logs[[1]]
    # return response as a list
    return(result)
  }


#' Monitor job status and retrieve its log when the job finishes running
#'
#' @param job_id unique job identificator
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param poll_interval_seconds Total time to wait before timeout in seconds
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY
#' @param ... Other arguments that will be ignored
#'
#' @return list with 2 attributes: job_id for submitted job`s id, and its logs
#' @export
#'
#' @examples \dontrun{
#' response <- abba_wait_for_job_log('sdfj4-asdjlk-bjslk')}
abba_wait_for_job_log <-
  function(job_id,
           poll_interval_seconds = 3,
           timeout_seconds = 600,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY"),
           ...) {

    start_time <- Sys.time()

    # Poll for job status in the specified batch group
    while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

      # Get the status
      job_details <- abba_get_job_status(job_id,
                                         api_address=api_address,
                                         api_key=api_key)

      # Get the names of the outer list in job_details
      status_names <- names(job_details)

      # Check if neither "Pending" nor "Running" is a name in job_details
      if (!"Pending" %in% status_names && !"Running" %in% status_names) {
        break # Break if no "Pending" or "Running" in the names of job_details
      }

      # Wait for the specified interval before polling again
      Sys.sleep(poll_interval_seconds)
    }

    # get logs after job is no long in pending/running stage
    logs <- abba_get_job_log(job_id,
                             api_address = api_address,
                             api_key=api_key)

    # return response as a list
    return(logs)
  }


#' Monitor batch status and retrieve its log when the all jobs in batch finish
#' running
#'
#' @param batch_id unique batch identificator
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param poll_interval_seconds Total time to wait before timeout in seconds
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY
#'
#' @return list with 2 sublists: job_ids and their logs
#' @export
#'
#' @examples \dontrun{
#' response <- abba_wait_for_batch_log('batch-sdtm-sdfj4-asdjlk-bjslk')}
abba_wait_for_batch_log <-
  function(batch_id,
           poll_interval_seconds = 3,
           timeout_seconds = 600,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    start_time <- Sys.time()

    # Poll for job status in the specified batch group
    while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

      # Get the status
      batch_details <- abba_get_batch_status(batch_id,
                                             api_address=api_address,
                                             api_key=api_key)

      # Get the names of the outer list in job_details
      status_names <- names(batch_details)

      # Check if neither "Pending" nor "Running" is a name in job_details
      if (!"Pending" %in% status_names && !"Running" %in% status_names) {
        break # Break if no "Pending" or "Running" in the names of job_details
      }

      # Wait for the specified interval before polling again
      Sys.sleep(poll_interval_seconds)
    }

    # get logs after job is no long in pending/running stage
    logs <- abba_get_batch_log(batch_id,
                               api_address = api_address,
                               api_key=api_key)

    # return response as a list
    return(logs)
  }


#' Send GET request to get logs of specified Jobs
#'
#' @param job_ids A list of job IDs to get logs for
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY
#'
#' @return body of request`s response in a list format
#' @export
#'
#' @examples \dontrun{
#' response <- abba_get_job_log('1234j-13j4l5k-ajslfd')}
abba_get_job_log <-
  function(job_ids,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    # address for a submit_job_and_watch endpoint
    req <- httr2::request(paste(api_address, 'job-log', sep='/'))

    # add a json body with all parameters
    req <- httr2::req_body_json(req, list(job_ids=job_ids)) %>%
      httr2::req_method("GET") %>%
      add_api_key_to_header(api_key=api_key)
    # send the request to API
    resp <- httr2::req_error(req, body = submit_job_error_body) %>% httr2::req_perform()
    result <- httr2::resp_body_json(resp)
    # unlist the log
    for (i in 1:length(result)){
      result[[i]]$job_id <- unlist(result[[i]]$job_id)
      result[[i]]$log <- unlist(result[[i]]$log)
    }
    # return response as a list
    return(result)
  }


#' Send GET request to get logs of all jobs in a batch
#'
#' @param batch_id unique batch identificator
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY
#'
#' @return body of request`s response in a list format
#' @export
#'
#' @examples \dontrun{
#' response <- abba_get_job_log('1234j-13j4l5k-ajslfd')}
abba_get_batch_log <-
  function(batch_id,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    # address for a submit_job_and_watch endpoint
    req <- httr2::request(paste(api_address, 'batch-log', sep='/'))

    # add a json body with all parameters
    req <- httr2::req_body_json(req, list(batch_id=batch_id)) %>%
      httr2::req_method("GET") %>%
      add_api_key_to_header(api_key=api_key)
    # send the request to API
    resp <- httr2::req_error(req, body = submit_job_error_body) %>% httr2::req_perform()
    result <- httr2::resp_body_json(resp)
    # unlist the log
    for (i in 1:length(result)){
      result[[i]]$pod_id <- unlist(result[[i]]$pod_id)
      result[[i]]$log <- unlist(result[[i]]$log)
    }
    # return response as a list
    return(result)
  }


#' Send GET request to get batch job statuses
#'
#' @param batch_id batch ID to get status for
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY
#'
#' @return body of request`s response in a list format
#' @export
#'
#' @examples \dontrun{
#' response <- abba_get_batch_status('1234j-13j4l5k-ajslfd')}
abba_get_batch_status <-
  function(batch_id,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    # address for a submit_job_and_watch endpoint
    req <- httr2::request(paste(api_address, 'batch-status', sep='/'))

    # add a json body with all parameters
    req <- httr2::req_body_json(req, list(batch_id=batch_id)) %>%
      httr2::req_method("GET") %>%
      add_api_key_to_header(api_key=api_key)
    # send the request to API
    resp <- httr2::req_error(req, body = submit_job_error_body) %>% httr2::req_perform()

    # return response as a list
    result <- httr2::resp_body_json(resp)
    # unlist id and path
    statuses <- names(result)
    for (status in statuses){
      for (i in 1:length(result[[status]]$Jobs)){
        for (name in names(result[[status]]$Jobs[[i]])){
          result[[status]]$Jobs[[i]][[name]]=unlist(result[[status]]$Jobs[[i]][[name]])
        }
      }
    }

    return(result)
  }

#' Send GET request to get job status
#'
#' @param job_id job IDs to get status for
#' @param api_address URL to send requests to, hosted in Posit Connect. Defaults
#'   to environment variable ABBA_API_ADDRESS.
#' @param api_key API Key for accessing restricted endpoints. Defaults to
#'   environment variable ABBA_API_KEY.
#'
#' @return body of request`s response in a list format
#' @export
#'
#' @examples \dontrun{
#' response <- abba_get_job_status('1234j-13j4l5k-ajslfd')}
abba_get_job_status <-
  function(job_id,
           api_address=Sys.getenv("ABBA_API_ADDRESS"),
           api_key=Sys.getenv("ABBA_API_KEY")) {

    # address for a submit_job_and_watch endpoint
    req <- httr2::request(paste(api_address, 'job-status', sep='/'))

    # add a json body with all parameters
    req <- httr2::req_body_json(req, list(job_id=job_id)) %>%
      httr2::req_method("GET") %>%
      add_api_key_to_header(api_key=api_key)
    # send the request to
    resp <- httr2::req_error(req, body = submit_job_error_body) %>% httr2::req_perform()

    # return response as a list
    result <- httr2::resp_body_json(resp)
    # unlist id and path
    statuses <- names(result)
    for (status in statuses){
      for (i in 1:length(result[[status]]$Jobs)){
        for (name in names(result[[status]]$Jobs[[i]])){
          result[[status]]$Jobs[[i]][[name]]=unlist(result[[status]]$Jobs[[i]][[name]])
        }
      }
    }

    return(result)
  }

#' Function used to extract error message from response body
#'
#' @param resp httr2`s response
#'
#' @return message attribute from response body
#' @noRd
submit_job_error_body <- function(resp) {
  httr2::resp_body_json(resp)$message
}

library(mockery)

test_that("Request headers update properly with auth keys", {
  req <- httr2::request("www.google.com")
  expect_error(add_api_key_to_header(req, api_key = list()))

  old_key <- Sys.getenv("CONNECT_API_KEY")
  Sys.setenv(CONNECT_API_KEY = "12345ABCDE")
  updated_req <- add_api_key_to_header(req, api_key=Sys.getenv("CONNECT_API_KEY"))

  expect_true("Authorization" %in% names(updated_req$headers))

  Sys.setenv(CONNECT_API_KEY = old_key)
})

test_that("abba_submit_job properly unpacks the response", {
  mock_request <- mock('request1')
  mock_add_api_key_to_header <- mock('request2')
  mock_req_body_json <- mock('request3')
  mock_req_error <- mock('request4')
  mock_req_perform <- mock('request5')
  mock_resp_body_json <- mock(list(job_id=list('unique-job-id'),
                                   batch_id=list('unique-batch-id')))

  stub(abba_submit_job, "httr2::request", mock_request)
  stub(abba_submit_job, "add_api_key_to_header", mock_add_api_key_to_header)
  stub(abba_submit_job, "httr2::req_body_json", mock_req_body_json)
  stub(abba_submit_job, "httr2::req_error", mock_req_error)
  stub(abba_submit_job, "httr2::req_perform", mock_req_perform)
  stub(abba_submit_job, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- abba_submit_job('/path/to/program.R',
                                   batch_group_id='unique-batch-id',
                                   user_tag = 'unique-user-tag',
                                   cpu_limit=1,
                                   memory_limit='500M',
                                   container='container',
                                   mounts='mounts',
                                   auto_mount_home=TRUE,
                                   home_nfs_ip_address='0.0.0.0',
                                   api_address='0.0.0.0',
                                   api_key='ABBA_API_KEY')

  expect_called(mock_request, 1)
  expect_called(mock_add_api_key_to_header, 1)
  expect_called(mock_req_body_json, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_req_perform, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_resp_body_json, 1)

  expect_args(mock_request, 1, '0.0.0.0/submit-job')

  expect_args(mock_add_api_key_to_header, 1, "request1", api_key='ABBA_API_KEY')

  expect_args(mock_req_body_json, 1,
              "request2",
              list(file_path='/path/to/program.R',
                   batch_group_id='unique-batch-id',
                   user_tag='unique-user-tag',
                   cpu_limit= 1,
                   memory_limit='500M',
                   mounts='mounts',
                   container='container',
                   auto_mount_home=TRUE,
                   home_nfs_ip_address='0.0.0.0')
  )

  expect_equal(result_actual, list(job_id='unique-job-id', batch_id='unique-batch-id'))
})

test_that("abba_submit_and_get_log properly waits for job to finish and returns log", {

  # create mocks for inside functions
  mock_abba_submit_job <- mock(list(job_id='job-id1', batch_id='batch_id1'))
  mock_abba_get_job_status <- mock(
    list(Pending=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R")))),
    list(Running=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R")))),
    list(Succeded=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"))))
  )
  mock_abba_get_job_log <- mock(list(list(job_id='job-id1', log=c('log1', 'log2'))))

  # mock inside functions
  stub(abba_submit_and_get_log, "abba_submit_job", mock_abba_submit_job)
  stub(abba_submit_and_get_log, "abba_get_job_status", mock_abba_get_job_status)
  stub(abba_submit_and_get_log, "abba_get_job_log", mock_abba_get_job_log)

  result_actual <- abba_submit_and_get_log(
    '/path/to/program.R',
    batch_group_id='unique-batch-id',
    user_tag = 'unique-user-tag',
    cpu_limit=1,
    memory_limit='500M',
    container='container',
    mounts='mounts',
    auto_mount_home=FALSE,
    poll_interval_seconds=0.1,
    timeout_seconds=5,
    api_address='0.0.0.0',
    api_key='ABBA_API_KEY')

  expect_called(mock_abba_submit_job, 1)
  expect_called(mock_abba_get_job_status, 3)
  expect_called(mock_abba_get_job_log, 1)

  expect_args(mock_abba_submit_job, 1,
              file_path='/path/to/program.R',
              batch_group_id='unique-batch-id',
              user_tag='unique-user-tag',
              cpu_limit= 1,
              memory_limit='500M',
              mounts='mounts',
              container='container',
              auto_mount_home=FALSE,
              home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')

  expect_args(mock_abba_get_job_status, 1,
              "job-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')
  expect_args(mock_abba_get_job_log, 1,
              "job-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')


  expect_equal(result_actual,
               list(script1=list(job_id='job-id1', log=c('log1', 'log2'))))
})

test_that("abba_wait_for_job_log properly waits for job to finish", {

  # create mocks for inside functions
  mock_abba_get_job_status <- mock(
    list(Pending=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R")))),
    list(Running=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R")))),
    list(Succeded=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"))))
  )
  mock_abba_get_job_log <- mock(list(list(job_id='job-id1', log=c('log1', 'log2'))))

  # mock inside functions
  stub(abba_wait_for_job_log, "abba_get_job_status", mock_abba_get_job_status)
  stub(abba_wait_for_job_log, "abba_get_job_log", mock_abba_get_job_log)

  result_actual <- abba_wait_for_job_log(
    'job-id1',
    poll_interval_seconds=0.1,
    timeout_seconds=5,
    api_address='0.0.0.0',
    api_key='ABBA_API_KEY')

  expect_called(mock_abba_get_job_status, 3)
  expect_called(mock_abba_get_job_log, 1)

  expect_args(mock_abba_get_job_status, 1,
              "job-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')
  expect_args(mock_abba_get_job_log, 1,
              "job-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')


  expect_equal(result_actual,
               list(list(job_id='job-id1', log=c('log1', 'log2'))))
})


test_that("abba_wait_for_batch_log  properly waits for jobs to finish", {

  # create mocks for inside functions
  mock_abba_get_batch_status <- mock(
    list(Pending=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"))),
         Succeeded=list(Jobs=list(list(id='job-id2', path="/path/to/script2.R")))),
    list(Running=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"))),
         Succeeded=list(Jobs=list(list(id='job-id2', path="/path/to/script2.R")))),
    list(Succeded=list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"),
                                 list(id='job-id2', path="/path/to/script2.R"))))
    )
  mock_abba_get_batch_log <- mock(list(list(job_id='job-id1', log=c('log11', 'log12')),
                                       list(job_id='job-id2', log=c('log21', 'log22'))))

  # mock inside functions
  stub(abba_wait_for_batch_log, "abba_get_batch_status", mock_abba_get_batch_status)
  stub(abba_wait_for_batch_log, "abba_get_batch_log", mock_abba_get_batch_log)

  result_actual <- abba_wait_for_batch_log(
    'batch-id1',
    poll_interval_seconds=0.1,
    timeout_seconds=5,
    api_address='0.0.0.0',
    api_key='ABBA_API_KEY')

  expect_called(mock_abba_get_batch_status, 3)
  expect_called(mock_abba_get_batch_log, 1)

  expect_args(mock_abba_get_batch_status, 1,
              "batch-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')
  expect_args(mock_abba_get_batch_log, 1,
              "batch-id1",
              api_address='0.0.0.0',
              api_key='ABBA_API_KEY')


  expect_equal(result_actual,
               list(list(job_id='job-id1', log=c('log11', 'log12')),
                    list(job_id='job-id2', log=c('log21', 'log22'))))
})

test_that("abba_get_job_log properly submits request and unpacks the response", {
  mock_request <- mock('request1')
  mock_req_method <- mock('request2')
  mock_add_api_key_to_header <- mock('request3')
  mock_req_body_json <- mock('request4')
  mock_req_error <- mock('request5')
  mock_req_perform <- mock('request6')
  mock_resp_body_json <- mock(list(list(job_id=list('job-id1'), log=list(list('log11'), list('log12'))),
                              list(job_id=list('job-id2'), log=list(list('log21'), list('log22')))))

  stub(abba_get_job_log, "httr2::request", mock_request)
  stub(abba_get_job_log, "httr2::req_method", mock_req_method)
  stub(abba_get_job_log, "add_api_key_to_header", mock_add_api_key_to_header)
  stub(abba_get_job_log, "httr2::req_body_json", mock_req_body_json)
  stub(abba_get_job_log, "httr2::req_error", mock_req_error)
  stub(abba_get_job_log, "httr2::req_perform", mock_req_perform)
  stub(abba_get_job_log, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- abba_get_job_log(c('job-id1', 'job-id2'),
                                    api_address='0.0.0.0',
                                    api_key='ABBA_API_KEY')

  expect_called(mock_request, 1)
  expect_called(mock_add_api_key_to_header, 1)
  expect_called(mock_req_body_json, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_req_perform, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_resp_body_json, 1)

  expect_args(mock_request, 1, '0.0.0.0/job-log')

  expect_args(mock_add_api_key_to_header, 1, "request2", api_key='ABBA_API_KEY')

  expect_args(mock_req_body_json, 1,
              "request1",
              list(job_ids=c('job-id1', 'job-id2'))
  )

  expect_equal(result_actual,
               list(list(job_id='job-id1', log=c('log11', 'log12')),
                    list(job_id='job-id2', log=c('log21', 'log22')))
  )
})

test_that("abba_get_batch_log properly submits request and unpacks the response", {
  mock_request <- mock('request1')
  mock_req_method <- mock('request2')
  mock_add_api_key_to_header <- mock('request3')
  mock_req_body_json <- mock('request4')
  mock_req_error <- mock('request5')
  mock_req_perform <- mock('request6')
  mock_resp_body_json <- mock(list(list(pod_id=list('pod-id1'), log=list(list('log11'), list('log12'))),
                                   list(pod_id=list('pod-id2'), log=list(list('log21'), list('log22')))))

  stub(abba_get_batch_log, "httr2::request", mock_request)
  stub(abba_get_batch_log, "httr2::req_method", mock_req_method)
  stub(abba_get_batch_log, "add_api_key_to_header", mock_add_api_key_to_header)
  stub(abba_get_batch_log, "httr2::req_body_json", mock_req_body_json)
  stub(abba_get_batch_log, "httr2::req_error", mock_req_error)
  stub(abba_get_batch_log, "httr2::req_perform", mock_req_perform)
  stub(abba_get_batch_log, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- abba_get_batch_log(c('batch-id1'),
                                      api_address='0.0.0.0',
                                      api_key='ABBA_API_KEY')

  expect_called(mock_request, 1)
  expect_called(mock_add_api_key_to_header, 1)
  expect_called(mock_req_body_json, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_req_perform, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_resp_body_json, 1)

  expect_args(mock_request, 1, '0.0.0.0/batch-log')

  expect_args(mock_add_api_key_to_header, 1, "request2", api_key='ABBA_API_KEY')

  expect_args(mock_req_body_json, 1,
              "request1",
              list(batch_id='batch-id1')
  )

  expect_equal(result_actual,
               list(list(pod_id='pod-id1', log=c('log11', 'log12')),
                    list(pod_id='pod-id2', log=c('log21', 'log22')))
  )
})

test_that("abba_get_batch_status properly submits request and unpacks the response", {
  mock_request <- mock('request1')
  mock_req_method <- mock('request2')
  mock_add_api_key_to_header <- mock('request3')
  mock_req_body_json <- mock('request4')
  mock_req_error <- mock('request5')
  mock_req_perform <- mock('request6')
  mock_resp_body_json <- mock(list(
    Pending = list(Jobs=list(list(id=list('job-id1'), path=list("/path/to/script1.R")))),
    Running = list(Jobs=list(list(id=list('job-id2'), path=list("/path/to/script2.R")))),
    Failed =  list(Jobs=list(list(id=list('job-id3'), path=list("/path/to/script3.R")))),
    Succeeded=list(Jobs=list(list(id=list('job-id4'), path=list("/path/to/script4.R")))),
    Unknown = list(Jobs=list(list(id=list('job-id5'), path=list("/path/to/script5.R"))))
  ))

  stub(abba_get_batch_status, "httr2::request", mock_request)
  stub(abba_get_batch_status, "httr2::req_method", mock_req_method)
  stub(abba_get_batch_status, "add_api_key_to_header", mock_add_api_key_to_header)
  stub(abba_get_batch_status, "httr2::req_body_json", mock_req_body_json)
  stub(abba_get_batch_status, "httr2::req_error", mock_req_error)
  stub(abba_get_batch_status, "httr2::req_perform", mock_req_perform)
  stub(abba_get_batch_status, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- abba_get_batch_status('batch-id1',
                                        api_address='0.0.0.0',
                                        api_key='ABBA_API_KEY')

  expect_called(mock_request, 1)
  expect_called(mock_add_api_key_to_header, 1)
  expect_called(mock_req_body_json, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_req_perform, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_resp_body_json, 1)

  expect_args(mock_request, 1, '0.0.0.0/batch-status')

  expect_args(mock_add_api_key_to_header, 1, "request2", api_key='ABBA_API_KEY')

  expect_args(mock_req_body_json, 1,
              "request1",
              list(batch_id='batch-id1')
  )

  expect_equal(result_actual,
               list(Pending = list(Jobs=list(list(id='job-id1', path="/path/to/script1.R"))),
                    Running = list(Jobs=list(list(id='job-id2', path="/path/to/script2.R"))),
                    Failed =  list(Jobs=list(list(id='job-id3', path="/path/to/script3.R"))),
                    Succeeded=list(Jobs=list(list(id='job-id4', path="/path/to/script4.R"))),
                    Unknown = list(Jobs=list(list(id='job-id5', path="/path/to/script5.R")))
  ))
})

test_that("abba_get_job_status properly submits request and unpacks the response", {
  mock_request <- mock('request1')
  mock_req_method <- mock('request2')
  mock_add_api_key_to_header <- mock('request3')
  mock_req_body_json <- mock('request4')
  mock_req_error <- mock('request5')
  mock_req_perform <- mock('request6')
  mock_resp_body_json <- mock(list(
    Pending = list(Jobs=list(list(id=list('job-id1'), path=list("/path/to/script1.R"))))
  ))

  stub(abba_get_job_status, "httr2::request", mock_request)
  stub(abba_get_job_status, "httr2::req_method", mock_req_method)
  stub(abba_get_job_status, "add_api_key_to_header", mock_add_api_key_to_header)
  stub(abba_get_job_status, "httr2::req_body_json", mock_req_body_json)
  stub(abba_get_job_status, "httr2::req_error", mock_req_error)
  stub(abba_get_job_status, "httr2::req_perform", mock_req_perform)
  stub(abba_get_job_status, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- abba_get_job_status('job-id1',
                                       api_address='0.0.0.0',
                                       api_key='ABBA_API_KEY')

  expect_called(mock_request, 1)
  expect_called(mock_add_api_key_to_header, 1)
  expect_called(mock_req_body_json, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_req_perform, 1)
  expect_called(mock_req_error, 1)
  expect_called(mock_resp_body_json, 1)

  expect_args(mock_request, 1, '0.0.0.0/job-status')

  expect_args(mock_add_api_key_to_header, 1, "request2", api_key='ABBA_API_KEY')

  expect_args(mock_req_body_json, 1,
              "request1",
              list(job_id='job-id1')
  )

  expect_equal(result_actual,
               list(Pending = list(Jobs=list(list(id='job-id1', path="/path/to/script1.R")))
               ))
})

test_that("submit_job_error_body returns message attribute from response body", {

  mock_resp_body_json <- mock(list(error='test_error', message='test error message'))
  stub(submit_job_error_body, "httr2::resp_body_json", mock_resp_body_json)

  result_actual <- submit_job_error_body('resp')

  expect_called(mock_resp_body_json, 1)
  expect_args(mock_resp_body_json, 1, 'resp')
  expect_equal(result_actual, 'test error message')
})

#' Get job log from Kubernetes for a single job
#'
#' @param job_id a string that uniquely identifies the job
#' @param namespace Kubernetes namespace to search for job
#'
#' @noRd
#' @return A list containing job_id and a character vector with jobs log
#'
get_k8s_job_log0 <- function(job_id,
                             namespace=getOption('abba.k8s.namespace')){

  # run the command for outputting job log
  # this command will generate warning in case job with a given ID does not exist
  # that's why there is a suppressWarnings in place
  log <- suppressWarnings(system2(command="kubectl",
                 args=c("logs", "-n" , namespace, paste0("jobs/", job_id)),
                 stdout=TRUE, stderr=TRUE))

  # check if log was retrieved without errors and raise an R error if command's exit code is not 0
  system_command_error_check(log, "Error getting job log")

  return(list(job_id=job_id, log=as.character(log)))
}


#' Get log for a pod
#'
#' @param pod_id Pod ID to get logs for
#' @param namespace Kubernetes namespace to search for pod
#'
#' @noRd
#' @return A character vector containing pod log.
#'
get_k8s_pod_log0 <- function(pod_id,
                             namespace=getOption('abba.k8s.namespace')){
  # run the command for outputting job log
  # this command will generate warning in case job with a given ID does not exist
  # that's why there is a suppressWarnings in place
  log <- suppressWarnings(system2(command="kubectl",
                                  args=c("logs", "-n" , namespace, pod_id),
                                  stdout=TRUE, stderr=TRUE))

  # check if log was retrieved without errors and raise an R error if command's exit code is not 0
  system_command_error_check(log, "Error getting job log")

  return(list(pod_id=pod_id, log=as.character(log)))
}


#' Get log for every job specified in an input vector/list
#'
#' @param job_ids A list of job IDs to get logs for
#' @param namespace Kubernetes namespace to search for job
#'
#' @export
#' @return A list of job logs. Each list entry will contain complete log for a job
#'
#' @examples \dontrun{
#' logs <- abba_get_k8s_job_log_local(c("job-sdtm-abc123", "job-adam-def456"))
#' }
abba_get_k8s_job_log_local <- function(job_ids,
                                       namespace=getOption('abba.k8s.namespace')){
  logs <- lapply(job_ids, function(x) get_k8s_job_log0(x, namespace=namespace))
  return(logs)
}


#' Return list of logs for jobs that are marked with a given batch ID
#'
#' @param batch_id string containing batch ID
#' @param namespace Kubernetes namespace to search for batch jobs
#'
#' @export
#' @return Job logs in a from of list consisting of character vectors
#'
#' @examples \dontrun{
#' logs <- abba_get_k8s_batch_log_local("batch-sdtm-abc123")
#' }
abba_get_k8s_batch_log_local <- function(batch_id,
                                         namespace=getOption('abba.k8s.namespace')){

  # get all pod IDs belonging to a given batch
  pod_ids <- get_k8s_job_ids_from_batch(batch_id, namespace=namespace)

  logs <- lapply(pod_ids, function(x) get_k8s_pod_log0(x, namespace=namespace))

  return(logs)
}

test_that("Supplying non-existing job id produces an error", {
  expect_error(get_k8s_job_log0("non_exst_job_id"))
})

test_that("Given a non-existing pod id, get_k8s_pod_log0 produces an error", {
  expect_error(get_k8s_pod_log0("non-existing-pod-id-2135"))
})

test_that("Supplying more than 1 job id returns a list with amount of elements equal to amount of supplied job ids", {
  m <- mock('job 1 log line 1', 'job 2 log line 1')
  stub(abba_get_k8s_job_log_local, 'system2', m, depth=2)
  log <- abba_get_k8s_job_log_local(c("job_id1", "job_id2"))
  expect_equal(length(log), 2L)
})

test_that("Given a non-existing batch id, get_batch_log returns an empty list", {
  expect_error(abba_get_k8s_batch_log_local("non-existing-batch-id-2135"))
})

test_that("get_k8s_pod_log0 returns log for an existing pod", {
  m <- mock(c('log line 1', 'log line 2'))
  stub(get_k8s_pod_log0, 'system2', m)
  log <- get_k8s_pod_log0("unique_pod_id")
  expected <- list(pod_id="unique_pod_id", log=c('log line 1', 'log line 2'))
  expect_equal(log, expected)
})

#' Calculate and Save File's Hash Sum.
#'
#' Read in the file's contents, calculate its hash sum and save it under
#' `cache_folder`. Later this information could be used to decide if the
#' program needs to be re-run (due to updates made to the code).
#'
#' @param file_path A full path to the file we want to read in and generate
#' a hash sum.
#' @param cache_folder A full path to the folder where the hash sum file will
#' be written. Required; must be supplied explicitly by the caller.
#' @return A hash sum of the file's contents.
#' @noRd
abba_save_file_cache <- function(file_path, cache_folder, ...) {
  if (is.null(cache_folder)){
    stop("cache_folder must be supplied; abba does not write caches to a default location.")
  }
  if (is.null(file_path)||is.na(file_path)||file_path==""||!file.exists(file_path)){
    # silently return NULL if supplied invalid file path
    message(sprintf("Could not calculate file hash, file %s does not exist.", file_path))
    return(NULL)
  }
  hash_value <- digest::digest(file_path, algo = "md5", file = TRUE)

  if (!dir.exists(cache_folder)) {
    dir.create(cache_folder, recursive = TRUE)
  }

  cache_file <- file.path(cache_folder, paste0(basename(file_path), ".cache"))
  writeLines(hash_value, cache_file)

  return(hash_value)
}


#' Check the Current File's Hash Sum versus the Hash Sum for the File Generated
#' Earlier.
#'
#' This function checks if the hash of a current file matches the stored
#' hash in the cache. If not - it calls `abba_save_file_cache` to update the
#' stored hash for the file.
#'
#' @param file_path A full path to the file we want to read in and generate
#' a hash sum.
#' @param cache_folder A full path to the folder containing hash sum for the
#' input file. Required; must be supplied explicitly by the caller.
#' @param update_cache Controls whether hash sum would be updated if program update is detected. TRUE by default
#' @return `NA` if file_path does not exist, `TRUE` if hash sum matches with what is stored in cache, `FALSE` otherwise.
#' @noRd
cache_match <- function(file_path, cache_folder, update_cache=TRUE, ...) {

  if (is.null(cache_folder)){
    stop("cache_folder must be supplied; abba does not read caches from a default location.")
  }

  # produce a warning if file_path does not exist and return FALSE
  if (is.null(file_path) || !file.exists(file_path)){

    report_str <- sprintf("file_path=%s", file_path)
    if (is.null(file_path)){report_str<-sprintf("file_path=%s", 'NULL')}

    # warning(sprintf("%s does not exist. Cache would not be calculated", report_str))
    return(NA)
  }

  current_hash <- digest::digest(file_path, algo = "md5", file = TRUE)

  cache_file <- file.path(cache_folder, paste0(basename(file_path), ".cache"))

  # save hash for a file and return FALSE if cash does not exist
  if (!file.exists(cache_file)) {
    if(update_cache){abba_save_file_cache(file_path, cache_folder=cache_folder)}
    return(FALSE)
  }

  # compare hashes of current program and program when it was last run in a batch
  cached_hash <- readLines(cache_file)

  if (current_hash == cached_hash) {return(TRUE)}
  else {
    if(update_cache){abba_save_file_cache(file_path, cache_folder=cache_folder)}
    return(FALSE)
    }
}

test_that("abba_save_file_cache creates correct cache file", {
  # Create a temporary program file and cache folder
  program_path <- tempfile()
  cache_folder <- file.path(tempdir(), "abba_cache_test")
  writeLines("Test content", program_path)

  # Run the function
  abba_save_file_cache(program_path, cache_folder=cache_folder)

  # Check if cache file exists with correct content
  cache_file <- file.path(cache_folder, paste0(basename(program_path), ".cache"))
  expect_true(file.exists(cache_file))
  expect_equal(readLines(cache_file), digest::digest(program_path, algo = "md5", file = TRUE))

  # Clean up
  unlink(cache_folder, recursive = TRUE)
  unlink(program_path)
})


test_that("abba_save_file_cache errors when cache_folder is not supplied", {
  program_path <- tempfile()
  writeLines("Test content", program_path)
  expect_error(abba_save_file_cache(program_path, cache_folder=NULL),
               "cache_folder must be supplied")
  unlink(program_path)
})


test_that("cache_match behaves correctly", {
  # Create a temporary program file and cache folder
  program_path <- tempfile()
  cache_folder <- file.path(tempdir(), "abba_cache_test")
  writeLines("Test content", program_path)

  # Initial run, no cache file exists
  expect_false(cache_match(program_path, cache_folder=cache_folder))

  # Second run, cache file exists, content hasn't changed
  expect_true(cache_match(program_path, cache_folder=cache_folder))

  # Modify the file and run again
  writeLines("Modified content", program_path)
  expect_false(cache_match(program_path, cache_folder=cache_folder))

  # Clean up
  unlink(cache_folder, recursive = TRUE)
  unlink(program_path)
})


test_that("cache_match returns NA when input program_path is NULL or does not exist", {
  cache_folder <- tempdir()
  expect_equal(is.na(cache_match(NULL, cache_folder=cache_folder)), TRUE)
  expect_equal(is.na(cache_match('/non/existing/program/path.definitelyNotAnRprogram',
                                 cache_folder=cache_folder)), TRUE)
})


test_that("cache_match errors when cache_folder is not supplied", {
  program_path <- tempfile()
  writeLines("Test content", program_path)
  expect_error(cache_match(program_path, cache_folder=NULL),
               "cache_folder must be supplied")
  unlink(program_path)
})

#' Extract Pod Name from Line
#'
#' This function extracts the pod name from a given line of `kubectl` output.
#' Assumes each line is a comma-separated string where the first part is the pod name.
#'
#' @param line A single line of output from `kubectl`, formatted as "name,status,args".
#' @return A character string representing the name of the pod.
#' @noRd
get_k8s_pod_name <- function(line) {
  parts <- strsplit(line, ",")[[1]]
  return(parts[1])
}

#' Extract Pod Status from Line
#'
#' This function extracts the pod status from a given line of `kubectl` output.
#'
#' @param line A single line of output from `kubectl`, formatted as "name,status,args".
#' @return A character string representing the status of the pod.
#' @noRd
get_k8s_pod_status <- function(line) {
  parts <- strsplit(line, ",")[[1]]
  return(parts[2])
}

#' Extract Pod Args String from Line
#'
#' This function extracts the entire args string from a given line of `kubectl` output.
#'
#' @param line A single line of output from `kubectl`, formatted as "name,status,args".
#' @return A character string representing the args of the pod.
#' @noRd
get_k8s_pod_args_string <- function(line) {
  parts <- strsplit(line, ",")[[1]]
  return(paste(parts[-c(1,2)], collapse = ","))
}

#' Extract Pod Program Name from Line
#'
#' This function extracts the program name from the pod args in a given line of `kubectl` output.
#' It specifically looks for the pattern between '-f' and '.R'.
#'
#' @param line A single line of output from `kubectl`, formatted as "name,status,args".
#' @return A character string representing the program name used by the pod.
#' @noRd
get_k8s_pod_program_name <- function(line) {
  args_string <- get_k8s_pod_args_string(line)
  pattern <- "-f ([^,]+\\.R)"
  matches <- regmatches(args_string, regexpr(pattern, args_string))
  program_name <- ifelse(length(matches) > 0, sub(pattern, "\\1", matches), NA)
  return(program_name)
}

library(testthat)

# Test for abba_get_k8s_pod_name_local function
test_that("abba_get_k8s_pod_name_local extracts pod name correctly", {
  line <- "pod1,Succeeded,/path/to/script.R"
  expect_equal(get_k8s_pod_name(line), "pod1")
})

# Test for abba_get_k8s_pod_status_local function
test_that("abba_get_k8s_pod_status_local extracts pod status correctly", {
  line <- "pod1,Succeeded,/path/to/script.R"
  expect_equal(get_k8s_pod_status(line), "Succeeded")
})

# Test for abba_get_k8s_pod_args_string_local function
test_that("abba_get_k8s_pod_args_string_local extracts args string correctly", {
  line <- "pod1,Succeeded,-f /path/to/script.R"
  expect_equal(get_k8s_pod_args_string(line), "-f /path/to/script.R")
})

# Test for abba_get_k8s_pod_program_name_local function
test_that("abba_get_k8s_pod_program_name_local extracts program name correctly", {
  line <- "pod1,Succeeded,-f /path/to/script.R"
  expect_equal(get_k8s_pod_program_name(line), "/path/to/script.R")

  line_with_additional_args <- "pod1,Succeeded,-f /path/to/script.R, extra, args"
  expect_equal(get_k8s_pod_program_name(line_with_additional_args), "/path/to/script.R")

  line_without_program_name <- "pod1,Succeeded,-f"
  expect_equal(get_k8s_pod_program_name(line_without_program_name), NA)
})

# simple function to create a data frame with results
# WORKS ONLY FOR BATCHES SUBMITTED TO WORKBENCH FOR NOW
compose_batch_results <- function(job_ids=NULL,
                                  prog_names=NULL,
                                  status_func=rslauncher_get_job_display_status,
                                  ...){
  # all components must be specified
  if (is.null(job_ids) & is.null(prog_names)){
    warning("Job IDs and program names were not provided. Batch results will not be composed.")
    return(stats::setNames(data.frame(matrix(ncol = 3, nrow = 0)), c("job_ID", "program_name", "status")))
  }
  if (is.null(status_func)){
    warning("status function was not provided. Batch results will not be composed.")
    return(stats::setNames(data.frame(matrix(ncol = 3, nrow = 0)), c("job_ID", "program_name", "status")))
  }

  statuses <- status_func(job_ids, ...)
  names(job_ids) <- c()
  names(statuses) <- c()

  # pad vectors with NAs so that job-ids and statuses would not be repeated
  # for programs that had not been run
  max_len <- max(lengths(list(job_ids, prog_names)))
  # job_ids <- c(job_ids, rep(NA, max_len - length(job_ids)))
  # statuses <- c(statuses, rep(NA, max_len - length(statuses)))

  results <- as.data.frame(list(job_ID=rep(NA, max_len),
                                program_name=unlist(prog_names),
                                status=rep(NA, max_len)))

  for(i in 1:length(job_ids)){
    job_path <- get_program_path_by_id(job_ids[[i]])
    results[results$program_name == job_path,]$job_ID <- job_ids[[i]]
  }
  results$status <- status_func(results$job_ID)
  return(results)

}


get_id_by_program_path <- function(x){
  jobs <- rstudioapi::launcherGetJobs(tags=paste('rstudio-r-script-job', x, sep=':'))
  results <- jobs[order(sapply(jobs,'[[','submissionTime'), decreasing = TRUE)]
  return(results[[1]]$id)
}


get_program_path_by_id <- function(x){
  job_tag <- tryCatch({rstudioapi::launcherGetJob(x)$tags[[1]]},
           error=function(e){" : "}
  )

  results <- unlist(strsplit(job_tag, ":", fixed=TRUE))[[2]]
  return(results)
}

test_that("compose_batch_results creates a data frame consisting of program names, job ids and execution statuses", {

  status_func <- mock(c('Completed', 'Completed with errors'), cycle=TRUE)
  get_job_path_by_id_func <- mock('/path/to/prog1.R', '/path/to/prog2.R')

  stub(compose_batch_results, "get_program_path_by_id", get_job_path_by_id_func)

  actual_results <- compose_batch_results(job_ids=c('job-id1', 'job-id2'),
                                          prog_names=c('/path/to/prog1.R', '/path/to/prog2.R'),
                                          status_func=status_func)
  expected_results <- as.data.frame(list(
    job_ID=c('job-id1', 'job-id2'),
    program_name=c('/path/to/prog1.R', '/path/to/prog2.R'),
    status=c('Completed', 'Completed with errors')))
  expect_equal(actual_results, expected_results)
})


test_that("compose_batch_results creates returns a NULL and produces a warning if any of input parameters is NULL", {

  status_func <- mock(c('Completed', 'Completed with errors'))
  get_job_path_by_id_func <- mock('/path/to/prog1.R', '/path/to/prog2.R')

  stub(compose_batch_results, "get_program_path_by_id", get_job_path_by_id_func)

  expect_warning(actual_result <-
                   compose_batch_results(job_ids=c('job-id1', 'job-id2'),
                                         prog_names=c('/path/to/prog1.R', '/path/to/prog2.R'),
                                         status_func=NULL))
  expected_result <- setNames(data.frame(matrix(ncol = 3, nrow = 0)), c("job_ID", "program_name", "status"))

  expect_equal(actual_result, expected_result)
})

#' Convert data frame of particular structure into a list suitable for batch submission
#' @param x A data frame. Must contain 2 columns named 'program_name' and 'run_group'
#' and have at least one record.
#' @return a list of vectors containing values from 'program_name' column. Values from
#' 'program_name' will end up in one vector if they have same value in 'run_group' column
#' @noRd
#' @examples
#' execution_list <- dataframe_to_batch_list(
#'   as.data.frame(list(program_name=c("program1.R", "program2.R", "program3.R", "program4.R", "program5.R"),
#'                      run_group=c(1, 2, 2, 3, 3))))
dataframe_to_batch_list <- function(x){

  # return a list/vector back unchanged
  if ((is.list(x) || is.character(x)) && !is.data.frame(x)){return(x)}

  # validate and process the data frame
  validate_batch_data_frame(x)

  result <- list()
  for (i in unique(x$run_group)){
    result <- c(result, list(x[x$run_group == i,]$program_name))
  }
  return(result)
}

#' Validate data frame containing information about grouping of programs for batch submission
#' @param x A data frame
#' @return TRUE if all checks pass
#' @noRd
#' @examples
#' validate_batch_data_frame(
#' as.data.frame(list(program_name=c("program1.R", "program2.R", "program3.R"),
#'                    run_group=c(1, 2, 2))))
validate_batch_data_frame <- function(x){
  if (!is.data.frame(x)){
    stop(sprintf('Batch dataset should be a data frame, not %s',
                 typeof(x)))
  }
  if (!all(c('program_name') %in% names(x))){
    stop(sprintf('Batch dataset is required to have program_name column, but it has %s',
                 names(x)))}
  if (nrow(x)==0){
    stop(sprintf('Batch dataset should have at least 1 record'))
  }
  return(TRUE)
}


#' Validate data frame for calculating run groups
#' @param x A data frame
#' @return TRUE if all checks pass
#' @noRd
#' @examples
#' validate_batch_data_frame_for_run_group(
#' as.data.frame(list(program_name=c("program1.R", "program2.R", "program3.R"),
#'                    inputs=c('inp1', 'inp2', 'inp3'),
#'                    outputs=c('out1', 'out2', 'out3'))))
validate_batch_data_frame_for_run_group <- function(x){
  if (!is.data.frame(x)){
    warning(sprintf('Batch dataset should be a data frame, not %s',
                 typeof(x)))
    return(FALSE)
  }
  if (!all(c('program_name', 'inputs', 'outputs') %in% names(x))){
    warning(sprintf('Run groups check: Batch dataset is required to have program_name, inputs and outputs columns, but it has %s. Run groups will not be calculated.',
                 paste(names(x), collapse=',')))
    return(FALSE)}
  if (any(is.na(x$inputs)) || any(is.na(x$outputs))){
    warning('Run groups check: There is at least one NA value in inputs/outputs columns. Run groups will not be calculated.')
    return(FALSE)}
  if (nrow(x)==0){
    warning('Run groups check: Batch dataset should have at least 1 record. Run groups will not be calculated')
    return(FALSE)
  }
  return(TRUE)
}

#' Calculate run_group variable using inputs and outputs of programs supplied by user
#'
#' @param x input data frame. Must contain columns 'inputs', 'outputs' that list
#' input/output datasets for each program
#' @param col_name name of newly created variable. Defaults to 'run_group_calculated'
#'
#' @return an input data frame with one new column
#' @export
#'
#' @examples
#' input_ds <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R'),
#'                                inputs=c('ds0.xpt', 'ds1.xpt'),
#'                                outputs=c('ds1.xpt', 'ds2.xpt')))
#' batch_ready <- calculate_run_group
calculate_run_group <- function(x, col_name='run_group'){
  input_validated <- validate_batch_data_frame_for_run_group(x)

  # if input data frame does not contain all required information - assign run_group
  # so that all programs would run sequentially
  if (!input_validated){
    warning("Batch will be executed in sequential mode because run_groups could not be calculated.")
    x[[col_name]] <- c(1:nrow(x))
    return(x)
  }

  x[[col_name]] <- 0
  # determine which program to run first - such program inputs are not on the outputs of any other program
  first_progs_index <- get_first_programs(x)
  x[first_progs_index,][[col_name]] <- 1
  # iteratively calculate group run order
  current_group <- 1

  #setup cutoff to avoid infinite loop
  x_cutoff=length(x)
  cutoff_count <- 1
  while (!identical(x[[col_name]],
                    calculate_next_group(x,
                                         current_group=current_group,
                                         col_name=col_name)[[col_name]])){
    x <- calculate_next_group(x, current_group=current_group, col_name=col_name)
    current_group <- current_group + 1
    cutoff_count <- cutoff_count + 1

    if (cutoff_count > x_cutoff){
      warning("Could not calculate run groups fully. Please check input dataframe for circular dependencies")
      break}

  }
  return(x)
}

# split comma-separated inputs
parse_inputs <- function(x, ...){

  # if input data frame does not contain 'inputs' column, produce a warning,
  # create one and fill it with NAs
  if (is.null(x)){
    warning("Input data frame does not contain 'inputs' column. Dependency tracing would not be performed.")
    return(NULL)
  }

  # custom strsplit function that would not produce an error when trying to split NULL/NA
  strsplit_ <- function(x, ...){
    if (is.null(x) || is.na(x)){return(x)}
    return(unlist(strsplit(x, ...)))
  }
  split_x <- lapply(x, function(x) {strsplit_(x, ',')})
  return(lapply(split_x, stringr::str_trim))
}

# function to get indexes of programs whose inputs are not produced by any programs in the x dataset
get_first_programs <- function(ds){
  return(sapply(parse_inputs(ds$inputs), function(x) all(!(x %in% ds$outputs))))
}

calculate_next_group <- function(x,
                                 current_group=1,
                                 col_name='run_group'){

  cur_group_outputs <- x[x[[col_name]] == current_group,]$outputs
  # next group definition: any dataset that has one or more outputs of current group
  # as its inputs
  next_group <- sapply(parse_inputs(x$inputs), function(y) any(y %in% cur_group_outputs) && is.character(y))
  # if there is no next group - return unmodified dataset
  if (all(!next_group)) {return (x)}

  #
  x[next_group,][[col_name]] <- current_group + 1
  return(x)
}

test_that("dataframe_to_batch_list correctly converts dataframe to batch list", {
  ds <- as.data.frame(list(program_name=c("program1.R", "program2.R", "program3.R", "program4.R", "program5.R"),
                           run_group=c(1, 2, 2, 3, 3)))
  expected <- list(c("program1.R"),
                   c("program2.R", "program3.R"),
                   c("program4.R", "program5.R"))
  actual <- dataframe_to_batch_list(ds)

  expect_equal(actual, expected)
})

test_that("dataframe_to_batch_list returns a list back unchanged", {
  expected <- list(c("program1.R"),
                   c("program2.R", "program3.R"),
                   c("program4.R", "program5.R"))
  actual <- dataframe_to_batch_list(expected)

  expect_equal(actual, expected)
})

test_that("dataframe_to_batch_list returns a character vector back unchanged", {
  expected <- c("program1.R", "program2.R", "program3.R")
  actual <- dataframe_to_batch_list(expected)

  expect_equal(actual, expected)
})

test_that("calculate_run_group correctly calculates run groups", {
  inputs <- as.data.frame(list(program_name=c("program1.R", "program2.R", "program3.R", "program4.R", "program5.R"),
                               inputs=c("raw.DM", "sdtm.DM", "sdtm.DM", "sdtm.DM,sdtm.EX", "sdtm.DM,sdtm.EX"),
                               outputs=c("sdtm.DM", "sdtm.EX", "sdtm.AE", "sdtm.SE", "sdtm.SV")))
  expected <- c(1, 2, 2, 3, 3)
  actual <- calculate_run_group(inputs)$run_group
  expect_equal(expected, actual)
})

test_that("function errors when batch dataset is not a data frame", {
  expect_error(validate_batch_data_frame(1))
})

test_that("function errors when the input data frame does not have required columns", {
  expect_error(validate_batch_data_frame(as.data.frame(test=c(1, 2))))
})

test_that("function errors when input dataframe is empty", {
  expect_error(validate_batch_data_frame(as.data.frame(program_name=c(), run_group=c())))
})

# detect mode for batch control
detect_batch_mode_from_input <- function(x){
  if (is.data.frame(x)){return(c('data_frame'))}
  else if (is.list(x) || is.character(x)){return(c('list'))}
  else {stop(sprintf("Expecting a data frame, character vector or a list, not %s", typeof(x)))}
}

# return unique group numbers
get_run_groups <- function(x, col_name='run_group'){
  # data frame is expected to have run_group pre-populated/pre-calculated
  if (is.data.frame(x)){
    # check if col_name is in x data frame
    if (!(col_name %in% names(x))){
      stop(sprintf("%s variable is not in input data frame. Cannot run batch if input data frame does not have run groups.", col_name))
    }
    return(unique(x[[col_name]]))
  }
  # if input is a list/vector, then each element is a separate run group, therefore
  # length of list/vector is equal to amount of run groups
  else if (is.list(x) || is.character(x)){
    return(c(1:length(x)))
  }
}

# control batch execution. Function is meant to be called in a loop
batch_run_control <- function(x,
                              current_group=1,
                              ...){
  mode <- detect_batch_mode_from_input(x)
  # don't check anything if this is the first iteration of the loop
  if (current_group==1){return(list(prog_list=x, stop=FALSE))}

  if (mode == 'data_frame'){
    return(control_batch_flow_data_frame(x, current_group=current_group, ...))
  }
  else if (mode == 'list'){
    return(control_batch_flow_list(x, ...))
  }
}

# stop batch execution in 'list' input mode(when a list of programs is supplied to batch runner)
# meant to be used in a loop, otherwise 'break' will throw an error
control_batch_flow_list <- function(x,
                                    previous_run_programs=NULL,
                                    previous_run_ok=NULL,
                                    halt_on_error=TRUE,
                                    ...){

  result <- list(prog_list=x, stop=FALSE)

  if (!all(previous_run_ok)){
    if (halt_on_error){
      warning(sprintf("At least one of the following programs have errors in their logs:\n\t%s\nBatch execution halted.\n",
                      paste(previous_run_programs[!previous_run_ok], collapse='\n\t')))
      result$stop=TRUE
    }
    else {
      warning(sprintf("At least one of the following programs have errors in their logs:\n\t%s\n",
                      paste(previous_run_programs[!previous_run_ok], collapse='\n\t')))}
  }
  return(result)
}

# stop batch execution in 'data frame' input mode(when a data frame of programs with inputs/outputs is supplied to batch runner)
control_batch_flow_data_frame <- function(x,
                                          current_group=NULL,
                                          previous_run_programs=NULL,
                                          previous_run_ok=NULL,
                                          halt_on_error=TRUE,
                                          ...){

  result <- list(prog_list=x, stop=FALSE)

  if (!all(previous_run_ok)){
    x_filtered <- remove_failed_program_dependencies(
      x,
      failed_programs <- previous_run_programs[!previous_run_ok])

    result$prog_list <- x_filtered

    if (halt_on_error){
      warning(sprintf("At least one of the following programs have errors in their logs:\n\t%s\nPrograms that depend on failed programs will not be executed.\n",
                      paste(previous_run_programs[!previous_run_ok], collapse='\n\t')))
      result$stop=TRUE
    }
    else {
      warning(sprintf("At least one of the following programs have errors in their logs:\n\t%s\n",
                      paste(previous_run_programs[!previous_run_ok], collapse='\n\t')))
    }
  }
  return(result)
}

# filter list/dataset to select programs for submission
select_parallel_run <- function(x, run_group, col_name='run_group'){
  mode <- detect_batch_mode_from_input(x)
  # data frame is expected to have run_group pre-populated/pre-calculated
  if (mode=='data_frame'){
    return(x[x[[col_name]]==run_group,]$program_name)
  }
  # in case of list/character vector, run group is just element position
  else if (mode=='list'){
    return(unlist(x[[run_group]]))
  }
}

# remove programs that depend on failed programs from data frame
remove_failed_program_dependencies <- function(x,
                                               failed_programs,
                                               col_name='run_group'){

  x_filtered <- x
  failed_programs_group <- min(x[x$program_name %in% failed_programs, ][[col_name]]) + 1

  # if failures occurred in the last run group - return dataset without modifications
  if (!(failed_programs_group %in% x[[col_name]])){return(x)}

  all_groups <- get_run_groups(x_filtered, col_name=col_name)
  current_failed_programs <- failed_programs

  # iteratively find all dependents of failed programs
  for (rg in c(all_groups[failed_programs_group:length(all_groups)])){

    current_failed_programs <- calculate_next_dependencies(
      x_filtered,
      current_group=rg,
      failed_programs=current_failed_programs,
      col_name=col_name)

    failed_programs <- c(failed_programs, current_failed_programs)
  }

  # return dataset without dependents of failed programs
  return(x[!(x$program_name %in% failed_programs),])
}

#
calculate_next_dependencies <- function(x,
                                        current_group=NULL,
                                        failed_programs=NULL,
                                        col_name='run_group',
                                        ...){
  # only select programs from current group
  x_current <- x[x[[col_name]]==current_group,]
  # get outputs of all previously failed programs
  cur_group_outputs <- x[x[['program_name']] == failed_programs,]$outputs
  # next group definition: any dataset that has one or more outputs of current group
  # as its inputs
  dependents <- sapply(parse_inputs(x_current$inputs), function(y) any(y %in% cur_group_outputs))

  # if there is no next group - return empty vector
  if (all(!dependents)) {return(NULL)}

  # return program names of failed program dependents
  return(x_current[dependents,][['program_name']])

}


# remove programs from input if their inputs/program files have not changed
remove_unchanged_programs <- function(x,
                                      run_group=1,
                                      col_name='run_group',
                                      ...){
  mode <- detect_batch_mode_from_input(x)
  if (mode == 'data_frame'){
    return(remove_unchanged_programs_data_frame(x, run_group=run_group, col_name=col_name, ...))
  }
  else if (mode == 'list'){
    return(remove_unchanged_programs_list(x, run_group=run_group, ...))
  }
}


# recursively descend into input list and remove programs whose hash has not changed from
# the last time it was collected(usually since the last batch run)
remove_unchanged_programs_list <- function(x,
                                           run_group=1,
                                           ...){

  # function for recursively filtering out programs from nested list
  recursive_cache_match <- function(l){

    descend_ <-  function(x) {
      if(is.list(x)) {
        return(recursive_cache_match(x))
        }
      else if(is.character(x)) {
        return(x[!(sapply(x, function(x) cache_match(x, ...)))])
      }
      else {return(x)}
    }
    lapply(l, descend_)
  }

  x_filtered <- recursive_cache_match(x)

  # display a message about programs that were filtered out
  x_removed <- setdiff(unlist(x), unlist(x_filtered))
  if (length(x_removed) > 0){
    message(sprintf("Programs would not be re-run due to hash sum check:\n\t%s", paste(x_removed, collapse="\\n\\t")))
  }

  return(x_filtered)
}


remove_unchanged_programs_data_frame <- function(x,
                                                 col_name='run_group',
                                                 run_group=1, ...){

  # select current group
  x_filtered <- x[x[[col_name]] == run_group,]

  # prepare a data frame
  programs_and_inputs <- x_filtered[c('program_name', 'inputs')]
  programs_and_inputs$inputs <- parse_inputs(programs_and_inputs$inputs, ...)
  # check if program cache from previous batch run matches the current program
  programs_cache_match <- apply(programs_and_inputs, 1, function(x) lapply(unlist(x), function(x) cache_match(x, ...)))
  # remove NULL results as they are for non-existing files - we would not include
  # them when making a decision on whether to re-run this particular program
  programs_cache_match <- lapply(programs_cache_match, function(x) Filter(Negate(is.na), x))
  # only programs that have all inputs and program itself unchanged will be excluded from batch
  programs_cache_match <- sapply(programs_cache_match, function(x) all(unlist(x)))
  x_filtered2 <- x_filtered[!unlist(programs_cache_match),]

  # display a message about programs that were filtered out
  x_removed <- x_filtered[unlist(programs_cache_match),]$program_name
  if (length(x_removed) > 0){
    message(sprintf("Programs would not be re-run due to hash sum check:\n\t%s", paste(x_removed, collapse="\n\t")))
  }

  # update current run group
  x <- x[x[[col_name]] != run_group,]
  x <- rbind(x, x_filtered2)

  return(x)
}



# calculate and save program hashes for later use in batch running
update_program_hashes <- function(x, ...){
  mode <- detect_batch_mode_from_input(x)
  if (mode == 'data_frame'){
    update_program_hashes_data_frame(x, ...)
  }
  else if (mode == 'list'){
    update_program_hashes_list(x, ...)
  }
}


# function for calculating hashes for programs when input is a list
update_program_hashes_list <- function(x, ...){
  # update program hashes
  sapply(unlist(x), function(x) abba_save_file_cache(x, ...))
}


# function for calculating hashes for programs and their inputs when input is a data frame
update_program_hashes_data_frame <- function(x, ...){
  # update program hashes
  sapply(x$program_name, function(x) abba_save_file_cache(x, ...))
  # update hashes of program inputs
  sapply(unlist(parse_inputs(x$inputs)), function(x) abba_save_file_cache(x, ...))
}

library(mockery)

test_that("detect_batch_mode_from_input correctly detects type of input", {
  expect_equal(detect_batch_mode_from_input(list()), 'list')
  expect_equal(detect_batch_mode_from_input(c('1')), 'list')
  expect_equal(detect_batch_mode_from_input(as.data.frame(list())), 'data_frame')
})


test_that("detect_batch_mode_from_input errors when input is not character vector/list/data frame", {
  expect_error(detect_batch_mode_from_input(NULL))
})


test_that("get_run_groups returns unique values of run group variable", {
  df_input <- as.data.frame(list(run_group=c(1,1,1,1,2,2,3,3,3,3)))
  expect_equal(get_run_groups(df_input),  c(1,2,3))

  list_input <- list(c('prog1.R', 'prog2.R'),
                     c('prog3.R'),
                     c('Prog4.R', 'prog5.R'),
                     c('Prog6.R'))
  expect_equal(get_run_groups(list_input),  c(1,2,3,4))

  expect_equal(get_run_groups(c('prog1.R', 'prog2.R')),  c(1,2))
})


test_that("get_run_groups errors when input data frame does not contain col_name variable", {
  df_input <- as.data.frame(list(run_group=c(1,1,1,1,2,2,3,3,3,3)))
  expect_error(get_run_groups(df_input, col_name='rg'))
})


test_that("batch_run_control returns unmodified input when current_group equals to 1", {
  mock_control_flow_df <- mock("data_frame_flow")
  mock_control_flow_list <- mock("list_flow")
  stub(batch_run_control, "control_batch_flow_data_frame", mock_control_flow_df)
  stub(batch_run_control, "control_batch_flow_list", mock_control_flow_list)

  actual_result <- batch_run_control(as.data.frame(list()),
                                     current_group=1,
                                     arg3='arg3')
  expect_called(mock_control_flow_df, 0)
  expect_called(mock_control_flow_list, 0)

  expect_equal(actual_result, list(prog_list=as.data.frame(list()), stop=FALSE))
})


test_that("batch_run_control runs as expected when input is a data frame", {
  mock_control_flow_df <- mock("data_frame_flow")
  mock_control_flow_list <- mock("list_flow")
  stub(batch_run_control, "control_batch_flow_data_frame", mock_control_flow_df)
  stub(batch_run_control, "control_batch_flow_list", mock_control_flow_list)

  actual_result <- batch_run_control(as.data.frame(list()),
                                     current_group=2,
                                     arg3='arg3')
  expect_called(mock_control_flow_df, 1)
  expect_called(mock_control_flow_list, 0)

  expect_args(mock_control_flow_df, 1,
              as.data.frame(list()), current_group=2, arg3='arg3')
})


test_that("batch_run_control runs as expected when input is a list", {
  mock_control_flow_df <- mock("data_frame_flow")
  mock_control_flow_list <- mock("list_flow")
  stub(batch_run_control, "control_batch_flow_data_frame", mock_control_flow_df)
  stub(batch_run_control, "control_batch_flow_list", mock_control_flow_list)

  actual_result <- batch_run_control(list(),
                                     current_group=2,
                                     arg3='arg3')
  expect_called(mock_control_flow_df, 0)
  expect_called(mock_control_flow_list, 1)

  expect_args(mock_control_flow_list, 1,
              list(), arg3='arg3')
})


test_that("control_batch_flow_list does not halt execution if all programs from previous wave run without errors", {

  actual_result <- control_batch_flow_list(list(c('prog1.R', 'prog2.R')),
                                           previous_run_programs=c('prog1.R', 'prog2.R'),
                                           previous_run_ok=c(TRUE,TRUE))
  expected_result <- list(prog_list=list(c('prog1.R', 'prog2.R')), stop=FALSE)
  expect_equal(actual_result, expected_result)
})


test_that("control_batch_flow_list halts execution and produces warning if any programs from previous wave encountered errors during execution", {

  expect_warning(actual_result <- control_batch_flow_list(list(c('prog1.R', 'prog2.R')),
                                                          previous_run_programs=c('prog1.R', 'prog2.R'),
                                                          previous_run_ok=c(TRUE,FALSE)))
  expected_result <- list(prog_list=list(c('prog1.R', 'prog2.R')), stop=TRUE)
  expect_equal(actual_result, expected_result)
})


test_that("control_batch_flow_list does not halt execution but produces warning if halt_on_error is set to FALSE", {

  expect_warning(actual_result <- control_batch_flow_list(list(c('prog1.R', 'prog2.R')),
                                                          previous_run_programs=c('prog1.R', 'prog2.R'),
                                                          previous_run_ok=c(TRUE,FALSE),
                                                          halt_on_error=FALSE))
  expected_result <- list(prog_list=list(c('prog1.R', 'prog2.R')), stop=FALSE)
  expect_equal(actual_result, expected_result)
})


test_that("control_batch_flow_data_frame does not halt execution if all programs from previous wave ran without errors", {
  df <- as.data.frame(list(program=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R'),
                           inputs=c('raw.dm', 'sdtm.dm,raw.ae', 'sdtm.dm', 'sdtm.ae', 'sdtm.dm,sdtm.ae'),
                           outputs=c('sdtm.dm', 'sdtm.ae', 'tlf.dm', 'tlf.ae', 'tlf.sp'),
                           run_group=c(1,2,3,3,3)))

  actual_result <- control_batch_flow_data_frame(
    df,
    previous_run_programs=c('prog1.R', 'prog2.R'),
    previous_run_ok=c(TRUE,TRUE))

  expected_result <- list(prog_list=df, stop=FALSE)
  expect_equal(actual_result, expected_result)
})


test_that("control_batch_flow_data_frame halts execution and returns filtered data frame if any programs from previous wave ran without errors", {
  mock_remove_failed_program_dependencies <- mock("df_filtered")
  stub(control_batch_flow_data_frame,
       "remove_failed_program_dependencies",
       mock_remove_failed_program_dependencies)

  expect_warning(actual_result <- control_batch_flow_data_frame(
    df,
    previous_run_programs=c('prog1.R', 'prog2.R'),
    previous_run_ok=c(TRUE,FALSE)))

  expected_result <- list(prog_list="df_filtered", stop=TRUE)
  expect_equal(actual_result, expected_result)
})


test_that("control_batch_flow_data_frame does not halt execution and returns filtered data frame if any programs from previous wave ran without errors and halt_on_error is FALSE", {
  mock_remove_failed_program_dependencies <- mock("df_filtered")
  stub(control_batch_flow_data_frame,
       "remove_failed_program_dependencies",
       mock_remove_failed_program_dependencies)

  expect_warning(actual_result <- control_batch_flow_data_frame(
    df,
    previous_run_programs=c('prog1.R', 'prog2.R'),
    previous_run_ok=c(TRUE,FALSE),
    halt_on_error=FALSE))

  expected_result <- list(prog_list="df_filtered", stop=FALSE)
  expect_equal(actual_result, expected_result)
})


test_that("select_parallel_run correctly filters data frame", {
  df <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R'),
                           inputs=c('raw.dm', 'sdtm.dm,raw.ae', 'sdtm.dm', 'sdtm.ae', 'sdtm.dm,sdtm.ae'),
                           outputs=c('sdtm.dm', 'sdtm.ae', 'tlf.dm', 'tlf.ae', 'tlf.sp'),
                           run_group=c(1,2,3,3,3)))

  actual_result <- select_parallel_run (df, 2, col_name='run_group')

  expected_result <- df[2,]$program_name
  expect_equal(actual_result, expected_result)
})


test_that("select_parallel_run correctly filters list", {
  l_input <- list(c('prog1.R'), c('prog2.R'), c('prog3.R', 'prog4.R', 'prog5.R'))

  actual_result <- select_parallel_run (l_input, 3)

  expected_result <- unlist(l_input[[3]])
  expect_equal(actual_result, expected_result)
})


test_that("calculate_next_dependencies returns empty vector if the current group is the last one", {
  df <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R'),
                           inputs=c('raw.dm', 'sdtm.dm,raw.ae', 'sdtm.dm', 'sdtm.ae', 'sdtm.dm,sdtm.ae'),
                           outputs=c('sdtm.dm', 'sdtm.ae', 'tlf.dm', 'tlf.ae', 'tlf.sp'),
                           run_group=c(1,2,3,3,3)))

  result_actual <- calculate_next_dependencies(df, current_group=3,
                                               failed_programs=c('prog5.R'))
  expected_result <- NULL

  expect_equal(result_actual, expected_result)
})

test_that("calculate_next_dependencies properly returns programs that depend on failed programs", {
  df <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R'),
                           inputs=c('raw.dm', 'sdtm.dm,raw.ae', 'sdtm.dm', 'sdtm.ae', 'sdtm.dm,sdtm.ae'),
                           outputs=c('sdtm.dm', 'sdtm.ae', 'tlf.dm', 'tlf.ae', 'tlf.sp'),
                           run_group=c(1,2,3,3,3)))

  result_actual <- calculate_next_dependencies(df, current_group=3,
                                               failed_programs=c('prog2.R'))
  expected_result <- c('prog4.R', 'prog5.R')

  expect_equal(result_actual, expected_result)
})


test_that("remove_failed_program_dependencies properly filters input dataset from programs that depend on failed programs", {
  df <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R', 'prog6.R', 'prog7.R', 'prog8.R', 'prog9.R'),
                           inputs=c('raw.dm', 'raw.ds', 'raw.ae', 'sdtm.dm,sdtm.ds', 'sdtm.ae', 'adam.adsl,sdtm.ds', 'adam.adsl', 'adam.adds', 'adam.adae'),
                           outputs=c('sdtm.dm', 'sdtm.ds', 'sdtm.ae', 'adam.adsl', 'adam.adae', 'adam.adds', 'tlf.dm', 'tlf.ds', 'tlf.ae'),
                           run_group=c(1,1,1,2,2,3,4,4, 4)))

  result_actual <- remove_failed_program_dependencies (df, c('prog2.R'))
  expected_result <- df[c(1,3,5,7,9),]

  expect_equal(result_actual, expected_result)
})


test_that("remove_failed_program_dependencies returns dataset unchanged if failed programs are in the last run_group", {
  df <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R', 'prog3.R', 'prog4.R', 'prog5.R', 'prog6.R', 'prog7.R', 'prog8.R', 'prog9.R'),
                           inputs=c('raw.dm', 'raw.ds', 'raw.ae', 'sdtm.dm,sdtm.ds', 'sdtm.ae', 'adam.adsl,sdtm.ds', 'adam.adsl', 'adam.adds', 'adam.adae'),
                           outputs=c('sdtm.dm', 'sdtm.ds', 'sdtm.ae', 'adam.adsl', 'adam.adae', 'adam.adds', 'tlf.dm', 'tlf.ds', 'tlf.ae'),
                           run_group=c(1,1,1,2,2,3,4,4, 4)))

  result_actual <- remove_failed_program_dependencies (df, c('prog7.R'))
  expected_result <- df

  expect_equal(result_actual, expected_result)
})

#' Submit a job profile for execution on a Kubernetes cluster
#'
#' @param yaml_full_path Path to YAML config file
#'
#' @return A string containing Job ID
#' @export
#'
#' @examples \dontrun{
#' job_id <- submit_k8s_yaml("/tmp/my_job.yaml")
#' }
submit_k8s_yaml <- function(yaml_full_path){

  # send the job for execution
  suppressWarnings(system2(command="kubectl",
                           args=c("apply", "-f" , yaml_full_path),
                           stdout=TRUE, stderr=TRUE))
  # read job id from config and return it for further tracking and reporting
  yaml_config <- yaml::read_yaml(yaml_full_path)
  return(yaml_config$metadata$name)
}


#' Get status of all pods that belong to a job or batch
#'
#' @param unit_id unique identifier for a job/batch
#' @param namespace Kubernetes namespace
#' @param unit_type type of unit - can be 'job' or 'batch'
#' @return list of statuses for every pod in a job/batch.
#' @noRd
#'
abba_get_k8s_unit_status_local <- function(unit_id, unit_type='job', namespace=getOption('abba.k8s.namespace')){

  status_descriptions <- list(
    Pending = "The Pod has been accepted by the Kubernetes cluster, but one or more of the containers has not been set up and made ready to run. This includes time a Pod spends waiting to be scheduled as well as the time spent downloading container images over the network.",
    Running = "The Pod has been bound to a node, and all of the containers have been created. At least one container is still running, or is in the process of starting or restarting.",
    Succeeded = "All containers in the Pod have terminated in success, and will not be restarted.",
    Failed = "All containers in the Pod have terminated, and at least one container has terminated in failure. That is, the container either exited with non-zero status or was terminated by the system.",
    Unknown = "For some reason the state of the Pod could not be obtained. This phase typically occurs due to an error in communicating with the node where the Pod should be running."
  )
  job_details <- list()

  # determine selector depending on unit_type
  validate_unit_type(unit_type)
  if (unit_type=='job') {unit_selector <- "--selector=batch.kubernetes.io/job-name"}
  else if (unit_type=='batch') {unit_selector <- "-l batch-group"}
  # Get the status and args of all pods in the batch group
  cmd_args <- sprintf("get pods -n %s %s=%s -o=jsonpath='{range .items[*]}{.metadata.name}{\",\"}{.status.phase}{\",\"}{.spec.containers[].args}{\"\\n\"}{end}'",
                      namespace, unit_selector, shQuote(unit_id))
  pod_info <- suppressWarnings(system2("kubectl",
                                       args=cmd_args,
                                       stdout=TRUE, stderr=TRUE))
  pod_lines <- unlist(strsplit(pod_info, "\n"))

  # Reset job_details for each iteration
  job_details <- list()

  for (line in pod_lines) {
    if (line != "") {
      pod_name <- get_k8s_pod_name(line)
      pod_status <- get_k8s_pod_status(line)
      program_name <- get_k8s_pod_program_name(line)

      # Ensure the list for this status exists
      if (!is.list(job_details[[pod_status]])) {
        job_details[[pod_status]] <- list("Jobs" = list(), "Description" = status_descriptions[[pod_status]])
      }

      # Append the job details
      job_details[[pod_status]]$Jobs <-
        c(job_details[[pod_status]]$Jobs,
          list(list(id=unlist(pod_name),
                    path=unlist(program_name))))
    }
  }

  return(job_details)
}

#' Get status of all pods that belong to a job
#'
#' @param job_id unique identifier for a job
#' @param namespace Kubernetes namespace
#'
#' @return list of statuses for every pod in a job(typically just one).
#' @export
#'
#' @examples \dontrun{
#' status <- abba_get_k8s_job_status_local("job-sdtm-abc123")
#' }
abba_get_k8s_job_status_local <- function(job_id, namespace=getOption('abba.k8s.namespace')){

  job_details <- abba_get_k8s_unit_status_local(unit_id=job_id, unit_type='job', namespace=namespace)

  return(job_details)
}


#' Get status of all jobs in a batch
#'
#' @param batch_id unique identifier for a batch
#' @param namespace Kubernetes namespace
#'
#' @return list of statuses for every job in a batch
#' @export
#'
#' @examples \dontrun{
#' status <- abba_get_k8s_batch_status_local("batch-sdtm-abc123")
#' }
abba_get_k8s_batch_status_local <- function(batch_id, namespace=getOption('abba.k8s.namespace')){

  job_details <- abba_get_k8s_unit_status_local(unit_id=batch_id, unit_type='batch', namespace=namespace)

  return(job_details)
}


#' Watch a K8S job/batch that has been submitted to Workbench, periodically polling it's execution status.
#'
#' @param unit_id job/batch ID that was specified when submitting job/batch
#' @param unit_type specify whether to watch a job or a batch
#' @param poll_interval_seconds Time interval for polling batch status in seconds
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param namespace Kubernetes namespace
#'
#' @return a list of job ID(s) and status(es)
#' @noRd
#'
#' @examples \dontrun{
#' result <- abba_watch_k8s_unit_local("safety-tfls-f0bf6848-46de-45b8-9fae-0e732b104760", 10, 3000)
#' }
#'
abba_watch_k8s_unit_local <- function(unit_id='',
                                      unit_type='job',
                                      poll_interval_seconds = 3,
                                      timeout_seconds = 600,
                                      namespace=getOption('abba.k8s.namespace')){
  # Initialize variables for tracking job status
  start_time <- Sys.time()

  validate_unit_type(unit_type)
  if (unit_type=='job'){get_status <- abba_get_k8s_job_status_local}
  else if (unit_type=='batch'){get_status <- abba_get_k8s_batch_status_local}
  # Poll for job status in the specified batch group
  while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

    job_details <- get_status(unit_id, namespace=namespace)

    # Get the names of the outer list in job_details
    status_names <- names(job_details)

    # Check if neither "Pending" nor "Running" is a name in job_details
    if (!"Pending" %in% status_names && !"Running" %in% status_names) {
      break # Break if no "Pending" or "Running" in the names of job_details
    }

    # Wait for the specified interval before polling again
    Sys.sleep(poll_interval_seconds)
  }

  return(job_details)
}


#' Watch a K8S batch that has been submitted to Workbench, periodically polling it's execution status.
#'
#' @param batch_group_id Batch ID that was specified when submitting a group of jobs
#' @param poll_interval_seconds Time interval for polling batch status in seconds
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param namespace Kubernetes namespace
#'
#' @return a list of jobs IDs and statuses that belong to batch named batch_group_id
#' @export
#'
#' @examples \dontrun{
#' result <- abba_watch_k8s_batch_local("safety-tfls-f0bf6848-46de-45b8-9fae-0e732b104760", 10, 3000)
#' }
#'
abba_watch_k8s_batch_local <- function(batch_group_id='',
                                       poll_interval_seconds = 3,
                                       timeout_seconds = 600,
                                       namespace=getOption('abba.k8s.namespace')){

  job_details <- abba_watch_k8s_unit_local(unit_id=batch_group_id,
                                           unit_type='batch',
                                           poll_interval_seconds=poll_interval_seconds,
                                           timeout_seconds=timeout_seconds,
                                           namespace=namespace)

  return(job_details)
}


#' Watch a K8S job that has been submitted to Workbench, periodically polling it's execution status.
#'
#' @param job_id Job ID. Typically obtained as a return value from submit_job and similar functions
#' @param poll_interval_seconds Time interval for polling job status in seconds
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param namespace Kubernetes namespace
#'
#' @return a list of pods, their IDs and execution statuses
#' @export
#'
#' @examples \dontrun{
#' result <- abba_watch_k8s_job_local("job-sdtm-f0bf6848-46de-45b8-9fae-0e732b104760", 10, 3000)
#' }
#'
abba_watch_k8s_job_local <- function(job_id='',
                                     poll_interval_seconds = 3,
                                     timeout_seconds = 600,
                                     namespace=getOption('abba.k8s.namespace')){

  job_details <- abba_watch_k8s_unit_local(unit_id=job_id,
                                           unit_type='job',
                                           poll_interval_seconds=poll_interval_seconds,
                                           timeout_seconds=timeout_seconds,
                                           namespace=namespace)

  return(job_details)
}


#' Submit an R program for execution on a Kubernetes cluster
#'
#' @param file_path Full path to R file
#' @param batch_group_id Group ID for batch processing
#' @param user_tag Optional; a string that describes what kind of job will be scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container A valid container image name provided as a character string. Defaults to the option abba.default.container.
#' @param mounts Specifically formatted list with information bout volumes that container would have access to during the run
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_ip_address IP address for mounting service user home directory
#' @param namespace Kubernetes namespace to put the job in
#' @param username user whose authority will be used to run the program
#'
#' @return A list with job_id and batch_id attributes in case of successful submission
#' @export
#'
#' @examples \dontrun{
#' job_info <- abba_submit_k8s_job_local("path/to/your/program.R", batch_group_id='SDTM')
#' }
#'
abba_submit_k8s_job_local <- function(file_path,
                                      batch_group_id='',
                                      user_tag='',
                                      cpu_limit=1L,
                                      memory_limit='512M',
                                      container=getOption('abba.default.container'),
                                      mounts='',
                                      auto_mount_home=FALSE,
                                      home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
                                      namespace=getOption('abba.k8s.namespace'),
                                      username=NULL) {

  # Check if batch_group_id is a vector with more than one element
  validate_batch_id(batch_group_id)
  # By default let 'batch-group' be the lowest possible level - the program name.
  # Otherwise, keep what user has specified.
  if (is.null(batch_group_id) || batch_group_id == '') {
    batch_group_id <- unlist(strsplit(basename(file_path), '.', fixed = TRUE))[1]

    # Make the BATCH_GROUP_ID unique by appending the UUID to it.

    # Because the "group-name" doesn't need to be something unique, there is a 100% chance
    # to poll already completed jobs from past life. To prevent this we add UUID to a
    # "group-name".
    # A NOTE FOR THE GUESTS FROM THE FUTURE.
    # For multiple concurrent or parallel submissions that 'truly' may share same "group-name",
    # one should assign UUID outside of this function (at the batch level) and simply
    # pass the generated value as a parameter for this function.

    batch_group_id <- paste0(batch_group_id, '-', uuid::UUIDgenerate())
  }

  # Configure job to run our program
  job_config <- configure_k8s_yaml(file_path=file_path,
                                   batch_group_id=batch_group_id,
                                   user_tag=user_tag,
                                   cpu_limit= cpu_limit,
                                   memory_limit=memory_limit,
                                   container=container,
                                   mounts=mounts,
                                   namespace=namespace,
                                   username=username,
                                   auto_mount_home=auto_mount_home,
                                   home_nfs_ip_address=home_nfs_ip_address
                                   )

  # Save yaml to temp folders
  job_config_path <- save_yaml(job_config)

  # Send the job for execution and read job id
  job_id <- submit_k8s_yaml(job_config_path)

  return(list(job_id=job_id, batch_id=batch_group_id))

}

#' Submit a job profile for execution on a Kubernetes cluster and poll for completion
#'
#' @param file_path Full path to R file
#' @param batch_group_id Group ID for batch processing
#' @param user_tag Optional; a string that describes what kind of job will be scheduled to run
#' @param cpu_limit Maximum number of cores available for Kubernetes container
#' @param memory_limit Maximum amount of RAM available for Kubernetes container
#' @param container A valid container image name provided as a character string. Defaults to the option abba.default.container.
#' @param mounts Specifically formatted list with information bout volumes that container would have access to during the run
#' @param auto_mount_home set to TRUE to mount service user home directory
#' @param home_nfs_ip_address IP address for mounting service user home directory
#' @param namespace Kubernetes namespace to put the job in
#' @param username user whose authority will be used to run the program
#' @param poll_interval_seconds Time interval for polling job status in seconds
#' @param timeout_seconds Total time to wait before timeout in seconds
#'
#' @return A "Job completed successfully" message, or a list of failed jobs and their ID's.
#' @export
#'
#' @examples \dontrun{
#' result <- abba_submit_k8s_job_and_poll_local("path/to/your/job.yaml")
#' }
#'
#'
abba_submit_k8s_job_and_poll_local <- function(file_path,
                                               batch_group_id='',
                                               user_tag='',
                                               cpu_limit=1L,
                                               memory_limit='512M',
                                               container=getOption('abba.default.container'),
                                               mounts='',
                                               auto_mount_home=FALSE,
                                               home_nfs_ip_address=getOption('abba.home.nfs.ip.address'),
                                               namespace=getOption('abba.k8s.namespace'),
                                               username=NULL,
                                               poll_interval_seconds = 3,
                                               timeout_seconds = 600) {

  # Send the job for execution and read job id
  job_info <- abba_submit_k8s_job_local(file_path=file_path,
                                        batch_group_id=batch_group_id,
                                        user_tag=user_tag,
                                        cpu_limit=cpu_limit,
                                        memory_limit=memory_limit,
                                        container=container,
                                        mounts=mounts,
                                        namespace=namespace,
                                        username=username,
                                        auto_mount_home=auto_mount_home,
                                        home_nfs_ip_address=home_nfs_ip_address)

  result <- abba_watch_k8s_job_local(job_info$job_id,
                                     poll_interval_seconds = poll_interval_seconds,
                                     timeout_seconds = timeout_seconds,
                                     namespace=namespace)

  return(result)

}

library(mockery)

# Mock function to simulate 'system' calls to 'kubectl'
mock_system2 <- function(command, ...) {
  args = list(...)
  if (any(sapply(args$args, function(x){grepl('unique-job-id', x, fixed=TRUE)}))){
    return("job1,Succeeded,R -f /path/to/script1.R\njob2,Failed,R -f /path/to/script2.R\njob3,Running,R -f /path/to/script3.R\njob4,Pending,R -f /path/to/script4.R\njob5,Unknown,R -f /path/to/script5.R")
  }
  else if (any(sapply(args$args, function(x){grepl('unique-batch-id', x, fixed=TRUE)}))){
    # Simulate a scenario with various pod statuses including 'Unknown'
    return("pod1,Succeeded,-R -f /path/to/script1.R\npod2,Failed,R -f /path/to/script2.R\npod3,Running,R -f /path/to/script3.R\npod4,Pending,R -f /path/to/script4.R\npod5,Unknown,R -f /path/to/script5.R")
  }
}

# define some expected results
status_descriptions <- list(
  Pending = "The Pod has been accepted by the Kubernetes cluster, but one or more of the containers has not been set up and made ready to run. This includes time a Pod spends waiting to be scheduled as well as the time spent downloading container images over the network.",
  Running = "The Pod has been bound to a node, and all of the containers have been created. At least one container is still running, or is in the process of starting or restarting.",
  Succeeded = "All containers in the Pod have terminated in success, and will not be restarted.",
  Failed = "All containers in the Pod have terminated, and at least one container has terminated in failure. That is, the container either exited with non-zero status or was terminated by the system.",
  Unknown = "For some reason the state of the Pod could not be obtained. This phase typically occurs due to an error in communicating with the node where the Pod should be running."
)

expected_jobs <- list(Succeeded=list(Jobs=list(list(id='job1', path="/path/to/script1.R")), Description=status_descriptions$Succeeded),
                      Failed=list(Jobs=list(list(id='job2', path="/path/to/script2.R")), Description=status_descriptions$Failed),
                      Running=list(Jobs=list(list(id='job3', path="/path/to/script3.R")), Description=status_descriptions$Running),
                      Pending=list(Jobs=list(list(id='job4', path="/path/to/script4.R")), Description=status_descriptions$Pending),
                      Unknown=list(Jobs=list(list(id='job5', path="/path/to/script5.R")), Description=status_descriptions$Unknown))

expected_batch <- list(Succeeded=list(Jobs=list(list(id='pod1', path="/path/to/script1.R")), Description=status_descriptions$Succeeded),
                       Failed=list( Jobs=list(list(id='pod2', path="/path/to/script2.R")), Description=status_descriptions$Failed),
                       Running=list(Jobs=list(list(id='pod3', path="/path/to/script3.R")), Description=status_descriptions$Running),
                       Pending=list(Jobs=list(list(id='pod4', path="/path/to/script4.R")), Description=status_descriptions$Pending),
                       Unknown=list(Jobs=list(list(id='pod5', path="/path/to/script5.R")), Description=status_descriptions$Unknown))

test_that("abba_get_k8s_unit_status_local correctly returns information about jobs", {
  stub(abba_get_k8s_unit_status_local, "system2", mock_system2)
  actual <- abba_get_k8s_unit_status_local('unique-job-id', unit_type='job')
  expect_equal(actual, expected_jobs)
})

test_that("abba_get_k8s_unit_status_local correctly returns information about batch", {
  stub(abba_get_k8s_unit_status_local, "system2", mock_system2)
  actual <- abba_get_k8s_unit_status_local('unique-batch-id', unit_type='batch')

  expect_equal(actual, expected_batch)
})

test_that("abba_get_k8s_job_status_local correctly returns information about jobs", {
  stub(abba_get_k8s_job_status_local, "system2", mock_system2, depth=2)
  actual <- abba_get_k8s_job_status_local('unique-job-id')
  expect_equal(actual, expected_jobs)
})

test_that("abba_get_k8s_batch_status_local correctly returns information about batch", {
  stub(abba_get_k8s_batch_status_local, "system2", mock_system2, depth=2)
  actual <- abba_get_k8s_batch_status_local('unique-batch-id')

  expect_equal(actual, expected_batch)
})

library(mockery)

# Mock function to simulate 'system' calls to 'kubectl'
mock_system <- function(command, ...) {
  if (grepl("kubectl", command)) {
    if (grepl("some-group", command)) {
      # Simulate a scenario with various pod statuses including 'Unknown'
      return("pod1,Succeeded,R -f /path/to/script1.R\npod2,Failed,R -f /path/to/script2.R\npod3,Running,R -f /path/to/script3.R\npod4,Pending,R -f /path/to/script4.R\npod5,Unknown,R -f /path/to/script5.R")
    } else {
      # Default response for other groups (all succeeded)
      return("pod1,Succeeded,R -f /path/to/script1.R\npod2,Succeeded,R -f /path/to/script2.R")
    }
  }
}

# Unit test for watch_job function
test_that("get_batch_status returns statuses in expected format", {
  stub(abba_get_k8s_batch_status_local, "system2", mock_system, depth=2)

  # Test case for a batch group with various job statuses
  result <- abba_get_k8s_batch_status_local("batch-group")

  # Manually define expected result
  desc_success <- "All containers in the Pod have terminated in success, and will not be restarted."

  expected_result <- list(
    Succeeded=list(
      Jobs=list(list(id='pod1', path='/path/to/script1.R'), list(id='pod2', path='/path/to/script2.R')),
      "Description"=desc_success))

  expect_equal(result, expected_result)

})

test_that("abba_watch_k8s_batch_local functions correctly", {
  mock_batch_validation <- mock()
  mock_configure_k8s_yaml <- mock('configured_yaml_file')
  mock_save_yaml <- mock('yaml_file_path')
  mock_submit_k8s_yaml <- mock()

  stub(abba_submit_k8s_job_local, "validate_batch_id", mock_batch_validation)
  stub(abba_submit_k8s_job_local, "configure_k8s_yaml", mock_configure_k8s_yaml)
  stub(abba_submit_k8s_job_local, "save_yaml", mock_save_yaml)
  stub(abba_submit_k8s_job_local, "submit_k8s_yaml", mock_submit_k8s_yaml)

  result <- abba_submit_k8s_job_local('/path/to/program.R',
                                      batch_group_id='batch-id',
                                      user_tag='user_tag',
                                      cpu_limit=1,
                                      memory_limit='500M',
                                      container='unique_container',
                                      namespace='unique_namespace',
                                      username='unique_username',
                                      auto_mount_home=TRUE,
                                      home_nfs_ip_address='0.0.0.0')

  expect_called(mock_batch_validation, 1)
  expect_called(mock_configure_k8s_yaml, 1)
  expect_called(mock_save_yaml, 1)
  expect_called(mock_submit_k8s_yaml, 1)

  expect_args(mock_batch_validation, 1, 'batch-id')
  expect_args(mock_configure_k8s_yaml, 1,
              file_path='/path/to/program.R',
              batch_group_id='batch-id',
              user_tag='user_tag',
              cpu_limit= 1,
              memory_limit='500M',
              container='unique_container',
              mounts='',
              namespace='unique_namespace',
              username='unique_username',
              auto_mount_home=TRUE,
              home_nfs_ip_address='0.0.0.0')

  expect_args(mock_save_yaml, 1, 'configured_yaml_file')
  expect_args(mock_submit_k8s_yaml, 1, 'yaml_file_path')
})

test_that("abba_watch_k8s_batch_local assigns batch_group_id when it is omitted", {
  mock_batch_validation <- mock()
  mock_configure_k8s_yaml <- mock('configured_yaml_file')
  mock_save_yaml <- mock('yaml_file_path')
  mock_submit_k8s_yaml <- mock()

  stub(abba_submit_k8s_job_local, "validate_batch_id", mock_batch_validation)
  stub(abba_submit_k8s_job_local, "configure_k8s_yaml", mock_configure_k8s_yaml)
  stub(abba_submit_k8s_job_local, "save_yaml", mock_save_yaml)
  stub(abba_submit_k8s_job_local, "submit_k8s_yaml", mock_submit_k8s_yaml)

  result <- abba_submit_k8s_job_local('/path/to/program.R')

  args <- mock_args(mock_configure_k8s_yaml)
  expect_true(grepl('program', args[[1]]$batch_group_id, fixed=TRUE))

})

library(testthat)
library(mockery)

# Mock function for 'configure_yaml'
mock_configure_yaml <- function(...) {
  return(list())  # Return a dummy list representing the YAML configuration
}

# Mock function for 'save_yaml'
mock_save_yaml <- function(...) {
  return(tempfile())  # Return a dummy file path
}

# Mock function for 'submit_yaml'
mock_submit_yaml <- function(...) {
  return("dummy-job-id")  # Return a dummy job ID
}

# Mock function for 'watch_job'
mock_watch_job <- function(...) {
  return("Job completed successfully")  # Return a success message
}

# Mock function for 'system'
mock_system <- function(command, ...) {
  return(invisible(NULL))  # Simulate successful execution of system commands
}

test_that("submit_job_and_poll works correctly", {
  stub(abba_submit_k8s_job_and_poll_local, "configure_k8s_yaml", mock_configure_yaml, depth=2)
  stub(abba_submit_k8s_job_and_poll_local, "save_yaml", mock_save_yaml, depth=3)
  stub(abba_submit_k8s_job_and_poll_local, "abba_submit_k8s_job_local", mock_submit_yaml, depth=3)
  stub(abba_submit_k8s_job_and_poll_local, "abba_watch_k8s_job_local", mock_watch_job, depth=3)
  stub(abba_submit_k8s_job_and_poll_local, "system2", mock_system, depth=3)

  result <- abba_submit_k8s_job_and_poll_local("/path/to/file.R")
  expect_equal(result, "Job completed successfully")
})

library(mockery)

# define some expected results
status_descriptions <- list(
  Succeeded = "All containers in the Pod have terminated in success, and will not be restarted.",
  Failed = "All containers in the Pod have terminated, and at least one container has terminated in failure. That is, the container either exited with non-zero status or was terminated by the system.",
  Unknown = "For some reason the state of the Pod could not be obtained. This phase typically occurs due to an error in communicating with the node where the Pod should be running."
)

expected_jobs <- list(Unknown=list(Jobs=list(list(id='job1', path="/path/to/script1.R")), Description=status_descriptions$Unknown),
                      Failed=list(Jobs=list(list(id='job2', path="/path/to/script2.R")), Description=status_descriptions$Failed),
                      Succeeded=list(Jobs=list(list(id='job3', path="/path/to/script3.R")), Description=status_descriptions$Succeeded))

test_that("abba_watch_k8s_unit_local correctly waits and returns information about jobs", {
  mock_wait_job <- mock("job1,Pending,R -f /path/to/script1.R\njob2,Running,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R",
                        "job1,Running,R -f /path/to/script1.R\njob2,Failed,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R",
                        "job1,Unknown,R -f /path/to/script1.R\njob2,Failed,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R")

  stub(abba_watch_k8s_unit_local, "system2", mock_wait_job, depth=3)
  actual <- abba_watch_k8s_unit_local('unique-job-id', unit_type='job',
                                      poll_interval_seconds = 0.1)
  expect_equal(actual, expected_jobs)
})

test_that("abba_watch_k8s_unit_local correctly waits and returns information about batch", {
  mock_wait_batch <- mock("job1,Pending,R -f /path/to/script1.R\njob2,Running,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R",
                         "job1,Running,R -f /path/to/script1.R\njob2,Failed,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R",
                         "job1,Unknown,R -f /path/to/script1.R\njob2,Failed,R -f /path/to/script2.R\njob3,Succeeded,R -f /path/to/script3.R")

  stub(abba_watch_k8s_unit_local, "system2", mock_wait_batch, depth=3)
  actual <- abba_watch_k8s_unit_local('unique-batch-id', unit_type='batch',
                                      poll_interval_seconds = 0.1)

  expect_equal(actual, expected_jobs)
})

test_that("abba_watch_k8s_job_local correctly returns information about jobs", {
  mock_unit <- mock('result')
  stub(abba_watch_k8s_job_local, "abba_watch_k8s_unit_local", mock_unit)
  result <- abba_watch_k8s_job_local('unique-job-id')
  expect_called(mock_unit, 1)
  expect_args(mock_unit, 1,
              unit_id='unique-job-id',
              unit_type='job',
              poll_interval_seconds = 3,
              timeout_seconds = 600,
              namespace=getOption('abba.k8s.namespace'))
})


test_that("abba_watch_k8s_batch_local correctly returns information about jobs", {
  mock_unit <- mock('result')
  stub(abba_watch_k8s_batch_local, "abba_watch_k8s_unit_local", mock_unit)
  result <- abba_watch_k8s_batch_local('unique-batch-id')
  expect_called(mock_unit, 1)
  expect_args(mock_unit, 1,
              unit_id='unique-batch-id',
              unit_type='batch',
              poll_interval_seconds = 3,
              timeout_seconds = 600,
              namespace=getOption('abba.k8s.namespace'))
})

library(testthat)
library(mockery)

# Mock function for 'yaml::read_yaml'
mock_read_yaml <- function(file_path) {
  list(metadata = list(name = "test-job-id"))
}

test_that("submit_yaml returns the correct job ID", {
  stub(submit_k8s_yaml, "system2", mock())
  stub(submit_k8s_yaml, "yaml::read_yaml", mock_read_yaml)

  job_id <- submit_k8s_yaml("path/to/job.yaml")
  expect_equal(job_id, "test-job-id")
})

#' Submit R program as a SLURM job
#'
#' @param program_path Full path to the R program file. Must be accessible from the SLURM node
#' @param log_path Parent folder for the program's log file. Required; must
#' be supplied explicitly so that logs are never written to an unexpected
#' location in the user's filespace.
#' @param r_version Version of R that will be used to run the program. Can be specified as a full path to Rscript executable, or as a label of R version that is displayed in the Workbench GUI.
#' @param user_tag custom string that will be added to the job name.
#' @param cpu_cores Amount of CPU cores that will be requested for the job.
#' @param memory Amount of RAM in megabytes that will be requested for the job.
#' @param username user whose permission level is used to execute the script. Defaults to user submitting the job.
#' @param working_dir working directory for the SLURM job. Defaults to parent directory of `program_path`.
#' @param job_timeout time limit for a job. Must be specified in a format of "days-hours:minutes:seconds" If exceeded, job will be cancelled.
#' @param ... additional arguments (currently unused)
#'
#' @return job ID
#' @export
#'
#' @examples \dontrun{
#' job_id <- abba_slurm_submit_job("/home/user/tfl/t1_dm.sas",
#'                                 log_path = "/home/user/tfl/logs")
#'  }
abba_slurm_submit_job <- function(program_path,
                                  log_path,
                                  r_version=NULL,
                                  user_tag=NULL,
                                  cpu_cores=getOption("abba.slurm.cpu.cores"),
                                  memory=getOption("abba.slurm.memory"),
                                  username=NULL,
                                  working_dir=NULL,
                                  job_timeout=3600,
                                  ...) {

  if (missing(log_path) || is.null(log_path) || log_path == '') {
    stop("log_path must be supplied; abba does not write SLURM job logs to a default location.")
  }

  # default to current session version of R if not provided by user
  rscript_path <- select_rscript_version(r_version)

  # Submit the job for the program and wait until its execution
  scriptPath <- path.expand(program_path)
  scriptFile <- basename(scriptPath)
  jobTag <- c(paste("rstudio-r-script-job", scriptPath, sep = ":"), user_tag)

  # put log file in r script folder if no log path is supplied
  log_path <- slurm_config_determine_log_path(log_path = log_path, program_path = program_path)

  # create log directory if it does not exist
  if (!file.exists(dirname(log_path))){dir.create(dirname(log_path))}

  # assign working directory to parent dir of submitted program
  if (is.null(working_dir)){working_dir <- dirname(program_path)}

  # Configure job to run our program
  job_config <- configure_slurm_job(program_path=program_path,
                                    log_path=log_path,
                                    rscript_path=rscript_path,
                                    user_tag=user_tag,
                                    cpu_cores=cpu_cores,
                                    memory=memory,
                                    username=username,
                                    job_timeout=job_timeout
  )

  # Save yaml to temp folders
  job_config_path <- save_slurm_template(job_config)

  # Send the job for execution and read job id
  job_id <- submit_slurm_job_config(job_config_path)

  # return path of the executed script along with execution status(anything other than 0 is a failure)
  return(job_id)

}


#' Function to get SLURM job log
#'
#' @param job_id string representing a valid SLURM job ID
#' @param ... other arguments that will be ignored
#'
#' @return character vector containing job log
#' @noRd
#'
get_slurm_job_log0 <- function(job_id, ...){

  # try and get log path for a given job
  log_path <- get_slurm_job_log_path(job_id, ...)

  if (!file.exists(log_path)){
    stop(sprintf('Cannot find log for job %s. File %s does not exist.', job_id, log_path))
  }

  return(readLines(con=log_path))
}


#' Return job log for each submitted job ID
#'
#' @param job_ids valid SLURM job IDs
#' @param ... other parameters that will be ignored
#'
#' @return list of character vectors. each character vector would represent program log.
#' @export
#'
#' @examples \dontrun{
#' job_log <- abba_slurm_get_job_log("156")}
#'
abba_slurm_get_job_log <- function(job_ids, ...){
  return(lapply(job_ids, get_slurm_job_log0))
}


# return TRUE if job exit code is 0, and FALSE otherwise
#' Get SLURM job exit code represented by a boolean variable
#'
#' @param job_id a valid SLURM job ID
#' @param ... other parameters that will be ignored
#'
#' @return TRUE if job finished running and has 'COMPLETED' status, FALSE otherwise.
#' If job has not finished running, a NULL will be returned.
#' @noRd
#'
abba_slurm_get_job_succeeded0 <- function(job_id, ...){
  output <- suppressWarnings(system2(command="squeue",
                                     args=c('--jobs', job_id, '--Format="UserName,Name:.60,JobID:.10,exit_code:.14,State:.20"', "--states=all"),
                                     stdout=TRUE, stderr=TRUE))

  system_command_error_check(output, sprintf("Error getting job exit code for job ID %s.", job_id))
  parsed_output <- slurm_parse_squeue_output(output)
  parsed_output <- parsed_output[parsed_output$JOBID == job_id, , drop = FALSE]
  if(parsed_output$STATE == 'COMPLETED') {job_succeeded <- TRUE}
  else if(parsed_output$STATE %in% c('CONFIGURING', 'COMPLETING', 'PENDING', 'RUNNING', 'SIGNALING', 'RESIZING')) {job_succeeded <- NULL}
  else {job_succeeded <- FALSE}
  return(job_succeeded)
}


#' Check whether SLURM jobs have been fully executed.
#'
#' @param job_ids a list/vector of  valid SLURM job IDs
#' @param ... other parameters that will be ignored
#'
#' @return a named boolean vector. TRUE if job finished running and has 'COMPLETED' status, FALSE otherwise.
#' If job has not finished running, a NULL will be returned.
#' @export
#'
#' @examples \dontrun{
#' job_statuses <- abba_slurm_get_job_succeeded(c('job-id-1', 'job-id-2'))
#' }
abba_slurm_get_job_succeeded <- function(job_ids, ...){
  results <- sapply(job_ids, abba_slurm_get_job_succeeded0)
  if (any(sapply(results, function(x) is.null(x)))){
    stop(sprintf(
      "Jobs %s are still executing. Try increasing timeout parameter to avoid this error.",
      paste(job_ids[is.null(results)], collapse='\n\t')))
  }
  return(results)
}


#' Return descriptive job status for SLURM job
#'
#' @param job_id a valid Slurm job ID
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a named character vector with job status as value and job ID as name.
#' @noRd
slurm_get_job_status0 <- function(job_id, ...){
  output <- suppressWarnings(system2(command="squeue",
                                     args=c('--jobs', job_id, '--format="%.18i %.20P %.60j %.25u %.15T %.12M %.9l"', "--states=all"),
                                     stdout=TRUE, stderr=TRUE))

  system_command_error_check(output, sprintf("Error getting job status for job ID %s.", job_id))
  parsed_output <- slurm_parse_squeue_output(output)
  parsed_output <- parsed_output[parsed_output$JOBID %in% job_id, , drop = FALSE]

  return(parsed_output$STATE)
}


#' Return descriptive job status for slurm jobs
#'
#' @param job_ids a list/vector of Slurm job IDs
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a named character vector with job statuses as values and job IDs as names.
#' @export
#'
#' @examples \dontrun{
#' job_statuses <- slurm_get_job_status(c('job-id-1', 'job-id-2'))
#' }
abba_slurm_get_job_status <- function(job_ids, ...){
  result <- sapply(job_ids, slurm_get_job_status0)
  return(result)
}


#' Watch SLURM job, periodically polling its execution status.
#'
#' @param job_id job ID that was specified when submitting jobs
#' @param poll_interval_seconds Time interval for polling job status in seconds
#' @param timeout_seconds Total time to wait before timeout in seconds
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a list of job ID(s) and status(es)
#' @export
#'
#' @examples \dontrun{
#' result <- abba_slurm_watch_job("5195", 10, 3000)
#' }
#'
abba_slurm_watch_job <- function(job_id='',
                                 poll_interval_seconds = 3,
                                 timeout_seconds = 300,
                                 ...){
  # Initialize variables for tracking job status
  start_time <- Sys.time()

  # Poll for job status in the specified batch group
  while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

    job_details <- abba_slurm_get_job_status(job_id)

    job_still_running_statuses <- c("CONFIGURING", "RUNNING", "COMPLETING", "PENDING")
    # Check if all jobs finished running
    if (!any(sapply(job_details, function(x){x %in% job_still_running_statuses}))) {
      break # Break if no "Pending" or "Running" in the names of job_details
    }

    # Wait for the specified interval before polling again
    Sys.sleep(poll_interval_seconds)
  }

  return(job_details)
}

library(mockery)


test_that("abba_slurm_submit_job functions correctly", {
  mock_select_rscript_version <- mock('/path/to/bin/Rscript')
  mock_configure_slurm_job <- mock('configured_sbatch_file')
  mock_save_slurm_template <- mock('submission_script_file_path')
  mock_submit_slurm_job_config <- mock('job-id')
  mock_slurm_config_determine_log_path <- mock('/path/to/log/program.R')
  mock_dir_create <- mock()

  stub(abba_slurm_submit_job, "select_rscript_version", mock_select_rscript_version)
  stub(abba_slurm_submit_job, "configure_slurm_job", mock_configure_slurm_job)
  stub(abba_slurm_submit_job, "save_slurm_template", mock_save_slurm_template)
  stub(abba_slurm_submit_job, "submit_slurm_job_config", mock_submit_slurm_job_config)
  stub(abba_slurm_submit_job, "slurm_config_determine_log_path", mock_slurm_config_determine_log_path)
  stub(abba_slurm_submit_job, "dir.create", mock_dir_create)

  result <- abba_slurm_submit_job('/path/to/program.R',
                                  log_path='/path/to/log',
                                  r_version='4.2.2',
                                  user_tag='user_tag',
                                  cpu_cores=1,
                                  memory='500M',
                                  username='unique_username',
                                  job_timeout=100)

  expect_called(mock_select_rscript_version, 1)
  expect_called(mock_configure_slurm_job, 1)
  expect_called(mock_save_slurm_template, 1)
  expect_called(mock_submit_slurm_job_config, 1)

  expect_args(mock_select_rscript_version, 1, '4.2.2')
  expect_args(mock_configure_slurm_job, 1,
              program_path='/path/to/program.R',
              log_path='/path/to/log/program.R',
              rscript_path='/path/to/bin/Rscript',
              user_tag='user_tag',
              cpu_cores=1,
              memory='500M',
              username='unique_username',
              job_timeout=100)

  expect_args(mock_save_slurm_template, 1, 'configured_sbatch_file')
  expect_args(mock_submit_slurm_job_config, 1, 'submission_script_file_path')
})


test_that("Package .submit file for configuring slurm job loads correctly", {
  x <- load_slurm_template()

  expect_equal(length(x), 10L)
})


test_that("Placeholders are properly updated by configure_slurm_template function", {

  mock_system2 <- mock('uid=123(test_username) gid=456(domain users)', cycle=TRUE)

  stub(configure_slurm_job, "system2", mock_system2, depth=2)
  stub(configure_slurm_job, "uuid::UUIDgenerate", "uuid-generated")

  actual <- configure_slurm_job(program_path="/tst/path/test.R",
                                log_path='/tst/path/test.log',
                                rscript_path='/opt/R/421/bin/Rscript',
                                user_tag='user_tag',
                                username='test_username',
                                cpu_cores=1L,
                                memory=998,
                                job_timeout=100)

  expected <- c("#!/bin/bash",
                "#SBATCH --uid=123",
                "#SBATCH --time=100",
                "#SBATCH --cpus-per-task=1",
                "#SBATCH --mem=998",
                "#SBATCH --job-name=test-user_tag-uuid-generated",
                "#SBATCH --output=/tst/path/test.log",
                "#SBATCH --chdir=/tst/path",
                "",
                "/opt/R/421/bin/Rscript /tst/path/test.R")

  expect_equal(actual, expected)

})


test_that("save_slurm_template saves file when no file_path is specified", {
  config <- load_slurm_template()
  config_path <- save_slurm_template(config)
  expect_true(file.exists(config_path))
})

test_that("save_yaml saves file in a temporary directory when no file_path is specified", {
  config <- load_slurm_template()
  config_path <- save_slurm_template(config)
  expect_true(grepl(tempdir(), config_path, fixed=TRUE))
})


test_that("slurm_config_get_job_id correctly parses job id out of sbatch command output", {
  expect_equal(slurm_config_get_job_id('Submitted batch job 105573'), '105573')
})


test_that("submit_slurm_job_config submits script for execution", {

  mock_system2 <- mock("system2_output", cycle=TRUE)
  mock_get_job_id <- mock("job_id", cycle=TRUE)

  stub(submit_slurm_job_config, "system2", mock_system2)
  stub(submit_slurm_job_config, "slurm_config_get_job_id", mock_get_job_id)

  actual_result <- submit_slurm_job_config("/path/to/script.sh")

  expect_args(mock_system2, 1,
              command="sbatch",
              args=c("/path/to/script.sh"),
              stdout=TRUE, stderr=TRUE)

  expect_called(mock_get_job_id, 1)

  expect_equal(actual_result, "job_id")
})


test_that("abba_slurm_get_job_status gets job status", {

  mock_system2 <- mock(c("JOBID ExitCode STATE",
                         "2955  0   COMPLETED"), cycle=TRUE)

  stub(abba_slurm_get_job_status, "system2", mock_system2, depth=2)

  actual_result <- abba_slurm_get_job_status("2955")
  expected_result <- c("COMPLETED")
  names(expected_result) <- c("2955")

  expect_equal(actual_result, expected_result)
})


test_that("abba_slurm_get_job_succeeded returns TRUE if job had 0 exit code status and FALSE otherwise", {

  mock_system2 <- mock(c("JOBID EXIT_CODE STATE", "2955  0   COMPLETED", "2971  1   FAILED"), cycle = TRUE)

  stub(abba_slurm_get_job_succeeded, "system2", mock_system2, depth=2)

  actual_result <- abba_slurm_get_job_succeeded(c("2955", "2971"))
  expected_result <- c(TRUE, FALSE)
  names(expected_result) <- c("2955", "2971")

  expect_equal(actual_result, expected_result)
})



test_that("abba_slurm_watch_job waits for job to finish", {
  mock_get_job_status <- mock(c('CONFIGURING','COMPLETING'),
                              c('RUNNING','COMPLETED'),
                              c('CANCELLED','COMPLETED'), loop=TRUE)
  stub(abba_slurm_watch_job,
       "abba_slurm_get_job_status",
       mock_get_job_status)

  expected_result <- c('CANCELLED','COMPLETED')

  result_actual <- abba_slurm_watch_job(c('job-id6', 'job-id7'),
                                        poll_interval_seconds = 0.1,
                                        timeout_seconds = 10)

  expect_called(mock_get_job_status, 3)
  expect_args(mock_get_job_status, 3, c('job-id6', 'job-id7'))
  expect_equal(result_actual, expected_result)
})

#' Get IDs for all jobs marked with a given batch id
#'
#' @param group_batch_id batch ID tag that was used to mark submitted jobs
#' @param namespace Kubernetes namespace to search for jobs
#' @noRd
#' @return a character vector of job IDs
#'
get_k8s_job_ids_from_batch <- function(group_batch_id,
                                       namespace=getOption('abba.k8s.namespace')){

  # get list of job IDs with a given batch group
  output <- system2(command="kubectl",
                    args=c("get", "pods", "-n" , namespace, paste0("-l batch-group=", group_batch_id)),
                    stdout=TRUE, stderr=TRUE)

  # if there are no jobs with a given batch id, command will return a single string.
  # return an empty list in this case
  if (length(output)==1){return(list())}

  # remove header from output
  jobids <- utils::tail(output, -1)
  jobids_list <- sapply(jobids,
                        function(x) stringr::str_extract(x,  stringr::regex("^[\\d\\w\\-]+")),
                        USE.NAMES=FALSE)
  return(jobids_list)
}

test_that("get_batch_ids returns an empty list if no jobs exist with a given batch id", {
  mock_system <- mock(c("NAME READY STATUS RESTARTS AGE"))

  stub(get_k8s_job_ids_from_batch, "system2", mock_system)
  expect_equal(get_k8s_job_ids_from_batch('non-existing-batch-id-945387'), list())
})


test_that("get_batch_ids correctly returns job ids when given batch id", {
  mock_system <- mock(c("NAME READY STATUS RESTARTS AGE",
                        "job-id1 0/1 Completed 0 106s",
                        "job-id2 0/1 Completed 0 99s"))

  stub(get_k8s_job_ids_from_batch, "system2", mock_system)

  expect_equal(get_k8s_job_ids_from_batch('unique-batch-id'),
               c('job-id1', 'job-id2'))
})

#' Load the Slurm job template stored in package
#'
#' @return Imported bash script as a list of strings
#' @noRd
load_slurm_template <- function() {
  readLines(system.file("slurm_job.submit", package="abba"))
}

#' Save nested list representing the YAML file in specified location or in a
#' temporary directory.
#'
#' @param yaml_obj A nested list representing template YAML config
#' @param file_path Optional. Full filepath for to-be-saved YAML config
#'
#' @return Full path to the saved YAML file
#' @noRd
save_yaml <- function(yaml_obj, file_path=''){

  yaml_path <- if(file_path=='') tempfile(fileext = '.yaml') else file_path

  # a hack
  if (is.character(yaml_obj$spec$template$spec$containers[[1]]$command)){
    yaml_obj$spec$template$spec$containers[[1]]$command <-
      list(yaml_obj$spec$template$spec$containers[[1]]$command)
  }

  yaml::write_yaml(yaml_obj, yaml_path)

  return(yaml_path)
}

#' Create a Batch API or Job file template
#'
#' This function create a batch API or job file template at the specified
#' location. The Batch API file template is a plumber API with the necessary
#' REST API server side to interface with the \pkg{abba} package. This simplifies
#' the process of setting up the receiver API for which jobs are submitted. The
#' job template file is a markdown file with the necessary function calls to run
#' a batch job.
#'
#' Note that to deploy an R api to Posit Connect, the file must be named
#' plumber.R.
#'
#' @param path A file path where the target file will be created. Must be
#'   supplied explicitly; no default is provided so that files are never
#'   written to an unexpected location.
#'
#' @return No return value, called for side effects (file creation).
#' @rdname create_batch
#' @export
#'
#' @examples
#' # Write the template to a temporary directory
#' create_batch_api(tempdir())
#' create_batch_job(tempdir())
#'
#' \dontrun{
#' create_batch_api("~/api_directory")
#' create_batch_api("~/api_directory/plumber.R")
#'
#' create_batch_job("~/job_directory")
#' create_batch_job("~/job_directory/my_job.Rmd")
#' }
create_batch_api <- function(path) {
  create_file("plumber.R", path)
}

#' @rdname create_batch
#' @export
create_batch_job <- function(path) {
  create_file("job.Rmd", path)
}

#' Wrapper method for abba file creation
#'
#' @param src Package template to source
#' @param path Target file path
#' @noRd
create_file <- function(src = c("plumber.R", "job.Rmd"), path) {

  msg <- c("plumber.R" = "Batch API", "job.Rmd" = "Job")

  success <- file.copy(system.file(src, package="abba"),
                       path,
                       overwrite=FALSE)
  if(all(success)) {
    message(sprintf("%s file created at %s", msg[src], path.expand(path)))
  } else{
    warning(sprintf(
      "File could not be copied. Does the file %s already exist?", path
    ))
  }
  invisible()
}

library(mockery)

test_that("create_file copies plumber.R and job.Rmd files to target directory", {

  # setup testing area and files
  temp_dir <- tempdir()
  source_dir <- file.path(temp_dir, 'source')
  dir.create(source_dir, showWarnings = FALSE)
  plumber_file <- tempfile('plumber', tmpdir=source_dir, fileext='.R')
  job_file <- tempfile('job', tmpdir=source_dir, fileext='.Rmd')
  writeLines(text = "plumber API", con = plumber_file)
  writeLines(text = "markdown file", con = job_file)

  expect_plumber_file <- file.path(temp_dir, basename(plumber_file))
  expect_job_file <- file.path(temp_dir, basename(job_file))

  mock_system_file <- mock(c(plumber_file, job_file))
  stub(create_file, "system.file", mock_system_file)

  create_file(path=temp_dir)

  expect_true(file.exists(expect_plumber_file))
  expect_true(file.exists(expect_job_file))

})

test_that("create_file gives a warning when files already exist at destination", {

  # setup testing area and files
  temp_dir <- tempdir()
  source_dir <- file.path(temp_dir, 'source')
  dir.create(source_dir, showWarnings = FALSE)
  plumber_file <- tempfile('plumber', tmpdir=source_dir, fileext='.R')
  job_file <- tempfile('job', tmpdir=source_dir, fileext='.Rmd')
  writeLines(text = "plumber API", con = plumber_file)
  writeLines(text = "markdown file", con = job_file)

  expect_plumber_file <- file.path(temp_dir, basename(plumber_file))
  expect_job_file <- file.path(temp_dir, basename(job_file))

  mock_system_file <- mock(c(plumber_file, job_file), cycle = TRUE)
  stub(create_file, "system.file", mock_system_file)

  # create files for first time
  create_file(path=temp_dir)

  # a warning should appear about files not being copied
  expect_warning(create_file(path=temp_dir))
})

test_that("create_batch_api calls create_file function", {

  # setup mocks
  mock_create_file <- mock()
  stub(create_batch_api, "create_file", mock_create_file)

  # call the function
  create_batch_api(path='/unique/path')

  # test mock calls
  expect_called(mock_create_file, 1)
  expect_args(mock_create_file, 1, "plumber.R", path='/unique/path')
})

test_that("create_batch_job calls create_file function", {

  # setup mocks
  mock_create_batch_job <- mock()
  stub(create_batch_job, "create_file", mock_create_batch_job)

  # call the function
  create_batch_job(path='/unique/path')

  # test mock calls
  expect_called(mock_create_batch_job, 1)
  expect_args(mock_create_batch_job, 1, "job.Rmd", path='/unique/path')
})

rslauncher_submit_job <- function(p,
                                  execution_type='standard',
                                  log_path,
                                  user_tag=NULL,
                                  r_version=NULL,
                                  environment_vars=NULL,
                                  source_file=NULL,
                                  working_dir=NULL,
                                  ...) {

  if (missing(log_path) || is.null(log_path) || log_path == '') {
    stop("log_path must be supplied; abba does not write Workbench job logs to a default location.")
  }

  # default to current session version of R if not provided by user
  r_version <- select_r_version(r_version)

  # check execution type
  if (!(execution_type %in% c('standard', 'logrx'))){
    stop(sprintf("execution_type argument must be 'standard' or 'logrx', not %s", execution_type))
  }
  # fail if non-existing source_file was specified
  else if (!is.null(source_file) && !file.exists(source_file)){
    stop(sprintf("Source file was specified(%s) but does not exist.", source_file))
    }
  # Submit the job for the program and wait until its execution
  scriptPath <- path.expand(p)
  scriptFile <- basename(scriptPath)
  jobTag <- c(paste("rstudio-r-script-job", scriptPath, sep = ":"), user_tag)

  # interpret log_path as the target directory for the program's log file
  log_path <- file.path(log_path, paste0(tools::file_path_sans_ext(basename(scriptPath)), '.log'))

  # create log directory if it does not exist. supplying non-existing directory
  # to launcherSubmitJob would produce a silent error
  if (!file.exists(dirname(log_path))){dir.create(dirname(log_path), recursive = TRUE)}

  # assign working directory to parent dir of submitted program
  if (is.null(working_dir)){working_dir <- dirname(p)}

  # define scriptpath depending on the execution type
  if (execution_type == 'standard'){
    if (is.null(source_file)) {
      scriptArg <- paste("-f", scriptPath)
      }
    else {
      scriptArg <- sprintf("-f %s --args %s %s",
                           system.file('regular_workbench_submission.R', package="abba"),
                           scriptPath,
                           ifelse(is.null(source_file), "", source_file))
    }
  }
  else if (execution_type == 'logrx'){
    scriptArg <- sprintf("-f %s --args %s %s %s",
                         system.file('logrx_workbench_submission.R', package="abba"),
                         scriptPath,
                         log_path,
                         ifelse(is.null(source_file), "", source_file))
  }
  # submit program for execution via rstudioapi
  job_id <- rstudioapi::launcherSubmitJob(args =  c("--slave", "--no-save", "--no-restore", scriptArg),
                                          cluster = 'Local',
                                          exe = r_version,
                                          workingDirectory = working_dir,
                                          stdoutFile = log_path,
                                          stderrFile = log_path,
                                          name = scriptPath,
                                          tags = c(jobTag),
                                          environment=unlist(environment_vars)
  )
  # return path of the executed script along with execution status(anything other than 0 is a failure)
  return(job_id)

}

#' Create a job for executing an R program
#'
#' @param p path to program
#' @param log_path Path to the directory where the log will be saved.
#' Required; must be supplied explicitly so that logs are never written to an
#' unexpected location in the user's filespace.
#' @param user_tag optional; any user tags user might want to add to the job
#' @param ... other arguments that will be passed to internal rslauncher_submit_job function
#'
#' @return job id
#' @export
#'
#' @examples \dontrun{
#' job_id <- abba_rslauncher_submit_job_local("/path/to/program.R",
#'                                            log_path = "/path/to/logs")
#' }
abba_rslauncher_submit_job_local <- function(p,
                                             log_path,
                                             user_tag='',
                                             ...) {

  job_id <- rslauncher_submit_job(p,
                                  execution_type='standard',
                                  log_path=log_path,
                                  user_tag=user_tag,
                                  ...)

  # return path of the executed script along with execution status(anything other than 0 is a failure)
  return(job_id)

}


#' Execute programs via logrx
#'
#' @param p path to program
#' @param log_path Path to the directory where the log will be saved.
#' Required; must be supplied explicitly so that logs are never written to an
#' unexpected location in the user's filespace.
#' @param user_tag optional; any user tags user might want to add to the job
#' @param ... other arguments that will be passed to internal rslauncher_submit_job function
#'
#' @return job id
#' @export
#'
#' @examples \dontrun{
#' job_id <- abba_rslauncher_submit_logrx_job_local("/path/to/program.R",
#'                                                  log_path = "/path/to/logs")
#' }
abba_rslauncher_submit_logrx_job_local <- function(p,
                                                   log_path,
                                                   user_tag='',
                                                   ...) {

  rslauncher_submit_job(p,
                        execution_type='logrx',
                        log_path=log_path,
                        user_tag=user_tag,
                        ...)

}


# simple function to get the job status
#' Get Workbench job status for a given vector/list of job IDs
#'
#' @param job_ids A list/vector of job IDs
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a vector of job ID statuses
#' @export
#'
#' @examples \dontrun{
#' job_statuses <- abba_rslauncher_get_job_status_local(c('job-id-1', 'job-id-2'))
#' }
abba_rslauncher_get_job_status_local <- function(job_ids,
                                                 ...){
  return(sapply(job_ids, function(x) rstudioapi::launcherGetJob(x)[['status']]))
}


#' Periodically poll Workbench jobs for status and return their IDs when all job
#' statuses arrive at 'Finished' state
#'
#' @param job_ids a list/vector of Workbench job IDs
#' @param poll_interval_seconds how often job statuses should be updated
#' @param timeout_seconds maximum amount of time in seconds after which job IDs will be
#' returned regardless of job statuses
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a vector of job IDs
#' @export
#'
#' @examples \dontrun{
#' job_statuses <- abba_rslauncher_watch_job_local(c('job-id-1', 'job-id-2'))
#' }
abba_rslauncher_watch_job_local <- function(job_ids,
                                            poll_interval_seconds = 1,
                                            timeout_seconds = 300,
                                            ...){
  # Initialize variables for tracking job status
  start_time <- Sys.time()

  # Watch the job while it's executing
  while (difftime(Sys.time(), start_time, units = "secs") <= timeout_seconds) {

    statuses <- abba_rslauncher_get_job_status_local(job_ids)
    # Check if job has finished running
    if (all(statuses == "Finished")) {
      break # Break if job has reached 'Finished' status
    }
    # Wait for the specified interval before polling again
    Sys.sleep(poll_interval_seconds)
  }
  if (difftime(Sys.time(), start_time, units = "secs") > timeout_seconds) {
    warning(sprintf("Timeout limit exceeded for jobs:\n\t%s. \nJob IDs have been returned while jobs are still running.",
                    paste(job_ids[!(statuses == "Finished")], collapse='\n\t')))
  }
  return(job_ids)
}


# simple function to get job log
get_rslauncher_job_log0 <- function(job_id, ...){

  # default rstudio api error when non-existing job id is supplied is gibberish,
  # so it's better to display a simple error message to the user
  tryCatch({job_info <- rstudioapi::launcherGetJob(job_id)},
           error=function(e){stop(sprintf("Job with ID '%s' does not exist.", job_id))}
  )

  if (!file.exists(job_info$stdoutFile)){
    return(c(sprintf('Log file does not exist for %s', job_info$id)))
  }

  return(readLines(con=job_info$stdoutFile))
}


# vectorized version of get_workbench_job_log0
abba_rslauncher_get_job_log_local <- function(job_ids, ...){
  return(lapply(job_ids, get_rslauncher_job_log0))
}


# simple function to check whether job finished running without errors
# will return TRUE if exitCode is equal to 0, i.e. no errors occured during execution.
# does not check for warnings, only hard R errors
rslauncher_get_job_succeeded0 <- function(job_id, ...){
  # job_info <- rstudioapi::launcherGetJob(job_id)
  job_info <- tryCatch({rstudioapi::launcherGetJob(job_id)},
                      error=function(e){list(exitCode=NA)})

  if (is.null(job_info$exitCode)){return(NULL)}
  if (is.na(job_info$exitCode)){return(NA)}
  return(job_info$exitCode == 0)
}


#' Check whether Workbench jobs have been fully executed.
#'
#' @param job_ids a list/vector of Workbench job IDs
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a named boolean vector. FALSE value indicates that job did not fully execute
#' @export
#'
#' @examples \dontrun{
#' job_statuses <- abba_rslauncher_get_job_succeeded_local(c('job-id-1', 'job-id-2'))
#' }
abba_rslauncher_get_job_succeeded_local <- function(job_ids, ...){
  results <- sapply(job_ids, rslauncher_get_job_succeeded0)
  if (any(sapply(results, function(x) is.null(x)))){
    stop(sprintf(
      "Jobs %s are still executing. Try increasing timeout parameter to avoid this error.",
      paste(job_ids[is.null(results)], collapse='\n\t')))
  }
  return(results)
}


#' Return descriptive job status based on program exitCode
#'
#' @param job_ids a list/vector of Workbench job IDs
#' @param ... other positional/keyword arguments that will be ignored
#'
#' @return a named character vector. Possible statuses are 'Completed', 'Completed with errors'
#' @noRd
#'
#' @examples \dontrun{
#' job_statuses <- abba_rslauncher_get_job_display_status(c('job-id-1', 'job-id-2'))
#' }
rslauncher_get_job_display_status <- function(job_ids, ...){
  intermediate_results <- sapply(job_ids, rslauncher_get_job_succeeded0)
  if (any(sapply(intermediate_results, function(x) is.null(x)))){
    stop(sprintf(
      "Jobs %s are still executing. Try increasing timeout parameter to avoid this error.",
      paste(job_ids[is.null(intermediate_results)], collapse='\n\t')))
  }

  # function to convert TRUE/FALSE results into descriptive character statuses
  make_descriptive <- function(x){
    if (is.na(x)){'Not submitted'}
    else if(x==TRUE){'Completed'}
    else if(x==FALSE){'Completed with errors'}
  }

  results <- sapply(intermediate_results, make_descriptive)

  return(results)
}

library(mockery)

test_that("rslauncher_submit_job is producing an error when execution type is not 'logrx' or 'standard'", {

  expect_error(rslauncher_submit_job('/path/to/program.R',
                                     execution_type='special',
                                     log_path='/path/to'))
})

test_that("rslauncher_submit_job errors when log_path is not supplied", {
  expect_error(rslauncher_submit_job('/path/to/program.R'),
               "log_path must be supplied")
})

test_that("rslauncher_submit_job is working as expected in 'standard' mode", {
  mock_launcher_submit_job <- mock('job-id1')
  mock_select_r_version <- mock('/path/to/r/executable')
  mock_dir_create <- mock()

  stub(rslauncher_submit_job, "rstudioapi::launcherSubmitJob", mock_launcher_submit_job)
  stub(rslauncher_submit_job, "select_r_version", mock_select_r_version)
  stub(rslauncher_submit_job, "dir.create", mock_dir_create)

  result_actual <- rslauncher_submit_job('/path/to/program.R',
                                         log_path='/path/to')

  expect_called(mock_launcher_submit_job, 1)

  expect_args(mock_launcher_submit_job, 1,
              args =  c("--slave", "--no-save", "--no-restore", "-f /path/to/program.R"),
              cluster = 'Local',
              exe = "/path/to/r/executable",
              workingDir = "/path/to",
              stdoutFile = '/path/to/program.log',
              stderrFile = '/path/to/program.log',
              name = '/path/to/program.R',
              tags = c("rstudio-r-script-job:/path/to/program.R"),
              environment=NULL)

  expected_result <- c('job-id1')
  # names(expected_result) <- c('/path/to/program.R')

  expect_equal(result_actual, expected_result)
})

test_that("rslauncher_submit_job accepts log_path and user_tag arguments", {
  mock_launcher_submit_job <- mock('job-id1')
  mock_select_r_version <- mock('/path/to/r/executable')
  mock_dir_create <- mock()

  stub(rslauncher_submit_job, "rstudioapi::launcherSubmitJob", mock_launcher_submit_job)
  stub(rslauncher_submit_job, "select_r_version", mock_select_r_version)
  stub(rslauncher_submit_job, "dir.create", mock_dir_create)

  result_actual <- rslauncher_submit_job('/path/to/program.R',
                                         log_path='/path/to/log',
                                         user_tag='custom-user-tag')

  expect_called(mock_launcher_submit_job, 1)

  expect_args(mock_launcher_submit_job, 1,
              args =  c("--slave", "--no-save", "--no-restore", "-f /path/to/program.R"),
              cluster = 'Local',
              exe = "/path/to/r/executable",
              workingDir = "/path/to",
              stdoutFile = '/path/to/log/program.log',
              stderrFile = '/path/to/log/program.log',
              name = '/path/to/program.R',
              tags = c("rstudio-r-script-job:/path/to/program.R", "custom-user-tag"),
              environment=NULL)

  expected_result <- c('job-id1')
  # names(expected_result) <- c('/path/to/program.R')

  expect_equal(result_actual, expected_result)
})

test_that("rslauncher_submit_job is working as expected in 'logrx' mode", {
  mock_launcher_submit_job <- mock('job-id2')
  mock_select_r_version <- mock('/path/to/r/executable')
  mock_dir_create <- mock()

  stub(rslauncher_submit_job, "rstudioapi::launcherSubmitJob", mock_launcher_submit_job)
  stub(rslauncher_submit_job, "select_r_version", mock_select_r_version)
  stub(rslauncher_submit_job, "dir.create", mock_dir_create)

  result_actual <- rslauncher_submit_job('/path/to/program.R',
                                         execution_type='logrx',
                                         log_path='/path/to')
  expected_result <- c('job-id2')

  expect_called(mock_launcher_submit_job, 1)

  expect_args(mock_launcher_submit_job, 1,
              args =  c("--slave", "--no-save", "--no-restore",
                        sprintf("-f %s --args %s %s ",
                                system.file('logrx_workbench_submission.R', package="abba"),
                                '/path/to/program.R', '/path/to/program.log')),
              cluster = 'Local',
              exe = "/path/to/r/executable",
              workingDir = "/path/to",
              stdoutFile = '/path/to/program.log',
              stderrFile = '/path/to/program.log',
              name = '/path/to/program.R',
              tags = c("rstudio-r-script-job:/path/to/program.R"),
              environment=NULL)

  expect_equal(result_actual, expected_result)
})

test_that("abba_rslauncher_submit_job_local submits job in 'standard' mode and returns job id", {
  mock_rslauncher_submit_job <- mock('job-id3')
  stub(abba_rslauncher_submit_job_local, "rslauncher_submit_job", mock_rslauncher_submit_job)

  result_actual <- abba_rslauncher_submit_job_local('/path/to/program.R',
                                                    log_path='/path/to/log',
                                                    user_tag='user_tag',
                                                    arg4='arg4')

  expected_result <- c('job-id3')

  expect_called(mock_rslauncher_submit_job, 1)
  expect_args(mock_rslauncher_submit_job, 1,
              '/path/to/program.R',
              execution_type='standard',
              log_path='/path/to/log',
              user_tag='user_tag',
              arg4='arg4')
  expect_equal(result_actual, expected_result)
})

test_that("abba_rslauncher_submit_logrx_job_local submits job in 'logrx' mode and returns job id", {
  mock_rslauncher_submit_job <- mock('job-id4')
  stub(abba_rslauncher_submit_logrx_job_local,
       "rslauncher_submit_job",
       mock_rslauncher_submit_job)

  result_actual <- abba_rslauncher_submit_logrx_job_local('/path/to/program.R',
                                                          log_path='/path/to/log',
                                                          user_tag='user_tag',
                                                          arg4='arg4')
  expected_result <- c('job-id4')

  expect_called(mock_rslauncher_submit_job, 1)
  expect_args(mock_rslauncher_submit_job, 1,
              '/path/to/program.R',
              execution_type='logrx',
              log_path='/path/to/log',
              user_tag='user_tag',
              arg4='arg4')
  expect_equal(result_actual, expected_result)
})

test_that("abba_rslauncher_get_job_status_local returns job status", {
  mock_launcherGetJob <- mock(list(status='job-status-of-job-id5'))
  stub(abba_rslauncher_get_job_status_local,
       "rstudioapi::launcherGetJob",
       mock_launcherGetJob)

  result_actual <- abba_rslauncher_get_job_status_local('job-id5')

  expected_result <- c('job-status-of-job-id5')
  names(expected_result) <- c('job-id5')

  expect_called(mock_launcherGetJob, 1)
  expect_args(mock_launcherGetJob, 1, 'job-id5')
  expect_equal(result_actual, expected_result)
})

test_that("abba_rslauncher_watch_job_local waits for job to finish", {
  mock_rslauncher_get_job_status <- mock(c('Started','In Progress'),
                                         c('In Progress','Finished'),
                                         c('Finished','Finished'), loop=TRUE)
  stub(abba_rslauncher_watch_job_local,
       "abba_rslauncher_get_job_status_local",
       mock_rslauncher_get_job_status)

  result_actual <- abba_rslauncher_watch_job_local(c('job-id6', 'job-id7'),
                                                   poll_interval_seconds = 0.1,
                                                   timeout_seconds = 10)
  expected_result <- c('job-id6', 'job-id7')

  expect_called(mock_rslauncher_get_job_status, 3)
  expect_args(mock_rslauncher_get_job_status, 3, c('job-id6', 'job-id7'))
  expect_equal(result_actual, expected_result)
})

test_that("abba_rslauncher_watch_job_local produces warning when timeout is exceeded", {
  mock_rslauncher_get_job_status <- mock(c('Started','In Progress'),
                                         c('In Progress','Finished'),
                                         c('Finished','Finished'), loop=TRUE)
  stub(abba_rslauncher_watch_job_local,
       "abba_rslauncher_get_job_status_local",
       mock_rslauncher_get_job_status)

  expect_warning(abba_rslauncher_watch_job_local(c('job-id6', 'job-id7'),
                                                 poll_interval_seconds = 0.1,
                                                 timeout_seconds = 0.1))
})


test_that("get_rslauncher_job_log0 returns job log", {
  mock_launcher_get_job <- mock(list(stdoutFile='/path/to/log/program.log'))
  mock_read_lines <- mock(c('log line 1', 'log line 2'))
  stub(get_rslauncher_job_log0, "rstudioapi::launcherGetJob",
       mock_launcher_get_job)
  stub(get_rslauncher_job_log0, "base::readLines", mock_read_lines)

  result_actual <- get_rslauncher_job_log0(c('job-id8'))
  expected_result <- c('log line 1', 'log line 2')

  expect_called(mock_launcher_get_job, 1)
  expect_args(mock_launcher_get_job, 1, c('job-id8'))
  # cannot stub readLines properly, so leaving this sealed for now
  # expect_called(mock_read_lines, 1)
  # expect_args(mock_read_lines, 1, con='/path/to/log/program.log')
  # expect_equal(result_actual, expected_result)
})


test_that("get_rslauncher_job_log0 throws an error if job does not exist", {
  expect_error(get_rslauncher_job_log0(c('nonexst-job-id')))
})


test_that("abba_rslauncher_get_job_log_local returns job logs", {
  mock_get_rslauncher_job_log0 <- mock(c('log line 1', 'log line 2'),
                                       c('log line 3', 'log line 4'))
  stub(abba_rslauncher_get_job_log_local, "get_rslauncher_job_log0",
       mock_get_rslauncher_job_log0)

  result_actual <- abba_rslauncher_get_job_log_local(c('job-id9', 'job-id10'))
  expected_result <- c('log line 1', 'log line 2')

  expect_called(mock_get_rslauncher_job_log0, 2)
  expect_args(mock_get_rslauncher_job_log0, 1, 'job-id9')
  expect_args(mock_get_rslauncher_job_log0, 2, 'job-id10')
  expect_equal(result_actual, list(c('log line 1', 'log line 2'),
                                   c('log line 3', 'log line 4')))
})


test_that("rslauncher_get_job_succeeded0 returns TRUE when job exit code is zero", {
  mock_launcherGetJob <- mock(list(exitCode=0))
  stub(rslauncher_get_job_succeeded0, "rstudioapi::launcherGetJob",
       mock_launcherGetJob)

  result_actual <- rslauncher_get_job_succeeded0(c('job-id11'))
  expected_result <- c(TRUE)

  expect_called(mock_launcherGetJob, 1)
  expect_args(mock_launcherGetJob, 1, 'job-id11')
  expect_equal(result_actual, TRUE)
})


test_that("rslauncher_get_job_succeeded0 returns FALSE when job exit code is non-zero", {
  mock_launcherGetJob <- mock(list(exitCode=1))
  stub(rslauncher_get_job_succeeded0, "rstudioapi::launcherGetJob",
       mock_launcherGetJob)

  result_actual <- rslauncher_get_job_succeeded0(c('job-id12'))
  expected_result <- c(FALSE)

  expect_called(mock_launcherGetJob, 1)
  expect_args(mock_launcherGetJob, 1, 'job-id12')
  expect_equal(result_actual, expected_result)
})


test_that("rslauncher_get_job_succeeded0 returns NULL when job is still executing", {
  mock_launcherGetJob <- mock(list())
  stub(rslauncher_get_job_succeeded0, "rstudioapi::launcherGetJob",
       mock_launcherGetJob)

  result_actual <- rslauncher_get_job_succeeded0(c('job-id13'))
  expected_result <- NULL

  expect_called(mock_launcherGetJob, 1)
  expect_args(mock_launcherGetJob, 1, 'job-id13')
  expect_equal(result_actual, expected_result)
})


test_that("abba_rslauncher_get_job_succeeded_local returns job statuses according to their exit codes", {
  mock_rslauncher_get_job_succeeded0 <- mock(TRUE, FALSE)
  stub(abba_rslauncher_get_job_succeeded_local, "rslauncher_get_job_succeeded0",
       mock_rslauncher_get_job_succeeded0)

  result_actual <- abba_rslauncher_get_job_succeeded_local(c('job-id14', 'job-id15'))
  expected_result <- c(TRUE, FALSE)
  names(expected_result) <- c('job-id14', 'job-id15')
  expect_called(mock_rslauncher_get_job_succeeded0, 2)
  expect_args(mock_rslauncher_get_job_succeeded0, 1, 'job-id14')
  expect_args(mock_rslauncher_get_job_succeeded0, 2, 'job-id15')
  expect_equal(result_actual, expected_result)
})


test_that("abba_rslauncher_get_job_succeeded_local errors when at least one job is still running", {
  mock_rslauncher_get_job_succeeded0 <- mock(TRUE, FALSE, NULL)
  stub(abba_rslauncher_get_job_succeeded_local, "rslauncher_get_job_succeeded0",
       mock_rslauncher_get_job_succeeded0)

  expect_error(abba_rslauncher_get_job_succeeded_local(c('job-id16', 'job-id17', 'job-id18')))
})


test_that("rslauncher_get_job_display_status returns job statuses according to their exit codes", {
  mock_rslauncher_get_job_succeeded0 <- mock(TRUE, FALSE)
  stub(rslauncher_get_job_display_status, "rslauncher_get_job_succeeded0",
       mock_rslauncher_get_job_succeeded0)

  result_actual <- rslauncher_get_job_display_status(c('job-id16', 'job-id17'))
  expected_result <- c('Completed', 'Completed with errors')
  names(expected_result) <- c('job-id16', 'job-id17')
  expect_called(mock_rslauncher_get_job_succeeded0, 2)
  expect_args(mock_rslauncher_get_job_succeeded0, 1, 'job-id16')
  expect_args(mock_rslauncher_get_job_succeeded0, 2, 'job-id17')
  expect_equal(result_actual, expected_result)
})


test_that("rslauncher_get_job_display_status errors when at least one job is still running", {
  mock_rslauncher_get_job_succeeded0 <- mock(TRUE, FALSE, NULL)
  stub(rslauncher_get_job_display_status, "rslauncher_get_job_succeeded0",
       mock_rslauncher_get_job_succeeded0)

  expect_error(rslauncher_get_job_display_status(c('job-id16', 'job-id17', 'job-id18')))
})

#' Insert Connect API key into request header
#'
#' @param req request object
#' @param api_key Posit Connect API key
#'
#' @return if api_key is supplied - a modified req object with api key added to the request header
#'         if api_key is not supplied - an original, unmodified req object
#' @noRd
#' @examples
#' req <- httr2::request('base_url') %>% add_api_key_to_header(api_key=Sys.getenv("CONNECT_API_KEY"))
#'
add_api_key_to_header <- function(req, api_key=NULL){
  # throw an error at the user if api_key is not one of expected types
  if (!is.null(api_key) && !is.character(api_key)){
    stop(sprintf("api_key parameter must be a NULL or a character, not %s", typeof(api_key)))}
  # update request header if api key is supplied
  if(is.character(api_key) && nchar(api_key) > 0){
    req <- httr2::req_headers(req, Authorization = paste0("Key ", api_key))
  }
  # return the modified/unmodified request
  return(req)
}

test_that("Request is not modified if api_key is NULL", {
  req <- httr2::request("base_url") %>% add_api_key_to_header(api_key=NULL)
  actual <- list()
  expect_equal(req$headers, actual)
})

test_that("Request is not modified if api_key is an empty string", {
  req <- httr2::request("base_url") %>% add_api_key_to_header(api_key='')
  actual <- list()
  expect_equal(req$headers, actual)
})

test_that("function errors when api_key is not NULL or character", {
  expect_error(httr2::request("base_url") %>%
                 add_api_key_to_header(api_key=list(api_key='asdk')))
})

test_that("Function adds proper header when api_key is supplied", {
  req <- httr2::request("base_url") %>% add_api_key_to_header(api_key='test_key')
  expect_true("Authorization" %in% names(req$headers))
  })

---
title: "abba Options"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{abba Options}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

*abba* has several configurable options to control the scope of what users are permitted to do in their job submission. These options are applied **on the API side**, and should be considered by the individual setting up the API deployment. 

Note that the CPU and memory thresholds are exceptionally important considerations for Kubernetes. Kubernetes will accept any specification that means the schema of its API, and its on the ownership of the job submitter to consider if the submitted job can physically be executed within their specific cluster. If you specify a job with 512 bytes of RAM, the job will try to run and fail because the container cannot successfully launch in the first place. If you submit a job requesting 5000 cores, Kubernetes will continue waiting until is has an available node with 5000 cores. We say this from experience, so save yourself the trouble and make sure these options are configured appropriately! 

| Option                    | Default | Description                                                                                                                                                                                                |
|---------------------------|---------|------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| abba.lower.cpu.limit      | 0.001   | The lower threshold of CPU allowed during a job submission                                                                                                                                                 |
| abba.cpu.limit            | 2       | The upper threshold of CPU allowed during a job submission                                                                                                                                                 |
| abba.lower.memory.limit   | '128M'  | The lower threshold of memory allowed during a job submission                                                                                                                                              |
| abba.memory.limit         | '1G'    | The upper threshold of memory allowed during a job submission                                                                                                                                              |
| abba.permitted.containers | NULL    | A character vector containing the list of containers permitted for job executing. Use this option to prevent users from using unauthorized container environments. When NULL, no restrictions are applied. |
| abba.default.container    | NULL    | The default container used for job submission. When the `container` parameter is left empty, this container will be used.                                                                                  |
| abba.r.versions.path      | '/etc/rstudio/r-versions' | Path to the Posit Workbench R versions configuration file. Override this if your Workbench installation uses a non-standard location.                                              |

------
title: "{abba} Job Template"
author: "`r Sys.info()['user']`"
date: "`r Sys.Date()`"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r test}
abba::abba_submit_and_get_log(
  "/path/to/file.R",
  container = "<container_image_name>"
  )
```

---
title: "SLURM Job Submission"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{SLURM Job Submission}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(abba)
```

In support for SLURM, **abba** has function to schedule and manage R programs on connected SLURM cluster. At a high level, these functions break down into three different categories:

-   Submit a job
-   Get job status
-   Wait for job completion
-   Retrieve the log of the job

Functions specifically for interfacing with SLURM follow the naming convention `abba_slurm_*`.

## Submit a Job

Job submission is handled by the function `abba_slurm_submit_job()`.

```{r submit_job, eval=FALSE}
abba_slurm_submit_job(
  "/home/mike.stackhouse/repos/abba/test_programs/test_program.R",
  log_path = "/home/mike.stackhouse/test_logs"
  )
# [1] "38"
```

Note that this returns a vector with job id and name being path of the program. The Job ID is the identifier used downstream to interact with SLURM.

`log_path` is required and specifies the directory where the program's log file will be written. abba does not fall back to a default location in the user's filespace.

## Get Job Status

Once the job is running in SLURM, **abba** can poll its status. This can be done using the function `abba_slurm_get_job_status()`

```{r get_status, eval=FALSE}
abba_slurm_get_job_status("57")
# 57 
# "COMPLETED" 
```

The job statuses can be "RUNNING", "COMPLETED", FAILED etc. For a complete list of SLURM job statuses, please refer to [**documentation**](https://slurm.schedmd.com/squeue.html#SECTION_JOB-STATE-CODES).

# Get the Log Content

In this context, the "log" refers to the stdout/stderr of the program itself. This returns into a list object with the Job ID and the individual lines written out by the program. The content will contain all of the console output from the program itself.

```{r watch_job, eval=FALSE}
abba_slurm_get_job_log("38")
# [[1]]
# [1] "[1] 4"                   "Warning message:"        "This is a test warning " "La-di-da"                       
```

# Submit a batch of jobs to SLURM cluster

Here is an example of using SLURM submission functions in abba_submit_batch. Additional arguments like log_path will be passed on to submit_func

```{r submit batch, eval=FALSE}
abba_submit_batch(list("/home/mike.stackhouse/repos/abba/test_programs/test_program.R", "/home/mike.stackhouse/repos/abba/test_programs/test_program.R"),
                  submit_func=abba_slurm_submit_job,
                  wait_func=abba_slurm_watch_job,
                  succeed_func=abba_slurm_get_job_succeeded,
                  log_path="/home/mike.stackhouse/test_logs")
# Submitting program:
# 	/home/mike.stackhouse/repos/abba/test_programs/test_program.R
# Submitting program:
# 	/home/mike.stackhouse/repos/abba/test_programs/test_program.R
# /home/mike.stackhouse/repos/abba/test_programs/test_program.R /home/mike.stackhouse/repos/abba/test_programs/test_program.R
                                                          "48"                                                           "49"                    
```

---
title: "Kubernetes Batch Jobs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Kubernetes Batch Jobs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(abba)
```

If you're an end user simply trying to run your work using **abba**, then this vignette is for you. 

Our ultimate goal with **abba** is to make it as simple as possible for a user to user to do what they need to do. In doing so, there are a handful of methods to interact with the **abba** API. These methods do the work of submitting a job, waiting for it to complete, and collecting the log output. 

# End User Facing Functions

The **abba** end user methods all follow the naming convention `abba_`.

| **Function**                | **Description**                                                               |
|-----------------------------|-------------------------------------------------------------------------------|
| `abba_submit_job()`         | Submit a job to the **abba** API                                              |
| `abba_submit_and_get_log()` | Submit a job to the **abba** API, wait for completion, and return the job log |
| `abba_wait_for_job_log()`   | Continue polling until the job log is available                               |
| `abba_get_job_log()`        | Retrieve a job log                                                            |
| `abba_get_job_status()`     | Retrieve a job status                                                         |

As a user, this gives you some distinct control over your own workflow - whether you want to submit the job and handle it later with some of the other specific methods, or simply submit the job and hold the console until the job completes. The internals of each of these methods handle interaction with the API to ensure that HTTP requests don't time out. **Note the the methods do have their own configurable timeout limit.**. 

# Create a Scheduled Batch File

A key reason that we created **abba** was to allow users to take advantage of submitting jobs through Connect. A key benefit of this is Connect's ability to run reports on a set schedule, be it nightly or at whatever interval you need. **abba** leverages this capability and provides the extra component of sending your work out to a designated environment. We do this by using R Markdown reports, for which Connect has that scheduling capability. 

Batch files can be created using the function `create_batch_job()`. 

```{r create_batch_job, eval=FALSE}
create_batch_job("path/to/job_directory")
# Job file created at path/to/job_directory/job.Rmd
```

This creates a template R Markdown file with an example call to **abba** ready to go. 

```{r example_call, eval=FALSE}
abba::abba_submit_and_get_log(
  "/path/to/file.R",
  container = "<container_image_name>"
  )
```

There are some considerations for you should make when in your job submission:

- The file path must be visible to you as a user. 
- Make sure the job has enough resources to execute by using the `cpu_limit` and `memory_limit` arguments. Memory limits should be specified in Megabyte, so for example, a 4GB session would be listed as `'4000M'`.
- The timeout limits defaults to 10 minutes. If your job will likely run longer than 10 minutes, increase this using the `timeout_seconds` argument.
- Make sure you specify the appropriate container. This should likely match whatever you're already using within Posit Workbench. 

A user's home directory is mounted to the container automatically, but if you're referencing data outside of your home directory then you may have to add that path to the `mount` parameter. This needs to follow a specific structure to tell the container how the drive should be mounted. For example:

```{r mount_example, eval=FALSE}
list(
  volumes=list(
    list(name='mount1', nfs=list(
      server='0.0.0.0',
      path='/mnt/remote/mount1')
      )
    ),
  volumeMounts=list(
    list(name='mount1', mountPath='/mnt/local/mount1')
    )
  )
)
```

In this example, the `volumes` element references the physical server that needs to be mounted, and the targeted folder from the server that is being mounted. The `volumeMounts` element indicates the path that will be present locally within the container session. This example represents a standard NFS mount. More complex mount types, such as SMB, are not supported at this time. 

## Deployment Prep

Before you deploy to connect, you'll need to make sure that you have a Posit Connect API key. Note that you'll need at least a role of Publisher to do this (and to actually deploy to Connect itself). You can do this by following the instructions in the Posit Connect documentation.

With that in place, make sure you test our your job locally before deploying to connect. To do this, you'll need the **abba** API URL hosted in your environment, which you can get from your Posit support team (or by looking through content visible to you on Posit Connect). 

To test the job, we recommend setting two environment variables, as you'll be configuring these once deployed to connect:

- `ABBA_API_ADDRESS`, set to the URL of the **abba** API in your environment
- `ABBA_API_KEY`, set to the key you retrieved from Posit Connect. Make sure you keep this key secure. 

You can set these locally as follows:

```{r set_env_vars, eval=FALSE}
Sys.setenv(
  ABBA_API_ADDRESS = "https://your-connect-server.com/abba_apo",
  ABBA_API_KEY = "<your API key>"
)
```

Once set, knit the R Markdown file to make sure it executes.

## Deploy a Batch File to Connect

Now you can deploy your job to Posit Connect. This can be done following a standard R Markdown deployment process, either using the [publish button within the RStudio IDE](https://docs.posit.co/connect/how-to/publish-r-markdown/), or using [git backed content](https://docs.posit.co/connect/user/git-backed/) if you'd like to follow a more controlled deployment process. 

Once the Markdown file is in connect, you'll need to set the environment variables we specified above within the deployed environment. 

![Environment variable panel](images/env_vars.png)
Once you the environment variables are configured, you can hit the refresh button (visible in the screenshot above) to rerun the Markdown job and make sure it works. 

## Schedule Your Cron Job

Once your job is executing successfully, you can configure the run schedule. To do this, go to the Schedule tab of the deployed Markdown file on Connect.

![Schedule Panel](images/schedule_1.png)

Click the "Schedule output for default" box and then fill out the rest of the form:

![Schedule Panel Options](images/schedule_2.png)

---
title: "Kubernetes Job Submission"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Kubernetes Job Submission}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(abba)
```

In support for Kubernetes, **abba** has several local submission functions to manage interface with a cluster. At a high level, these functions break down into three different categories:

- Submit a job
- Get job status
- Retrieve the log of the job

Functions specifically for interfacing with Kubernetes follow the naming convention `abba_*_k8s_*`, and more generally functions intended for local submission (i.e. not interface with the *abba* API) are suffixed with `_local()`. 

## Submit a Job
Job submission is handled by the function `abba_submit_k8s_job_local()`. This is the function used to submit a job on the API side.

```{r submit_job, eval=FALSE}
abba_submit_k8s_job_local(
  "/home/user/programs/my_program.R",
  container = "registry.example.com/my_image:latest"
)
# job.batch/test-program-fbc3131d-a850-4081-8224-c06ef32821da created
# $job_id
# [1] "test-program-fbc3131d-a850-4081-8224-c06ef32821da"
# 
# $batch_id
# [1] "test_program-62e382a9-c1a7-4527-a32d-46600ee9ff80"
```

Note that this returns two fields: `job_id` and `batch_id`. The Job ID is the main field to consider and the identifier used downstream to interact with Kubernetes. Batch ID is a configurable identifier to group work within Kubernetes, and this is a placeholder for intended future functions that will help with the submission of several programs in a bundle.

## Get Job Status

Once the job is running in Kubernetes, **abba** can poll its status. This can be done using the function `abba_get_k8s_job_status_local()`

```{r get_status, eval=FALSE}
abba_get_k8s_job_status_local("test-program-fbc3131d-a850-4081-8224-c06ef32821da")
# $Succeeded
# $Succeeded$Jobs
# $Succeeded$Jobs[[1]]
# $Succeeded$Jobs[[1]]$id
# [1] "test-program-fbc3131d-a850-4081-8224-c06ef32821da-2dbr8"
# 
# $Succeeded$Jobs[[1]]$path
# [1] "/home/mike.stackhouse/repos/abba/test_programs/test_program.R"
# 
# 
# 
# $Succeeded$Description
# [1] "All containers in the Pod have terminated in success, and will not be restarted."
```

The job statuses can take on the following values:

| Status    | **Description**                                                                                                                                                                                                                                                          |
|-----------|--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Pending   | The Pod has been accepted by the Kubernetes cluster, but one or more of   the containers has not been set up and made ready to run. This includes time   a Pod spends waiting to be scheduled as well as the time spent downloading   container images over the network. |
| Running   | The Pod has been bound to a node, and all of the containers have been   created. At least one container is still running, or is in the process of   starting or restarting.                                                                                              |
| Succeeded | All containers in the Pod have terminated in success, and will not be   restarted.                                                                                                                                                                                       |
| Failed    | All containers in the Pod have terminated, and at least one container has   terminated in failure. That is, the container either exited with non-zero   status or was terminated by the system.                                                                          |
| Unknown   | For some reason the state of the Pod could not be obtained. This phase   typically occurs due to an error in communicating with the node where the Pod   should be running.                                                                                              |

The API interface methods will keep watching until they see a status other than "Pending" or "Running". "Succeeded", "Failed", or "Unknown" imply the job is "complete" in some watch. 

# Get the Log Content

In this context, the "log" refers to the stdout/stderr of the program itself. This returns into a list object with the Job ID and the individual lines written out by the program. Additionally, this could be some sort of failure in the pod. If the R program executes to completion, the content will contain all of the console output from the program itself. 

```{r watch_job, eval=FALSE}
abba_get_k8s_job_log_local("test-program-fbc3131d-a850-4081-8224-c06ef32821da")
# [[1]]
# [[1]]$job_id
# [1] "test-program-fbc3131d-a850-4081-8224-c06ef32821da"
# 
# [[1]]$log
# [1] "[1] 4"                   "Warning message:"        "This is a test warning " "La-di-da"                        
```

# Interacting with Kubernetes

Physical submission of work into Kubernetes is done using system calls to the `kubectl` utility. Note that for this to function, the calling user account (importantly, the user account executing the **abba** API), must have `kubectl` configured for the intended Kubernetes cluster. These local function will **not* work if `kubectl` is either not configured or not authorized. 

---
title: "Workbench Batch Jobs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Workbench Batch Jobs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(abba)
```

# End User Facing Functions

The main **abba** function to submit batches is `abba_submit_batch()`. It allows user to submit a number of jobs to run in order defined by structure of input list using runner functions supplied by user. In the example below, `r_job_test.R` and `r_job_test2.R` will run in parallel, and the `r_job_test3.R` will execute after both `r_job_test.R` and `r_job_test2.R` execution is finished. 

```{r submit_batch, eval=FALSE}
abba_submit_batch(
  list(c("/home/mike.stackhouse/test_progs/r_job_test.R",
         "/home/mike.stackhouse/test_progs/r_job_test2.R"),
       "/home/mike.stackhouse/test_progs/r_job_test3.R"),
  submit_func=abba_rslauncher_submit_job_local,
  wait_func=abba_rslauncher_watch_job_local,
  log_path="/home/mike.stackhouse/test_logs",
  timeout_seconds = 15
  )
#  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R 
#  "TG9jYWw6UitNSWpxN2FaQ3JyQXRqbGQ5TzVFUT09"  "TG9jYWw6eGVremgtbGRNMFF0VDFFVG5MTEEzdz09"  "TG9jYWw6eENZb2lBR2x1RVJyZVYyelo0Z040QT09" 
```

The functions that submit programs and wait for program execution should be specified by the user. This gives the ability to run program in batches on all platforms supported by abba. Alternatively, user can provide its own functions. All additional arguments besides program list will be passed to submit_func(in the example above, `timeout_seconds` parameter is then passed to and used in `abba_rslauncher_watch_job_local()` function to not wait too long before returning job IDs).

# Create a batch list from dataset 


To automatically calculate batch run_group order that defines order in which programs will be executed, `calculate_run_group()` function can be used: 

```{r create_batch_list, eval=FALSE}
ds <- as.data.frame(list(program_name=c("/home/yevhenii.boiko/k8s_test/r_job_test.R", "/home/yevhenii.boiko/k8s_test/r_job_test2.R", "/home/yevhenii.boiko/k8s_test/r_job_test3.R"),
                         inputs=c("raw.DM", "raw.EX,sdtm.DM", "raw.AE,sdtm.DM"),
                         outputs=c("sdtm.DM", "sdtm.EX", "sdtm.AE")))
calculate_run_group(ds, col_name='run_group')
#                                  program_name         inputs outputs run_group
# 1  /home/yevhenii.boiko/k8s_test/r_job_test.R         raw.DM sdtm.DM         1
# 2 /home/yevhenii.boiko/k8s_test/r_job_test2.R raw.EX,sdtm.DM sdtm.EX         2
# 3 /home/yevhenii.boiko/k8s_test/r_job_test3.R raw.AE,sdtm.DM sdtm.AE         2
```

This creates an additional(name configurable) column in input dataset that will group programs for execution using `inputs` and `outputs` values provided by user. 

This dataset can be supplied to `abba_submit_batch` for execution in the order of `run_group` column:

```{r example_batch_execution, eval=FALSE}
abba_submit_batch(calculate_run_group(ds, col_name='run_group'),
                  submit_func=abba_rslauncher_submit_job_local,
                  wait_func=abba_rslauncher_watch_job_local,
                  log_path="/home/mike.stackhouse/test_logs"
  )
#  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R 
#  "TG9jYWw6UGVPUVFoeVVMKys2c3M0TnZZcm9ZUT09"  "TG9jYWw6VmgrUnlueDBOLXZuVkVIYWxRa1NEZz09"  "TG9jYWw6bGlWcjdsVTRwQzIwcmlLUThzUUN4UT09" 
```

Batch log can be then obtained:

```{r get_batch_log, eval=FALSE}
abba_rslauncher_get_job_log_local(c("TG9jYWw6UGVPUVFoeVVMKys2c3M0TnZZcm9ZUT09", "TG9jYWw6VmgrUnlueDBOLXZuVkVIYWxRa1NEZz09", "TG9jYWw6bGlWcjdsVTRwQzIwcmlLUThzUUN4UT09"))
# [[1]]
# [1] "[1] \"Program r_job_test start run timestamp: 2024-03-28 16:48:39\"" "[1] \"Program r_job_test end run timestamp: 2024-03-28 16:48:40\""  
# [3] "Warning message:"                                                    "No useful work done here. "                                         
# 
# [[2]]
# [1] "[1] \"Program r_job_test2 start run timestamp: 2024-03-28 16:48:41\"" "[1] \"Program r_job_test2 end run timestamp: 2024-03-28 16:48:42\""  
# [3] "Warning message:"                                                     "No useful work done here. "                                          
# 
# [[3]]
# [1] "[1] \"Program r_job_test3 start run timestamp: 2024-03-28 16:48:41\"" "[1] \"Program r_job_test3 end run timestamp: 2024-03-28 16:48:42\""  
# [3] "Warning message:"                                                     "No useful work done here. "                       
```

To run programs sequentially regardless of supplied grouping, `sequential` parameter can be set to `TRUE`: 

```{r example_batch_execution_sequential, eval=FALSE}
abba_submit_batch(calculate_run_group(ds, col_name='run_group'),
                  submit_func=abba_rslauncher_submit_job_local,
                  wait_func=abba_rslauncher_watch_job_local,
                  log_path="/home/mike.stackhouse/test_logs",
                  sequential=TRUE
  )
#  /home/yevhenii.boiko/k8s_test/r_job_test.R /home/yevhenii.boiko/k8s_test/r_job_test2.R /home/yevhenii.boiko/k8s_test/r_job_test3.R 
#  "TG9jYWw6cGppOTZSOXgtR3FDbGdpZGtzaVh1UT09"  "TG9jYWw6VlFhVEs4MXNlWmRmOUh1dG9PeEZFZz09"  "TG9jYWw6Um5YbzY5ZEU5bHN0ZlRTMVBZNm1DZz09" 

abba_rslauncher_get_job_log_local(c("TG9jYWw6cGppOTZSOXgtR3FDbGdpZGtzaVh1UT09","TG9jYWw6VlFhVEs4MXNlWmRmOUh1dG9PeEZFZz09","TG9jYWw6Um5YbzY5ZEU5bHN0ZlRTMVBZNm1DZz09" ))
# [[1]]
# [1] "[1] \"Program r_job_test start run timestamp: 2024-03-28 16:52:01\"" "[1] \"Program r_job_test end run timestamp: 2024-03-28 16:52:02\""  
# [3] "Warning message:"                                                    "No useful work done here. "                                         
# 
# [[2]]
# [1] "[1] \"Program r_job_test2 start run timestamp: 2024-03-28 16:52:04\"" "[1] \"Program r_job_test2 end run timestamp: 2024-03-28 16:52:05\""  
# [3] "Warning message:"                                                     "No useful work done here. "                                          
# 
# [[3]]
# [1] "[1] \"Program r_job_test3 start run timestamp: 2024-03-28 16:52:06\"" "[1] \"Program r_job_test3 end run timestamp: 2024-03-28 16:52:07\""  
# [3] "Warning message:"                                                     "No useful work done here. "                                          
```

# Error generates properly when container name is invalid

    The container bad_name is not an permitted image. Permitted images are:
    	- path.io/container-1.0.0
    	- path.io/container2-1.0.0
    	- path.io/container3-1.0.0

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_slurm.R
\name{abba_slurm_watch_job}
\alias{abba_slurm_watch_job}
\title{Watch SLURM job, periodically polling its execution status.}
\usage{
abba_slurm_watch_job(
  job_id = "",
  poll_interval_seconds = 3,
  timeout_seconds = 300,
  ...
)
}
\arguments{
\item{job_id}{job ID that was specified when submitting jobs}

\item{poll_interval_seconds}{Time interval for polling job status in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{...}{other positional/keyword arguments that will be ignored}
}
\value{
a list of job ID(s) and status(es)
}
\description{
Watch SLURM job, periodically polling its execution status.
}
\examples{
\dontrun{
result <- abba_slurm_watch_job("5195", 10, 3000)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_slurm.R
\name{abba_slurm_get_job_status}
\alias{abba_slurm_get_job_status}
\title{Return descriptive job status for slurm jobs}
\usage{
abba_slurm_get_job_status(job_ids, ...)
}
\arguments{
\item{job_ids}{a list/vector of Slurm job IDs}

\item{...}{other positional/keyword arguments that will be ignored}
}
\value{
a named character vector with job statuses as values and job IDs as names.
}
\description{
Return descriptive job status for slurm jobs
}
\examples{
\dontrun{
job_statuses <- slurm_get_job_status(c('job-id-1', 'job-id-2'))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_workbench.R
\name{abba_rslauncher_watch_job_local}
\alias{abba_rslauncher_watch_job_local}
\title{Periodically poll Workbench jobs for status and return their IDs when all job
statuses arrive at 'Finished' state}
\usage{
abba_rslauncher_watch_job_local(
  job_ids,
  poll_interval_seconds = 1,
  timeout_seconds = 300,
  ...
)
}
\arguments{
\item{job_ids}{a list/vector of Workbench job IDs}

\item{poll_interval_seconds}{how often job statuses should be updated}

\item{timeout_seconds}{maximum amount of time in seconds after which job IDs will be
returned regardless of job statuses}

\item{...}{other positional/keyword arguments that will be ignored}
}
\value{
a vector of job IDs
}
\description{
Periodically poll Workbench jobs for status and return their IDs when all job
statuses arrive at 'Finished' state
}
\examples{
\dontrun{
job_statuses <- abba_rslauncher_watch_job_local(c('job-id-1', 'job-id-2'))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_workbench.R
\name{abba_rslauncher_get_job_succeeded_local}
\alias{abba_rslauncher_get_job_succeeded_local}
\title{Check whether Workbench jobs have been fully executed.}
\usage{
abba_rslauncher_get_job_succeeded_local(job_ids, ...)
}
\arguments{
\item{job_ids}{a list/vector of Workbench job IDs}

\item{...}{other positional/keyword arguments that will be ignored}
}
\value{
a named boolean vector. FALSE value indicates that job did not fully execute
}
\description{
Check whether Workbench jobs have been fully executed.
}
\examples{
\dontrun{
job_statuses <- abba_rslauncher_get_job_succeeded_local(c('job-id-1', 'job-id-2'))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{abba_watch_k8s_batch_local}
\alias{abba_watch_k8s_batch_local}
\title{Watch a K8S batch that has been submitted to Workbench, periodically polling it's execution status.}
\usage{
abba_watch_k8s_batch_local(
  batch_group_id = "",
  poll_interval_seconds = 3,
  timeout_seconds = 600,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{batch_group_id}{Batch ID that was specified when submitting a group of jobs}

\item{poll_interval_seconds}{Time interval for polling batch status in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{namespace}{Kubernetes namespace}
}
\value{
a list of jobs IDs and statuses that belong to batch named batch_group_id
}
\description{
Watch a K8S batch that has been submitted to Workbench, periodically polling it's execution status.
}
\examples{
\dontrun{
result <- abba_watch_k8s_batch_local("safety-tfls-f0bf6848-46de-45b8-9fae-0e732b104760", 10, 3000)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{abba_submit_k8s_job_local}
\alias{abba_submit_k8s_job_local}
\title{Submit an R program for execution on a Kubernetes cluster}
\usage{
abba_submit_k8s_job_local(
  file_path,
  batch_group_id = "",
  user_tag = "",
  cpu_limit = 1L,
  memory_limit = "512M",
  container = getOption("abba.default.container"),
  mounts = "",
  auto_mount_home = FALSE,
  home_nfs_ip_address = getOption("abba.home.nfs.ip.address"),
  namespace = getOption("abba.k8s.namespace"),
  username = NULL
)
}
\arguments{
\item{file_path}{Full path to R file}

\item{batch_group_id}{Group ID for batch processing}

\item{user_tag}{Optional; a string that describes what kind of job will be scheduled to run}

\item{cpu_limit}{Maximum number of cores available for Kubernetes container}

\item{memory_limit}{Maximum amount of RAM available for Kubernetes container}

\item{container}{A valid container image name provided as a character string. Defaults to the option abba.default.container.}

\item{mounts}{Specifically formatted list with information bout volumes that container would have access to during the run}

\item{auto_mount_home}{set to TRUE to mount service user home directory}

\item{home_nfs_ip_address}{IP address for mounting service user home directory}

\item{namespace}{Kubernetes namespace to put the job in}

\item{username}{user whose authority will be used to run the program}
}
\value{
A list with job_id and batch_id attributes in case of successful submission
}
\description{
Submit an R program for execution on a Kubernetes cluster
}
\examples{
\dontrun{
job_info <- abba_submit_k8s_job_local("path/to/your/program.R", batch_group_id='SDTM')
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_slurm.R
\name{abba_slurm_submit_job}
\alias{abba_slurm_submit_job}
\title{Submit R program as a SLURM job}
\usage{
abba_slurm_submit_job(
  program_path,
  log_path,
  r_version = NULL,
  user_tag = NULL,
  cpu_cores = getOption("abba.slurm.cpu.cores"),
  memory = getOption("abba.slurm.memory"),
  username = NULL,
  working_dir = NULL,
  job_timeout = 3600,
  ...
)
}
\arguments{
\item{program_path}{Full path to the R program file. Must be accessible from the SLURM node}

\item{log_path}{Parent folder for the program's log file. Required; must
be supplied explicitly so that logs are never written to an unexpected
location in the user's filespace.}

\item{r_version}{Version of R that will be used to run the program. Can be specified as a full path to Rscript executable, or as a label of R version that is displayed in the Workbench GUI.}

\item{user_tag}{custom string that will be added to the job name.}

\item{cpu_cores}{Amount of CPU cores that will be requested for the job.}

\item{memory}{Amount of RAM in megabytes that will be requested for the job.}

\item{username}{user whose permission level is used to execute the script. Defaults to user submitting the job.}

\item{working_dir}{working directory for the SLURM job. Defaults to parent directory of \code{program_path}.}

\item{job_timeout}{time limit for a job. Must be specified in a format of "days-hours:minutes:seconds" If exceeded, job will be cancelled.}

\item{...}{additional arguments (currently unused)}
}
\value{
job ID
}
\description{
Submit R program as a SLURM job
}
\examples{
\dontrun{
job_id <- abba_slurm_submit_job("/home/user/tfl/t1_dm.sas",
                                log_path = "/home/user/tfl/logs")
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_workbench.R
\name{abba_rslauncher_get_job_status_local}
\alias{abba_rslauncher_get_job_status_local}
\title{Get Workbench job status for a given vector/list of job IDs}
\usage{
abba_rslauncher_get_job_status_local(job_ids, ...)
}
\arguments{
\item{job_ids}{A list/vector of job IDs}

\item{...}{other positional/keyword arguments that will be ignored}
}
\value{
a vector of job ID statuses
}
\description{
Get Workbench job status for a given vector/list of job IDs
}
\examples{
\dontrun{
job_statuses <- abba_rslauncher_get_job_status_local(c('job-id-1', 'job-id-2'))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_job_log.R
\name{abba_get_k8s_batch_log_local}
\alias{abba_get_k8s_batch_log_local}
\title{Return list of logs for jobs that are marked with a given batch ID}
\usage{
abba_get_k8s_batch_log_local(
  batch_id,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{batch_id}{string containing batch ID}

\item{namespace}{Kubernetes namespace to search for batch jobs}
}
\value{
Job logs in a from of list consisting of character vectors
}
\description{
Return list of logs for jobs that are marked with a given batch ID
}
\examples{
\dontrun{
logs <- abba_get_k8s_batch_log_local("batch-sdtm-abc123")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{abba_get_k8s_batch_status_local}
\alias{abba_get_k8s_batch_status_local}
\title{Get status of all jobs in a batch}
\usage{
abba_get_k8s_batch_status_local(
  batch_id,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{batch_id}{unique identifier for a batch}

\item{namespace}{Kubernetes namespace}
}
\value{
list of statuses for every job in a batch
}
\description{
Get status of all jobs in a batch
}
\examples{
\dontrun{
status <- abba_get_k8s_batch_status_local("batch-sdtm-abc123")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_slurm.R
\name{abba_slurm_get_job_log}
\alias{abba_slurm_get_job_log}
\title{Return job log for each submitted job ID}
\usage{
abba_slurm_get_job_log(job_ids, ...)
}
\arguments{
\item{job_ids}{valid SLURM job IDs}

\item{...}{other parameters that will be ignored}
}
\value{
list of character vectors. each character vector would represent program log.
}
\description{
Return job log for each submitted job ID
}
\examples{
\dontrun{
job_log <- abba_slurm_get_job_log("156")}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_requests.R
\name{abba_get_batch_log}
\alias{abba_get_batch_log}
\title{Send GET request to get logs of all jobs in a batch}
\usage{
abba_get_batch_log(
  batch_id,
  api_address = Sys.getenv("ABBA_API_ADDRESS"),
  api_key = Sys.getenv("ABBA_API_KEY")
)
}
\arguments{
\item{batch_id}{unique batch identificator}

\item{api_address}{URL to send requests to, hosted in Posit Connect. Defaults
to environment variable ABBA_API_ADDRESS.}

\item{api_key}{API Key for accessing restricted endpoints. Defaults to
environment variable ABBA_API_KEY}
}
\value{
body of request`s response in a list format
}
\description{
Send GET request to get logs of all jobs in a batch
}
\examples{
\dontrun{
response <- abba_get_job_log('1234j-13j4l5k-ajslfd')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_requests.R
\name{abba_wait_for_job_log}
\alias{abba_wait_for_job_log}
\title{Monitor job status and retrieve its log when the job finishes running}
\usage{
abba_wait_for_job_log(
  job_id,
  poll_interval_seconds = 3,
  timeout_seconds = 600,
  api_address = Sys.getenv("ABBA_API_ADDRESS"),
  api_key = Sys.getenv("ABBA_API_KEY"),
  ...
)
}
\arguments{
\item{job_id}{unique job identificator}

\item{poll_interval_seconds}{Total time to wait before timeout in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{api_address}{URL to send requests to, hosted in Posit Connect. Defaults
to environment variable ABBA_API_ADDRESS.}

\item{api_key}{API Key for accessing restricted endpoints. Defaults to
environment variable ABBA_API_KEY}

\item{...}{Other arguments that will be ignored}
}
\value{
list with 2 attributes: job_id for submitted job`s id, and its logs
}
\description{
Monitor job status and retrieve its log when the job finishes running
}
\examples{
\dontrun{
response <- abba_wait_for_job_log('sdfj4-asdjlk-bjslk')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{abba_get_k8s_job_status_local}
\alias{abba_get_k8s_job_status_local}
\title{Get status of all pods that belong to a job}
\usage{
abba_get_k8s_job_status_local(
  job_id,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{job_id}{unique identifier for a job}

\item{namespace}{Kubernetes namespace}
}
\value{
list of statuses for every pod in a job(typically just one).
}
\description{
Get status of all pods that belong to a job
}
\examples{
\dontrun{
status <- abba_get_k8s_job_status_local("job-sdtm-abc123")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_requests.R
\name{abba_get_job_log}
\alias{abba_get_job_log}
\title{Send GET request to get logs of specified Jobs}
\usage{
abba_get_job_log(
  job_ids,
  api_address = Sys.getenv("ABBA_API_ADDRESS"),
  api_key = Sys.getenv("ABBA_API_KEY")
)
}
\arguments{
\item{job_ids}{A list of job IDs to get logs for}

\item{api_address}{URL to send requests to, hosted in Posit Connect. Defaults
to environment variable ABBA_API_ADDRESS.}

\item{api_key}{API Key for accessing restricted endpoints. Defaults to
environment variable ABBA_API_KEY}
}
\value{
body of request`s response in a list format
}
\description{
Send GET request to get logs of specified Jobs
}
\examples{
\dontrun{
response <- abba_get_job_log('1234j-13j4l5k-ajslfd')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{submit_k8s_yaml}
\alias{submit_k8s_yaml}
\title{Submit a job profile for execution on a Kubernetes cluster}
\usage{
submit_k8s_yaml(yaml_full_path)
}
\arguments{
\item{yaml_full_path}{Path to YAML config file}
}
\value{
A string containing Job ID
}
\description{
Submit a job profile for execution on a Kubernetes cluster
}
\examples{
\dontrun{
job_id <- submit_k8s_yaml("/tmp/my_job.yaml")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_k8s.R
\name{abba_watch_k8s_job_local}
\alias{abba_watch_k8s_job_local}
\title{Watch a K8S job that has been submitted to Workbench, periodically polling it's execution status.}
\usage{
abba_watch_k8s_job_local(
  job_id = "",
  poll_interval_seconds = 3,
  timeout_seconds = 600,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{job_id}{Job ID. Typically obtained as a return value from submit_job and similar functions}

\item{poll_interval_seconds}{Time interval for polling job status in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{namespace}{Kubernetes namespace}
}
\value{
a list of pods, their IDs and execution statuses
}
\description{
Watch a K8S job that has been submitted to Workbench, periodically polling it's execution status.
}
\examples{
\dontrun{
result <- abba_watch_k8s_job_local("job-sdtm-f0bf6848-46de-45b8-9fae-0e732b104760", 10, 3000)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_batch.R
\name{create_batch_api}
\alias{create_batch_api}
\alias{create_batch_job}
\title{Create a Batch API or Job file template}
\usage{
create_batch_api(path)

create_batch_job(path)
}
\arguments{
\item{path}{A file path where the target file will be created. Must be
supplied explicitly; no default is provided so that files are never
written to an unexpected location.}
}
\value{
No return value, called for side effects (file creation).
}
\description{
This function create a batch API or job file template at the specified
location. The Batch API file template is a plumber API with the necessary
REST API server side to interface with the \pkg{abba} package. This simplifies
the process of setting up the receiver API for which jobs are submitted. The
job template file is a markdown file with the necessary function calls to run
a batch job.
}
\details{
Note that to deploy an R api to Posit Connect, the file must be named
plumber.R.
}
\examples{
# Write the template to a temporary directory
create_batch_api(tempdir())
create_batch_job(tempdir())

\dontrun{
create_batch_api("~/api_directory")
create_batch_api("~/api_directory/plumber.R")

create_batch_job("~/job_directory")
create_batch_job("~/job_directory/my_job.Rmd")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_dataset.R
\name{calculate_run_group}
\alias{calculate_run_group}
\title{Calculate run_group variable using inputs and outputs of programs supplied by user}
\usage{
calculate_run_group(x, col_name = "run_group")
}
\arguments{
\item{x}{input data frame. Must contain columns 'inputs', 'outputs' that list
input/output datasets for each program}

\item{col_name}{name of newly created variable. Defaults to 'run_group_calculated'}
}
\value{
an input data frame with one new column
}
\description{
Calculate run_group variable using inputs and outputs of programs supplied by user
}
\examples{
input_ds <- as.data.frame(list(program_name=c('prog1.R', 'prog2.R'),
                               inputs=c('ds0.xpt', 'ds1.xpt'),
                               outputs=c('ds1.xpt', 'ds2.xpt')))
batch_ready <- calculate_run_group
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_requests.R
\name{abba_wait_for_batch_log}
\alias{abba_wait_for_batch_log}
\title{Monitor batch status and retrieve its log when the all jobs in batch finish
running}
\usage{
abba_wait_for_batch_log(
  batch_id,
  poll_interval_seconds = 3,
  timeout_seconds = 600,
  api_address = Sys.getenv("ABBA_API_ADDRESS"),
  api_key = Sys.getenv("ABBA_API_KEY")
)
}
\arguments{
\item{batch_id}{unique batch identificator}

\item{poll_interval_seconds}{Total time to wait before timeout in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{api_address}{URL to send requests to, hosted in Posit Connect. Defaults
to environment variable ABBA_API_ADDRESS.}

\item{api_key}{API Key for accessing restricted endpoints. Defaults to
environment variable ABBA_API_KEY}
}
\value{
list with 2 sublists: job_ids and their logs
}
\description{
Monitor batch status and retrieve its log when the all jobs in batch finish
running
}
\examples{
\dontrun{
response <- abba_wait_for_batch_log('batch-sdtm-sdfj4-asdjlk-bjslk')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_job_log.R
\name{abba_get_k8s_job_log_local}
\alias{abba_get_k8s_job_log_local}
\title{Get log for every job specified in an input vector/list}
\usage{
abba_get_k8s_job_log_local(
  job_ids,
  namespace = getOption("abba.k8s.namespace")
)
}
\arguments{
\item{job_ids}{A list of job IDs to get logs for}

\item{namespace}{Kubernetes namespace to search for job}
}
\value{
A list of job logs. Each list entry will contain complete log for a job
}
\description{
Get log for every job specified in an input vector/list
}
\examples{
\dontrun{
logs <- abba_get_k8s_job_log_local(c("job-sdtm-abc123", "job-adam-def456"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api_requests.R
\name{abba_submit_and_get_log}
\alias{abba_submit_and_get_log}
\title{Send job and poll for status. This function sends multiple requests so it
won't time out on heavy jobs}
\usage{
abba_submit_and_get_log(
  file_path,
  batch_group_id = "",
  user_tag = "",
  cpu_limit = 1L,
  memory_limit = "512M",
  container = "",
  mounts = "",
  auto_mount_home = FALSE,
  home_nfs_ip_address = getOption("abba.home.nfs.ip.address"),
  poll_interval_seconds = 3,
  timeout_seconds = 600,
  api_address = Sys.getenv("ABBA_API_ADDRESS"),
  api_key = Sys.getenv("ABBA_API_KEY")
)
}
\arguments{
\item{file_path}{Full path to R file}

\item{batch_group_id}{Group ID for batch processing}

\item{user_tag}{Optional; a string that describes what kind of job will be
scheduled to run}

\item{cpu_limit}{Maximum number of cores available for Kubernetes container}

\item{memory_limit}{Maximum amount of RAM available for Kubernetes container}

\item{container}{list that contains container name and image name}

\item{mounts}{Specifically formatted list with information bout volumes that
container would have access to during the run}

\item{auto_mount_home}{set to TRUE to mount service user home directory}

\item{home_nfs_ip_address}{IP address for mounting service user home directory}

\item{poll_interval_seconds}{Total time to wait before timeout in seconds}

\item{timeout_seconds}{Total time to wait before timeout in seconds}

\item{api_address}{URL to send requests to, hosted in Posit Connect. Defaults
to environment variable ABBA_API_ADDRESS.}

\item{api_key}{API Key for accessing restricted endpoints. Defaults to
environment variable ABBA_API_KEY.}
}
\value{
list with 2 attributes: job_id for submitted job`s id, and its logs
}
\description{
Send job and poll for status. This function sends multiple requests so it
won't time out on heavy jobs
}
\examples{
\dontrun{
response <- abba_submit_and_get_log('/path/to/R/program.R', batch_group_id='SDTM')}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_workbench.R
\name{abba_rslauncher_submit_job_local}
\alias{abba_rslauncher_submit_job_local}
\title{Create a job for executing an R program}
\usage{
abba_rslauncher_submit_job_local(p, log_path, user_tag = "", ...)
}
\arguments{
\item{p}{path to program}

\item{log_path}{Path to the directory where the log will be saved.
Required; must be supplied explicitly so that logs are never written to an
unexpected location in the user's filespace.}

\item{user_tag}{optional; any user tags user might want to add to the job}

\item{...}{other arguments that will be passed to internal rslauncher_submit_job function}
}
\value{
job id
}
\description{
Create a job for executing an R program
}
\examples{
\dontrun{
job_id <- abba_rslauncher_submit_job_local("/path/to/program.R",
                                           log_path = "/path/to/logs")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submit_job_slurm.R
\name{abba_slurm_get_job_succeeded}
\alias{abba_slurm_get_job_succeeded}
\title{Check whether SLURM jobs have been fully executed.}
\usage{
abba_slurm_get_job_succeeded(job_ids, ...)
}
\arguments{
\item{job_ids}{a list/vector of  valid SLURM job IDs}

\item{...}{other parameters that will be ignored}
}
\value{
a named boolean vector. TRUE if job finished running and has 'COMPLETED' status, FALSE otherwise.
If job has not finished running, a NULL will be returned.
}
\description{
Check whether SLURM jobs have been fully executed.
}
\examples{
\dontrun{
job_statuses <- abba_slurm_get_job_succeeded(c('job-id-1', 'job-id-2'))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_submission.R
\name{abba_submit_batch}
\alias{abba_submit_batch}
\title{Submit programs for execution in order defined by structure of input list.
Programs inside sublists will be executed in parallel, and sublists themselves
would be submitted sequentially.}
\usage{
abba_submit_batch(
  prog_list,
  sequential = FALSE,
  submit_func = abba_rslauncher_submit_job_local,
  wait_func = abba_rslauncher_watch_job_local,
  succeed_func = abba_rslauncher_get_job_succeeded_local,
  col_name = "run_group",
  halt_on_error = TRUE,
  rerun_unchanged_programs = TRUE,
  update_cache = FALSE,
  cache_folder = getOption("abba.default_cache_folder"),
  ...
)
}
\arguments{
\item{prog_list}{A list of R program paths to execute}

\item{sequential}{when sequential=TRUE, prog_list is flattened and everything is executed sequentially.}

\item{submit_func}{function that will be used to submit jobs}

\item{wait_func}{function that checks job status and returns when job finishes executing}

\item{succeed_func}{function that returns TRUE if job finished running without errors and FALSE otherwise}

\item{col_name}{Name of the column that contains run group numbers when prog_list is a data frame}

\item{halt_on_error}{If TRUE: if prog_list contains program inputs/outputs - programs that depend on failed
program will not be executed; if prog_list contains only program paths - when program fails,
entire batch will stop executing. TRUE by default}

\item{rerun_unchanged_programs}{If FALSE: will not re-run programs whose code
and inputs have not been modified since last batch run with update_cache=TRUE.}

\item{update_cache}{if TRUE, file hash for programs and their inputs will be
calculated and written to \code{cache_folder} after the batch run. If FALSE, the
cache will not be created or updated. FALSE by default.}

\item{cache_folder}{Path to the folder where hash-sums of programs and
their inputs will be stored. Required when \code{update_cache=TRUE} or
\code{rerun_unchanged_programs=FALSE}. The default value can be set via the
\code{abba.default_cache_folder} option; abba never falls back to writing caches
under the user's filespace.}

\item{...}{arguments that will be passed to submit_func and wait_func functions}
}
\value{
list job IDs associated with executed programs
}
\description{
Submit programs for execution in order defined by structure of input list.
Programs inside sublists will be executed in parallel, and sublists themselves
would be submitted sequentially.
}
\examples{
\dontrun{
job_ids <- abba_submit_batch(list(
  c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/dm.sas"),
  c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/sdtm/ae.sas"),
  c("/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_dm.sas",
    "/mnt/work_drive/proj/comp/prot/task/development/prod/program/tfl/t1_ae.sas")))
 }
}
